<?php

/**
 * ExcelExportOptions controller
 *
 * @version 1.0
 * @author Alvaro Torterola <alvaro.torterola@fengoffice.com>
 */
class ExcelExportController extends ApplicationController {

	var $plugin_name = "advanced_core";
	
	private $can_use_phpspreadsheet = false;
	
	/**
	 * Construct the ExcelExportOptions
	 *
	 * @access public
	 * @param void
	 */
	function __construct() {
		parent::__construct();
		prepare_company_website_controller($this, 'website');
		Env::useHelper('reporting');
		
		$this->can_use_phpspreadsheet = self::can_use_php_spreadsheet_lib();
		
		if ($this->can_use_phpspreadsheet) {
			Env::useLibrary('PhpSpreadsheet', 'advanced_core');
		} else {
			Env::useLibrary('PHPExcel', 'advanced_core');
		}
		
	} // __construct
	
	
	/**
	 * Returns true if all requirements are met to use PhpSpreadsheet library
	 * @return boolean
	 */
	static function can_use_php_spreadsheet_lib() {
		
		$requirements = array(
				'PHP 7.1.0' => version_compare(PHP_VERSION, '7.1.0', '>='),
				'PHP extension XML' => extension_loaded('xml'),
				'PHP extension xmlwriter' => extension_loaded('xmlwriter'),
				'PHP extension mbstring' => extension_loaded('mbstring'),
				'PHP extension ZipArchive' => extension_loaded('zip'),
				'PHP extension GD (optional)' => extension_loaded('gd'),
				'PHP extension dom (optional)' => extension_loaded('dom'),
		);
		
		$ok = true;
		
		foreach ($requirements as $req) {
			if (!$req) $ok = false;
		}
		
		return $ok;
	}
	
	
	function export_custom_report_excel() {
		ini_set('memory_limit', '1G');
		
		ajx_current("empty");
		$report_id = get_id();
		
		$report = Reports::findById($report_id);
		$filename = $report->getObjectName().".xlsx";

		$file_path = $this->generate_excel_file_from_report($report);
		$size = filesize($file_path);
		
		ajx_extra_data(array("filename" => $filename));
	}
	
	
	function generate_excel_file_from_report(Report $report) {
		// report params
		$params_str = array_var($_REQUEST, 'report_params');
		if ($params_str) $params = json_decode(str_replace("'", '"',$params_str), true);
		$_GET['limit'] = 0;
		
		// execute report
		$rc = new ReportingController();
		$_REQUEST['limit'] = 0;
		$results = $rc->view_custom_report($report->getId(), null, $params);

		// get report result in arrays
		$all_data = report_values_to_arrays($results, $report);
		$headers = array_var($all_data, 'headers');
		$all_data_rows = array_var($all_data, 'values');

		// build a sheet for the excel file with the report data 
		$excel_report_sheet_options = array('title' => $report->getObjectName());
		Hook::fire('override_excel_custom_report_sheet_options', array('report' => $report), $excel_report_sheet_options);
		$sheets = array(
			array('headers' => $headers, 'data' => $all_data_rows, 'options' => $excel_report_sheet_options),
		);
		
		$excel_report_options = array('filename' => $report->getObjectName().".xlsx");
		Hook::fire('override_excel_custom_report_options', array('report' => $report), $excel_report_options);
		
		// generate the excel file
		$filepath = $this->generate_excel_file($sheets, $excel_report_options);
		
		return $filepath;
	}

	// Auxiliar functions for excel export -------------------------------------
	
	private function get_spreadsheet_writer($spreadsheet) {
		
		if ($this->can_use_phpspreadsheet) {
			$writer = PhpOffice\PhpSpreadsheet\IOFactory::createWriter($spreadsheet, 'Xlsx');
			
		} else {
			// use this zip class because php7 may not include the ZipArchive class by default
			PHPExcel_Settings::setZipClass(PHPExcel_Settings::PCLZIP);
			$writer = PHPExcel_IOFactory::createWriter($spreadsheet, 'Excel2007');
		}
		
		return $writer;
	}
	
	private function create_spreadsheet_object() {
		
		if ($this->can_use_phpspreadsheet) {
			$e = new PhpOffice\PhpSpreadsheet\Spreadsheet();
		} else {
			$e = new PHPExcel();
		}
		
		return $e;
	}
	
	private function get_last_column_letter_by_index($total_cols) {
		
		if ($this->can_use_phpspreadsheet) {
			$last_column_letter = PhpOffice\PhpSpreadsheet\Cell\Coordinate::stringFromColumnIndex($total_cols);
		} else {
			$last_column_letter = PHPExcel_Cell::stringFromColumnIndex($total_cols-1);
		}
		
		return $last_column_letter;
	}
	
	private function set_cell_value_by_column_and_row($sheet, $col, $row, $cval) {
		
		if ($this->can_use_phpspreadsheet) {
			$index = $col + 1;
		} else {
			$index = $col;
		}
		$sheet->setCellValueByColumnAndRow($index, $row, html_entity_decode($cval));
		
	}
	
	
	function apply_borders_to_cells($sheet, $cell_range, $border_color, $border_position) {
		if ($this->can_use_phpspreadsheet) {
			
			$border_style = PhpOffice\PhpSpreadsheet\Style\Border::BORDER_THIN;
			if ($border_position == 'top_bottom') {
				$border_style_array = array('borders' => array(
						'bottom' => array('borderStyle' => $border_style, 'color' => array('rgb' => $border_color)),
						'top' => array('borderStyle' => $border_style, 'color' => array('rgb' => $border_color)),
				));
			} else {
				$border_style_array = array('borders' => array(
						$border_position => array('borderStyle' => $border_style, 'color' => array('rgb' => $border_color)),
				));
			}
			
		} else {
			
			$border_style = PHPExcel_Style_Border::BORDER_THIN;
			if ($border_position == 'top_bottom') {
				$border_style_array = array('borders' => array(
						'bottom' => array('style' => $border_style, 'color' => array('rgb' => $border_color)),
						'top' => array('style' => $border_style, 'color' => array('rgb' => $border_color)),
				));
			} else {
				$border_style_array = array('borders' => array(
						$border_position => array('style' => $border_style, 'color' => array('rgb' => $border_color))
				));
			}
		}
		
		$sheet->getStyle($cell_range)->applyFromArray($border_style_array);
		
	}
	
	// -------------------------------------
	

	function generate_excel_file($sheet_datas = array(), $options = array()) {
		$logged_user_name = logged_user() instanceof Contact ? logged_user()->getObjectName() : '';
		$properties = array_var($options, 'properties', array());
		$keywords = array_var($properties, 'keywords', "");

		$e = $this->create_spreadsheet_object();
		
		$e->getProperties()
			->setCreator($logged_user_name)
			->setTitle(isset($sheet_datas[0]) && isset($sheet_datas[0]['options']) ? array_var($sheet_datas[0]['options'],'title',"Book 1") : "Book 1")
			->setLastModifiedBy($logged_user_name)
			->setKeywords('excel feng office fengoffice '.$keywords);
		
		$sheet_count = 0;
		foreach ($sheet_datas as $sh_data) {
			if ($sheet_count > 0) {
				$e->createSheet($sheet_count);
			}
			$sheet = $e->getSheet($sheet_count);
			
			$this->generate_excel_file_sheet($sheet, array_var($sh_data, 'headers'), array_var($sh_data, 'data'), array_var($sh_data, 'options'), array_var($sh_data, 'grouped_data'));
			$sheet_count++;
		}
		
		$writer = $this->get_spreadsheet_writer($e);
		
		$filename = array_var($options, 'filename', 'excel_'.gen_id().'xlsx');
		$filepath = ROOT.'/tmp/'.$filename;
		$writer->save($filepath);
		
		return $filepath;
	}
	

	private function generate_excel_file_sheet($sheet, $headers = array(), $data = array(), $sheet_options = array(), $grouped_data = array()) {
		$title = array_var($sheet_options, 'title');
        $conditons = array_var($sheet_options, 'conditions');
		if (strlen_utf($title) > 30) {
			$title = substr_utf($title, 0, 30);
		}
		$title = str_replace(array("*", ":", "/", "\\", "?", "[", "]"), "", $title);
		
		if ($title) $sheet->setTitle($title);
        $data_start_row = 1;

       	/* @var $sheet PhpOffice\PhpSpreadsheet\Worksheet\Worksheet */

        // report conditions
        if (is_array($conditons) && count($conditons) > 0) {
            $total_cols=0;
            foreach ($conditons as $cval) {
            	$this->set_cell_value_by_column_and_row($sheet, $total_cols, $data_start_row, $cval);
                $sheet->mergeCells('A1:E1');
                $total_cols++;
            }

            $data_start_row ++;

        }
        
        $data_groups = array();
        if (!empty($headers) || !empty($data)) {
        	$data_groups[] = array('headers' => $headers, 'data' => $data);
        }
        if (!empty($grouped_data)) {
        	$data_groups = array_merge($data_groups, $grouped_data);
        }

        foreach ($data_groups as $data_group) {
        	
        	$headers = array_var($data_group, 'headers', array());
			$data = array_var($data_group, 'data');
			$totals = array_var($data_group, 'totals');
			$totals_after_data = array_var($data_group, 'totals_after_data');
        	
        	$total_cols = 0;
        	
        	// report main headers
        	if (isset($data_group['main_header']) && count($data_group['main_header']) > 0) {
        		$total_cols=0;
        		foreach ($data_group['main_header'] as $hval) {
        			$this->set_cell_value_by_column_and_row($sheet, $total_cols, $data_start_row, $hval);
        			$total_cols++;
        		}
        		
        		$last_column_letter = $this->get_last_column_letter_by_index($total_cols);
        		$main_header = 'A'.$data_start_row.':'.$last_column_letter.$data_start_row;
        		
        		$data_start_row++;
        		
        		$main_header_options = array();
        		$main_header_options['bg_color'] = array_var($sheet_options, 'main_header_bg_color', 'ACD0A6');
        		$main_header_options['border_color'] = array_var($sheet_options, 'main_header_border_color', 'ACD0A6');
        		$main_header_options['border_position'] = array_var($sheet_options, 'main_header_border_position', 'bottom');
        		
        		$this->format_table_header($sheet, $main_header, $main_header_options);
        	}
        	
			// report headers
			if (is_array($headers) && count($headers) > 0) {
				$total_cols=0;
				
				$total_rows = $this->generate_header($sheet, $sheet_options);
				$data_start_row += $total_rows;

				foreach ($headers as $hval) {
					$hval = str_replace(",", "\n", $hval); // when one header is composed by several, separate them to allow text wrap to format the header
					$this->set_cell_value_by_column_and_row($sheet, $total_cols, $data_start_row, $hval);
					$total_cols++;
				}
				
				$last_column_letter = $this->get_last_column_letter_by_index($total_cols);
	
				$header = 'A'.$data_start_row.':'.$last_column_letter.$data_start_row;
	
				$data_start_row++;
	
				$header_options = array();
				if (isset($sheet_options['header_bg_color'])) $header_options['bg_color'] = $sheet_options['header_bg_color'];
				if (isset($sheet_options['header_border_color'])) $header_options['border_color'] = $sheet_options['header_border_color'];
				if (isset($sheet_options['header_border_position'])) $header_options['border_position'] = $sheet_options['header_border_position'];
				
				$this->format_table_header($sheet, $header, $header_options);
			}

			// report totals, if defined to be printed before the data
			if (!$totals_after_data && is_array($totals) && count($totals) > 0) {
				$total_cols=0;
				foreach ($totals as $hval) {
					$this->set_cell_value_by_column_and_row($sheet, $total_cols, $data_start_row, $hval);
					$total_cols++;
				}
	
				$last_column_letter = $this->get_last_column_letter_by_index($total_cols);
				$total = 'A'.$data_start_row.':'.$last_column_letter.$data_start_row;
	
				$data_start_row++;
	
				$total_options = array();
				$total_options['bg_color'] = array_var($sheet_options, 'total_bg_color', 'e7e7e7');
				$total_options['border_color'] = array_var($sheet_options, 'total_border_color');
				$total_options['border_position'] = array_var($sheet_options, 'total_border_position', 'top');
				
				$this->format_table_header($sheet, $total, $total_options);
			}
			
			if (is_array($data)) {
				foreach ($data as &$row) {
					$row = str_replace(array("\r\n","\r","\n"), " ", $row);
					foreach ($row as &$rowitem) {
						$decoded = html_entity_decode($rowitem);
						if ($decoded != $rowitem) {
							$decoded = utf8_encode($decoded);
						}
						$rowitem = $decoded;
					}
				}
				
				// put data after title
				$sheet->fromArray($data, ' ', 'A'.$data_start_row);
				
				$data_start_row += count($data);
			}

			// report totals, if defined to be printed after the data
			if ($totals_after_data && is_array($totals) && count($totals) > 0) {
				$total_cols=0;
				foreach ($totals as $hval) {
					$this->set_cell_value_by_column_and_row($sheet, $total_cols, $data_start_row, $hval);
					$total_cols++;
				}
	
				$last_column_letter = $this->get_last_column_letter_by_index($total_cols);
				$total = 'A'.$data_start_row.':'.$last_column_letter.$data_start_row;
	
				$data_start_row++;
	
				$total_options = array();
				$total_options['bg_color'] = array_var($sheet_options, 'total_bg_color', 'e7e7e7');
				$total_options['border_color'] = array_var($sheet_options, 'total_border_color');
				$total_options['border_position'] = array_var($sheet_options, 'total_border_position', 'bottom');
				
				$this->format_table_header($sheet, $total, $total_options);
			}
        }
        
		// set column autowidth
		try {
			$cellIterator = $sheet->getRowIterator()->current()->getCellIterator();
			$cellIterator->setIterateOnlyExistingCells(true);
			
			$fixed_cells_width = array_var($sheet_options, 'fixed_cells_width');
			
			if (!$fixed_cells_width) {
				foreach ($cellIterator as $cell) {
					$sheet->getColumnDimension($cell->getColumn())->setAutoSize(true);
				}
			} else {
				foreach ($cellIterator as $cell) {
					$sheet->getColumnDimension($cell->getColumn())->setWidth($fixed_cells_width);
				}
			}
			
		} catch (Exception $e) {
			Logger::log("Error exporting to excel when setting auto width: ".$e->getMessage());
			Logger::log($e->getMessage());
		}
	}

	private function generate_header($sheet, $sheet_options){
		$data_start_row = 0;

		$params = array();
		$params_str = array_var($_REQUEST, 'report_params');
		if ($params_str) $params = json_decode(str_replace("'", '"',$params_str), true);

		$report_id = get_id();
		$filters = build_report_conditions_html($report_id, $params);
		$filters  = str_replace(['<li>','</li>','<ul>','</ul>'],' ',$filters);


		$company_name = owner_company()->getObjectName();
		$company_logo = null;
		if (FileRepository::isInRepository(owner_company()->getPictureFile())) {
			$company_logo = FileRepository::getBackend()->getFilePath(owner_company()->getPictureFile());
		}
		$now_exel = date("d")."/".date("m")."/".date("Y");
		$name_report = $sheet_options['title'];
		$name_project = "";
		
		$context = active_context();
		foreach ($context as $selection) :
			if ($selection instanceof Member) :
				$name_project = $selection->getName();
			endif;
		endforeach;

		if ($this->can_use_phpspreadsheet) {
			$drawing = new \PhpOffice\PhpSpreadsheet\Worksheet\Drawing();
		} else {
			$drawing = new PHPExcel_Worksheet_Drawing();
		}
		$drawing->setName('Paid');
		$drawing->setDescription('Paid');
		if ($company_logo) {
			$drawing->setPath($company_logo);
		}
		$drawing->setCoordinates('A1');
		$drawing->setOffsetX(10);
		$drawing->setHeight(50);
		$drawing->setWorksheet($sheet);

		$font_style_head = [
			'font' => [ 'size' => 18, 'color' => array('rgb' => '2E4760') ]
		];
		$font_style_executed = [
			'font' => [ 'size' => 11, 'bold'  =>  true, 'color' => array('rgb' => 'B1B1B1') ]
		];

		$row = $sheet->getHighestRow()+1;
		$sheet->insertNewRowBefore($row);
		$sheet->setCellValue('A1', '');
		$sheet->setCellValue('B1', $company_name);
		$sheet->setCellValue('C1', 'Executed on');
		$sheet->getStyle("B1")->applyFromArray($font_style_head);
		$sheet->getStyle("C1")->applyFromArray($font_style_executed);
		$sheet->getRowDimension('1')->setRowHeight(40);
		
		$data_start_row++;

		$font_style_executed_date = [
			'font' => [ 'size' => 10, 'bold' => false, 'color' => array('rgb' => 'B1B1B1') ]
		];
		$row = $sheet->getHighestRow()+1;
		$sheet->insertNewRowBefore($row);
		$sheet->setCellValue('C2', $now_exel);
		$sheet->getStyle("C2")->applyFromArray($font_style_executed_date);
		$data_start_row++;

		$right_alingment = $this->can_use_phpspreadsheet ? \PhpOffice\PhpSpreadsheet\Style\Alignment::HORIZONTAL_RIGHT : PHPExcel_Style_Alignment::HORIZONTAL_RIGHT;
		$left_alingment = $this->can_use_phpspreadsheet ? \PhpOffice\PhpSpreadsheet\Style\Alignment::HORIZONTAL_LEFT : PHPExcel_Style_Alignment::HORIZONTAL_LEFT;

		$font_style_project_title = [
			'font' => [ 'size' => 16, 'bold'  =>  true, 'color' => array('rgb' => '2A9D8F') ]
		];
		$font_style_filter_title = [
			'font' => [ 'size' => 12, 'bold'  =>  true, 'color' => array('rgb' => '435762') ],
			'alignment' => ['horizontal' => $right_alingment],
		];
		
		$row = $sheet->getHighestRow()+1;
		$sheet->insertNewRowBefore($row);
		$sheet->setCellValue('A3', $name_report);
		$sheet->setCellValue('C3', 'Filters');
		$sheet->getStyle("A3")->applyFromArray($font_style_project_title);
		$sheet->getStyle("C3")->applyFromArray($font_style_filter_title);
		$data_start_row++;

		$font_style_project_sbtitle = [
			'font' => [ 'size' => 14, 'bold'  =>  true, 'color' => array('rgb' => '304551') ]
		];
		$font_style_filter_text = [
			'font' => [ 'size' => 11, 'bold' => false, 'color' => array('rgb' => '435762') ],
			'alignment' => ['horizontal' => $left_alingment],
		];

		$row = $sheet->getHighestRow()+1;
		$sheet->insertNewRowBefore($row);
		$sheet->setCellValue('A4', $name_project);
		$sheet->setCellValue('C4', $filters);
		$sheet->getStyle("A4")->applyFromArray($font_style_project_sbtitle);
		$sheet->getStyle("C4")->applyFromArray($font_style_filter_text);
		$data_start_row++;

		$report = Reports::findById($report_id);
		Hook::fire("excel_export_additional_header_content", array('sheet' => $sheet, 'report' => $report), $data_start_row);

		return $data_start_row;
	}
	
	
	private function apply_bg_color_to_cells($sheet, $cell_range, $bg_color) {
		if ($this->can_use_phpspreadsheet) {
			$fill_solid = PhpOffice\PhpSpreadsheet\Style\Fill::FILL_SOLID;
		} else {
			$fill_solid = PHPExcel_Style_Fill::FILL_SOLID;
		}
		
		$sheet->getStyle($cell_range)->getFill()->setFillType($fill_solid)->getStartColor()->setRGB($bg_color);
	}
	
	
	private function format_table_header($sheet, $cell_range, $options = array()) {
		
		$bg_color = array_var($options, 'bg_color', 'DCE6F0');
		$border_color = array_var($options, 'border_color', '333333');
		$border_position = array_var($options, 'border_position', 'top_bottom');
		
		if (!in_array($border_position, array('top','bottom','top_bottom'))) $border_position = 'bottom';
		
		// title background color
		$this->apply_bg_color_to_cells($sheet, $cell_range, $bg_color);
		
		// font style
		$font_style = array('font' => array('bold' => true));
		$sheet->getStyle($cell_range)->applyFromArray($font_style);
		
		// wrap text
		$sheet->getStyle($cell_range)->getAlignment()->setWrapText(true);
		
		// vertical alignment
		if ($this->can_use_phpspreadsheet) {
			$sheet->getStyle($cell_range)->getAlignment()->setVertical(PhpOffice\PhpSpreadsheet\Style\Alignment::VERTICAL_CENTER);
		} else {
			$sheet->getStyle($cell_range)->getAlignment()->setVertical(PHPExcel_Style_Alignment::VERTICAL_CENTER);
		}
		
		// borders
		$this->apply_borders_to_cells($sheet, $cell_range, $border_color, $border_position);
	}

} // ExcelExportController
