<?php

/***************************************************************************************/
/**************************** BASIC AND AUTH FUNCTIONS *********************************/
/***************************************************************************************/

function get_ms_access_token($user_id) {
	
	$access_token = null;

	try {

		$serialized_token = user_config_option('ms_access_token', null, $user_id);
		if ($serialized_token != '') {
			$access_token = unserialize(base64_decode($serialized_token));
			
			// verify if expired and refresh if necesary
			is_valid_microsoft_access_token($user_id, $access_token);
		}
		
	} catch (Exception $e) {
		$access_token = null;
		delete_ms_access_token($user_id);
		Logger::log_r("MS ERROR at ".__FUNCTION__.": ".$e->getMessage()."\n".$e->getTraceAsString());

	} catch (Error $e) {
		$access_token = null;
		delete_ms_access_token($user_id);
		Logger::log_r("MS ERROR at ".__FUNCTION__.": ".$e->getMessage()."\n".$e->getTraceAsString());
	}
	
	
	return $access_token;
}

function set_ms_access_token($user_id, $access_token) {
	
	$serialized_token = base64_encode(serialize($access_token));
	
	set_user_config_option('ms_access_token', $serialized_token, $user_id);
}

function delete_ms_access_token($user_id) {
	set_user_config_option('ms_access_token', null, $user_id);
}

function get_microsoft_config() {
  $client_id = config_option('ms_client_id');
	
  $callback_file = with_slash(ROOT) . config_option('ms_redirect_uri_suffix');
  $webhook_callback_file = with_slash(ROOT) .  config_option('ms_webhook_redirect_uri_suffix');
	
	if (!file_exists($callback_file)) {
		@copy(ROOT . '/plugins/microsoft_integrations/application/helpers/microsoft_callback.php', $callback_file);
	}
	if (!file_exists($callback_file)) {
		throw new Exception('can\'t syncronize microsoft callback file dnx');
	}
	
	if (!file_exists($webhook_callback_file)) {
		@copy(ROOT . '/plugins/microsoft_integrations/application/helpers/microsoft_webhook_callback.php', $webhook_callback_file);
	}
	
	return array(
    'clientId'                => config_option('ms_client_id'),
    'clientSecret'            => config_option('ms_client_secret_value'),
    'redirectUri'             => with_slash(ROOT_URL) . config_option('ms_redirect_uri_suffix'),
    'urlAuthorize'            => config_option('ms_url_authorize'),
    'urlAccessToken'          => config_option('ms_url_access_token'),
    'urlResourceOwnerDetails' => config_option('ms_url_resource_owner_details'),
    'scopes'                  => config_option('ms_scopes')
  );
}

function configure_microsoft_oauth_client() {
	$config = get_microsoft_config();
  $oauthClient = new \League\OAuth2\Client\Provider\GenericProvider($config);
	
	return $oauthClient;
}

function get_microsoft_auth_url() {

	$oauthClient = configure_microsoft_oauth_client();
	
	$authUrl = $oauthClient->getAuthorizationUrl();

	// Save client state so we can validate in callback
	$_SESSION['microsoftOauthState'] = $oauthClient->getState();

	return $authUrl;
}

function is_valid_microsoft_access_token($user_id, &$accessToken) {
	/* @var $accessToken League\OAuth2\Client\Token\AccessToken */
	if (!$accessToken) return false;
	$is_valid = false;
	try {
		$is_valid = !$accessToken->hasExpired();
		
		if (!$is_valid) {
			// try to refresh the token
			$oauthClient = configure_microsoft_oauth_client();
			$newAccessToken = $oauthClient->getAccessToken('refresh_token', array(
					'refresh_token' => $accessToken->getRefreshToken()
			));
			
			//Logger::log_r("REFRESHING MS TOKEN: ".print_r($newAccessToken,true));
			
			if ($newAccessToken) {
				set_ms_access_token($user_id, $newAccessToken);
				$accessToken = $newAccessToken;
				$is_valid = !$accessToken->hasExpired();
			}
		}
		
	} catch (Exception $e) {
		Logger::log_r("MS ERROR at ".__FUNCTION__.": ".$e->getMessage()."\n".$e->getTraceAsString());
	}
	
	return $is_valid;
}

/**
 * gets the email address of the ms account linked to the user if the user has connected to ms
 */
function get_contact_ms_email($contact) {
	
	$email = $contact->getEmailAddress();
	
	$access_token = get_ms_access_token($contact->getId());
	if (is_valid_microsoft_access_token($contact->getId(), $access_token)) {
		try {
			$graph = new Microsoft\Graph\Graph();
			$graph->setAccessToken($access_token->getToken());
			
			$user_response = $graph->createRequest('GET', '/me?$select=displayName,mail,mailboxSettings,userPrincipalName')->execute();
			//Logger::log_r($user_response);
			
			$user_data = $user_response->getBody();
			if ($user_data) {
				$email = $user_data['userPrincipalName'];
			}
		} catch (Exception $e) {
			Logger::log("MS ERROR: could not get ms email for user ".$contact->getId()."\n".$e->getMessage());
		}
	}
	
	return $email;
}

/**
 * After saving an event in feng, this function calls ms graph api to create it also in the ms calendar associated to the user
 * @param ProjectEvent $object The event to sync
 * @param int $user_id the id of the user that added/edited the event
 */
function save_feng_event_in_ms_calendar(ProjectEvent $object, $user_id, $sync_all_attendees=true) {

	$accessToken = get_ms_access_token($user_id);
	if (!is_valid_microsoft_access_token($user_id, $accessToken)) {
		
		$login_event_data = array(
				'need_microsoft_login' => true,
				'controller' => 'microsoft',
				'action' => 'save_event_in_ms_calendar',
				'id' => $object->getId(),
				'auth_url' => get_microsoft_auth_url(),
		);
		
		if ($user_id == logged_user()->getId()) {
			evt_add("login_to_microsoft", $login_event_data);
		}
		return;
	}
	
	$now = DateTimeValueLib::now();
	$now_str = $now->toMySQL();
	$now_str = str_replace(' ', 'T', $now_str);

	$more_users_to_sync_this_event = array();
	
	$invitations = EventInvitations::instance()->findAll(array('conditions' => 'event_id='.$object->getId()));
	$attendees = array();
	foreach ($invitations as $inv) {
		/* @var $inv EventInvitation */
		$contact = Contacts::findById($inv->getContactId());
		if (!$contact) continue;
		
		if ($user_id == $contact->getId()) {
			$contact_ms_email = $contact->getEmailAddress();
		} else {
			$contact_ms_email = get_contact_ms_email($contact);
		}
		/*
		if ($inv->getInvitationState() == 1) $inv_response = 'Accepted';
		else if ($inv->getInvitationState() == 2) $inv_response = 'Declined';
		else $inv_response = 'None';
		*/
		$attendees[] = array(
				'emailAddress' => array(
						'address' => $contact_ms_email
				),
				'type' => 'optional',
			/*	This part is commented because it generates an error
				'status' => array(
						'response' => $inv_response,
						'time' => $now_str,
				),*/
		);
		if ($contact->getId() != $user_id) {
			$more_users_to_sync_this_event[] = $contact->getId();
		}
	}
	
	// get timezone info
	$tz_offset = $object->getTimezoneValue() / 3600;
	$tz_name = Timezones::getTimezoneName($object->getTimezoneId());

	// clone date objects to process
	$start_dt = new DateTimeValue($object->getStart()->getTimestamp());
	$end_dt = new DateTimeValue($object->getDuration()->getTimestamp());


	// apply timezone to dates
	$start_dt->add('h', $tz_offset);
	$end_dt->add('h', $tz_offset);
	
	// format dates
	$start = $start_dt->toMySQL();
	$start = str_replace(' ', 'T', $start);
	$end = $end_dt->toMySQL();
	$end = str_replace(' ', 'T', $end);

	$event_data = [
			'subject' => $object->getObjectName(),
			'attendees' => $attendees,
			'start' => [
					'dateTime' => $start,
					'timeZone' => $tz_name,
			],
			'end' => [
					'dateTime' => $end,
					'timeZone' => $tz_name,
			],
			'body' => [
					'content' => $object->getDescription(),
					'contentType' => 'TEXT'//TEXT or HTML
			],
			'isAllDay' => $object->getTypeId() == 2,
	];
	if ($event_data['isAllDay']) {
		$event_data['start']['dateTime'] = $start_dt->beginningOfDay()->toMySQL();
		$event_data['end']['dateTime'] = $start_dt->beginningOfDay()->add('d',1)->toMySQL();
	}

	/**
	 * Recurrence data for the events
	 */
	Env::useHelper('microsoft_utilities', 'microsoft_integrations');

	# Check if there is recurrence to add $event_data['recurrence'] 
	if($object->isRepetitive()) {
		# daily or weekly
		if(!empty($object->getRepeatD())) {
			
			$event_data['recurrence'] = [
				'pattern' => [
					'type' => 'daily',
					'interval' => $object->getRepeatD()
				],
				'range' => MicrosoftUtilities::buildRange($object, $start_dt)
			];
		}
		# monthly
		else if(!empty($object->getRepeatM())) {
			$event_data['recurrence'] = [
				'pattern' => [
					'type' => 'absoluteMonthly',
					'interval' => $object->getRepeatM(),
					'dayOfMonth' => $start_dt->getDay(),
				],
				'range' => MicrosoftUtilities::buildRange($object, $start_dt)
			];
		}
		# yearly
		else if(!empty($object->getRepeatY())) {
			$event_data['recurrence'] = [
				'pattern' => [
					'type' => 'absoluteYearly',
					'interval' => $object->getRepeatY(),
					'dayOfMonth' => $start_dt->getDay(),
					"month" => $start_dt->getMonth(),
				],
				'range' => MicrosoftUtilities::buildRange($object, $start_dt)
			]; 
		}
	}
	
	/**
	 * Check if there is any reminder
	 */
 	$minutesToTheReminder = MicrosoftUtilities::getMinutesOfReminder($object);
	if($minutesToTheReminder) {
		$event_data['isReminderOn'] = true;
		$event_data['reminderMinutesBeforeStart'] = $minutesToTheReminder;
	}
	
	$ms_event_id = MsCalendarObjects::getMsIdByUserAndObject($user_id, $object->getId());
	$is_new = !isset($ms_event_id) || $ms_event_id == '';

	try {
		
		$graph = new Microsoft\Graph\Graph();
		$graph->setAccessToken($accessToken->getToken());
		
		if ($is_new) {
			
			// create the request to add the event
			$response = $graph->createRequest('POST', '/me/events')
			->attachBody($event_data)
			->execute();
		} else {
			// create the request to edit the event
			$response = $graph->createRequest('PATCH', '/me/events/'.$ms_event_id)
			->attachBody($event_data)
			->execute();
		}
		
		$response_body = $response->getBody();
		$ms_event_id = $response_body['id'];
		//Logger::log_r("EVENT SYNCRONIZED: $ms_event_id");
		
		if ($is_new) {
			// save the external event id for this user and event
			$cal_obj_data = MsCalendarObjects::findOne(array("conditions" => array("object_id=? and user_id=?", $object->getId(), $user_id)));
			if (!$cal_obj_data) {
				$cal_obj_data = new MsCalendarObject();
				$cal_obj_data->setObjectId($object->getId());
				$cal_obj_data->setUserId($user_id);
			}
			$cal_obj_data->setMsObjectId($ms_event_id);
			$cal_obj_data->save();
		}
		
		// save the event in the attendees' calendar, if they are connected
/*		if ($sync_all_attendees) {
			foreach ($more_users_to_sync_this_event as $uid) {
				save_feng_event_in_ms_calendar($object, $uid, false);
			}
 		}
 		*/
		
	} catch (Exception $e) {
		
		Logger::log_r($e->getMessage());
		Logger::log_r($e->getTraceAsString());
		
		throw $e;// only for debug
	}
}

/**
 * After deleting an event in feng/evx this function deletes it from the ms calendar associated to the event
 * @param int $user_id the id of the user that added/edited the event
 * @param string $ms_event_id the id of the event in the ms calendar
 */
function delete_feng_event_in_ms_calendar($user_id, $ms_event_id, $sync_all_attendees=true) {
	
	if (!isset($ms_event_id) || $ms_event_id == '') {
		return;
	}
	
	$accessToken = get_ms_access_token($user_id);
	if (!is_valid_microsoft_access_token($user_id, $accessToken)) {
		
		$login_event_data = array(
				'need_microsoft_login' => true,
				'controller' => 'microsoft',
				'action' => 'delete_event_in_ms_calendar',
				'id' => $ms_event_id,
				'auth_url' => get_microsoft_auth_url(),
		);
		
		if ($user_id == logged_user()->getId()) {
			evt_add("login_to_microsoft", $login_event_data);
		}
		return;
	}
	
	try {

		$graph = new Microsoft\Graph\Graph();
		$graph->setAccessToken($accessToken->getToken());
		
		// create the request to delete the event
		$graph->createRequest('DELETE', '/me/events/'.$ms_event_id)
		->attachBody($event_data)
		->execute();
		
	/*	if ($sync_all_attendees) {
			
			$event_id = MsCalendarObjects::getIdByUserAndMsObjectId($user_id, $ms_event_id);
			
			$more_users_to_sync_this_event = array();
			$invitations = EventInvitations::instance()->findAll(array('conditions' => 'event_id='.$event_id));
			
			foreach ($invitations as $inv) {
				$contact = Contacts::findById($inv->getContactId());
				if (!$contact) continue;
				if ($contact->getId() != $user_id) {
					$more_users_to_sync_this_event[] = $contact->getId();
				}
			}
			foreach ($more_users_to_sync_this_event as $uid) {
				$uid_ms_event_id = MsCalendarObjects::getMsIdByUserAndObject($uid, $event_id);
				delete_feng_event_in_ms_calendar($uid, $uid_ms_event_id, false);
			}
		}
		*/
		
		MsCalendarObjects::delete("user_id='$user_id' AND ms_object_id='$ms_event_id'");
		
	} catch (Exception $e) {
		
		Logger::log_r($e->getMessage());
		Logger::log_r($e->getTraceAsString());
		
	}
	
	
}


