<?php

/**
 * When receiving an event change notification from ms, this function is called to update the event in feng
 * @param Microsoft\Graph\Model\Event $event_result the data received from the ms graph api
 * @param ProjectEvent $feng_event the event object to update
 * @param int $logged_user_id the user id associated to the action
 * @return ProjectEvent $feng_event the updated event object
 */
function apply_ms_event_data_to_feng_event(Microsoft\Graph\Model\Event $event_result, ProjectEvent $feng_event, $logged_user_id) {
	
	if ($event_result instanceof Microsoft\Graph\Model\Event) {

		$logged_user = Contacts::findById($logged_user_id);
		CompanyWebsite::instance()->logUserIn($logged_user);
		
		$feng_event->setObjectName($event_result->getSubject());
		
		$description = html_to_text($event_result->getBody()->getContent());
		$feng_event->setDescription($description);
		
		$start_time = strtotime($event_result->getStart()->getDateTime());
		if ($event_result->getStart()->getTimeZone() != "UTC") {
			// apply sent tz
		}
		$start_dt = new DateTimeValue($start_time);
		$feng_event->setStart($start_dt);
		
		$end_time = strtotime($event_result->getEnd()->getDateTime());
		if ($event_result->getEnd()->getTimeZone() != "UTC") {
			// apply sent tz
		}
		$end_dt = new DateTimeValue($end_time);
		$feng_event->setDuration($end_dt);
		
		if ($event_result->getIsAllDay()) {
			$feng_event->setTypeId(2);
			$start_dt->setHour(9);
			$start_dt->setMinute(0);
			$start_dt->setSecond(0);
			$end_dt = new DateTimeValue($start_dt->getTimestamp() + 3600);
			$feng_event->setStart($start_dt);
			$feng_event->setDuration($end_dt);
		} else {
			$feng_event->setTypeId(1);
		}
		
		
		$feng_event->save();
		
		$all_attendees = $event_result->getAttendees();
		// add the organizer to the attendee list
		$all_attendees[] = array(
				'emailAddress' => array(
						'address'=>$event_result->getOrganizer()->getEmailAddress()->getAddress(),
				)
		);
		
		//Logger::log_r($event_result);
		//Logger::log_r($all_attendees);
		
		// foreach attendee create an event_invitation at feng/evx
		foreach ($all_attendees as $attendee) {
			$inv_email = $attendee['emailAddress']['address'];
			$user = Contacts::getByEmail($inv_email);
			if (!$user) {
				if (!isset($all_users)) {
					$all_users = Contacts::getAllUsers();
				}
				foreach ($all_users as $u) {
					$ms_email_address = get_contact_ms_email($u);
					if ($ms_email_address == $inv_email) {
						$user = $u;
						break;
					}
				}
			}
			if ($user instanceof Contact) {
				$inv = EventInvitations::instance()->findByEventAndContact($feng_event->getId(), $user->getId());
				if (!$inv instanceof EventInvitation) {
					$inv = new EventInvitation();
					$inv->setEventId($feng_event->getId());
					$inv->setContactId($user->getId());
				}
				//$inv_response = $attendee->getStatus()->getResponse()->value();
				//$inv_response = $attendee['status']['response'];
				$inv_response = 'none';
				switch (strtolower($inv_response)) {
					case 'organizer':
					case 'tentativelyaccepted':
					case 'accepted':
						$invitation_state = 1;
						break;
					case 'declined':
						$invitation_state = 2;
						break;
					default:
						$invitation_state = 0;
				}
				$inv->setInvitationState($invitation_state);
				$inv->save();
			}
		}
	}
	
	return $feng_event;
}

/**
 * This function is called when receving a change notification from ms graph api
 * @param string $ms_event_id the event id at ms calendar of the added/modified event
 * @param string $action the action performed at the external calendar
 */
function process_incoming_data_change_event($ms_event_id, $action) {
	$_SESSION['sync_incoming_objects'] = true; // to prevent sync loops

	// include general sync functions file
	Env::useHelper('microsoft_functions', 'microsoft_integrations');
	
	// process event update action
	if ($action == 'updated') {
		$ms_cal_obj = MsCalendarObjects::getByMsObjectId($ms_event_id);
		if (!$ms_cal_obj) {
			$action = 'created'; // if event not found then create one
		} else {
		
			$user_id = $ms_cal_obj->getUserId();
			$feng_event_id = $ms_cal_obj->getObjectId();
			
			$feng_event = ProjectEvents::findById($feng_event_id);
			if (!$feng_event) {
				$action = 'created'; // if event not found then create one
			} else {
				
				try {
					$access_token = get_ms_access_token($user_id);
					if ($access_token) {

						$graph = new Microsoft\Graph\Graph();
						$graph->setAccessToken($access_token->getToken());
						
						//query the event data
						$event_result = $graph->createRequest("GET", "/me/events/".$ms_event_id)
						->setReturnType(Microsoft\Graph\Model\Event::class)
						->execute();
						
					}
					
				} catch (Exception $e) {
					Logger::log_r("MS ERROR WHEN GETTING EVENT ($user_id - ".$feng_event->getId()."): ".$e->getMessage()."\n".$e->getTraceAsString());
				}
				
				try {
					
					if ($event_result instanceof Microsoft\Graph\Model\Event) {

						$ms_event_mod_time = $event_result->getLastModifiedDateTime();
						
						// verify that the event in ms has a newer version that the one in feng
						if ($ms_event_mod_time->getTimestamp() > $feng_event->getUpdatedOn()->getTimestamp()) {

							// update the event
							$feng_event = apply_ms_event_data_to_feng_event($event_result, $feng_event, $user_id);
						
						} else {
							// only for debug
						//	Logger::log_r("IGNORING NOTIFICATION, FENG EVENT IS NEWER");
						}
					}
					
				} catch (Exception $e) {
					Logger::log_r("MS ERROR WHEN UPDATING EVENT IN FENG: ".$e->getMessage()."\n".$e->getTraceAsString());
				}

			}
		}
	}
	
	// process event creation action
	if ($action == 'created') {

		// if the event was created by us it may happen that this notification arrive before we could save the ms_event_id 
		// so we can give some time to do that to our controller
		sleep(5);
		// -------------
		
		$all_users = Contacts::getAllUsers();
		$user_found = false;
		
		foreach ($all_users as $user) {
			$access_token = get_ms_access_token($user->getId());
			if ($access_token) {
				$graph = new Microsoft\Graph\Graph();
				$graph->setAccessToken($access_token->getToken());
				
				$event_result = null;
				
				try {
					//query the event data
					$event_result = $graph->createRequest("GET", "/me/events/".$ms_event_id)
					->setReturnType(Microsoft\Graph\Model\Event::class)
					->execute();
					
				} catch (Exception $e) {
					// if not found then continue trying with another user
					//Logger::log_r("MS ERROR WHEN GETTING EVENT: ".$e->getMessage()."\n".$e->getTraceAsString());
				}
				
				if ($event_result instanceof Microsoft\Graph\Model\Event) {

					$user_found = true; // this stops the iteration

					// check if ms event id already exists, to prevent duplicates
					$feng_event_id = MsCalendarObjects::getIdByUserAndMsObjectId($user->getId(), $ms_event_id);

					// only create the event if not found in previous query
					if (!$feng_event_id) {

						// if we created the event and the notification arrived before we save the ms_event_id, we need to verify by name, date and creator in order to prevent duplicates
						$start_time_str = strtotime($event_result->getStart()->getDateTime());
						$st = new DateTimeValue($start_time_str);
						$event_exists = ProjectEvents::instance()->findOne(array(
							"conditions" => array("name=? AND start=? AND created_by_id=?", $event_result->getSubject(), $st, $user->getId())
						));

						if (!$event_exists) {

							$feng_event = new ProjectEvent();
							$feng_event = apply_ms_event_data_to_feng_event($event_result, $feng_event, $user->getId());
							
							$feng_event->addToSharingTable();
							$feng_event->addToSearchableObjects(true);
							
							$cal_obj_data = MsCalendarObjects::findOne(array("conditions" => array("object_id=? and user_id=?", $feng_event->getId(), $user->getId())));
							if (!$cal_obj_data) {
								$cal_obj_data = new MsCalendarObject();
								$cal_obj_data->setObjectId($feng_event->getId());
								$cal_obj_data->setUserId($user->getId());
							}
							$cal_obj_data->setMsObjectId($ms_event_id);
							$cal_obj_data->save();

							Logger::log_r("*** EVENT CREATED FROM MS: ". $feng_event->getObjectName());

						} else {
							Logger::log_r("*** EVENT ALREADY EXISTS: ". $event_exists->getObjectName());
						}

					} else {
						// only for debug
						Logger::log_r("IGNORING EXISTING EVENT CREATION: $ms_event_id");
					}
					
				}
			}
			if ($user_found) break;
		}
		
	}

	// process event deletion action
	if ($action == 'deleted') {
		$ms_cal_obj = MsCalendarObjects::getByMsObjectId($ms_event_id);
		if (!$ms_cal_obj) {
			$action = 'deleted'; // if event not found then delete one
		} else {
		
			$user_id = $ms_cal_obj->getUserId();
			$feng_event_id = $ms_cal_obj->getObjectId();
			
			$feng_event = ProjectEvents::findById($feng_event_id);
			
			try {
				
				if ($feng_event) {
					$feng_event->trash();
				}
					
			} catch (Exception $e) {
				Logger::log_r("MS ERROR WHEN DELETING EVENT ($user_id - ".$feng_event->getId()."): ".$e->getMessage()."\n".$e->getTraceAsString());
			}
			
		}
	}

	unset($_SESSION['sync_incoming_objects']); // remove the flag that prevents loops
	
}

