<?php

/**
 * This is the main function triggered by the hooks after adding, editing or untrashing a timeslot
 * It splits the entered timeslot im the overtime hours dimension members and then triggers the recalculation process for the rest of the week.
 * @param Timeslot $object
 */
function split_timeslots_by_overtime_hours_and_recalculate_the_rest_of_the_week($object) {
	// This is a flag for avoiding that this hooked is triggered by the calculation processes.
    // It is set here, and unset after the process is finished.
	$now = DateTimeValueLib::now();
	if (isset($_SESSION['after_add_to_members_processing']) && ($now->getTimestamp() - 10 > $_SESSION['after_add_to_members_processing'])) {
		// unset this flag if it is older than 10 seconds
		unset($_SESSION['after_add_to_members_processing']);
	}
	
	if (!array_var($_SESSION, 'after_add_to_members_processing') && $object instanceof Timeslot) {
		
		// Setting the flag...
		$now = DateTimeValueLib::now();
		$_SESSION['after_add_to_members_processing'] = $now->getTimestamp();
		
		// Ensure that the timeslot is the original one
		$object = find_original_timeslot($object);
		
		// Delete the related timeslots before processing this one
		delete_related_timeslots($object);
		
		// Step 1) Split the original timeslot, in case it already has more hours than the total regular hours it's allowed 
		// calculating which hours are normal, extra, etc.
		$generated_timeslots_data = split_timeslots_by_overtime_hours($object, json_decode(array_var($_REQUEST, 'members')));
		
		// Step 2) Save the timeslot(s), and recalculate the entire week
		save_timeslots_and_recalculate_entire_week($object, $generated_timeslots_data);
		
		unset($_SESSION['after_add_to_members_processing']);
	}
}


/**
 * Splits the timeslot by the hour types defined in the configuration.
 * Checks all the user's timeslots in current day and current week to split the timeslot between regular and overtime hours.
 * 
 * This is Step 1)
 * This is the first function of the process, which is triggered by hook in overtime_calculations, with the hook 'after_add_to_members' in the ObjectController
 * 
 * It is also called from within other functions in this file 
 * 
 * Notes:
 *   * Note that "normal hours" and "regular hours" are used interchangably
 * 
 * @return array The array of timeslots data after splitting the original timeslot using the overtime hours dimension
 * @param Timeslot $object
 */

function split_timeslots_by_overtime_hours(Timeslot $object, $timeslot_members = null) {

    //====================================
    // A. Initializations and Pre-checks
    //====================================
    
    // Pre-check #1) Does the timeslot have an end date? If not, we can't process it.
	if (!$object->getEndTime() instanceof DateTimeValue) return;
		
	$all_timeslots = array();
	
	$tz_offset = logged_user()->getUserTimezoneHoursOffset();
	
	//Initialize values from configuration variables
	$max_normal_work_minutes = 60 * config_option('normal_work_hours');
	$max_overtime_work_minutes = 60 * config_option('overtime_work_hours');
	$max_normal_work_minutes_per_week = 60 * config_option('normal_work_hours_per_week');
	$normal_work_member_id = config_option('normal_work_member_id');
	$overtime_work_member_id = config_option('overtime_work_member_id');
	$double_overtime_work_member_id = config_option('double_overtime_work_member_id');
    //This is an array with all the members that correspond to worked hours (the ones that participate in the calculations	
	$worked_hours_member_ids = array($normal_work_member_id, $overtime_work_member_id, $double_overtime_work_member_id);

	// Initialize member_ids (Make sure all the members in which the timeslot is classified are loaded).
	if (is_null($timeslot_members)) {
		$original_member_ids = $object->getMemberIds();
	} else {
		$original_member_ids = $timeslot_members;
	}

	// Pre-check #2) Is this "worked" time? 
	// Ignore this timeslot for calculations if it is classified in a member of non-worked hours (holiday, sick, etc)
	$overtime_dimension = Dimensions::findByCode('overtime_hours');
	foreach ($original_member_ids as $original_member_id) {
		$m = Members::getMemberById($original_member_id);
		if ($m->getDimensionId() == $overtime_dimension->getId() && !in_array($m->getId(), array($normal_work_member_id, $overtime_work_member_id, $double_overtime_work_member_id))) {
			return;
		}
	}
	
	// Initial date (in GMT-0)
	$timeslot_start_date = new DateTimeValue($object->getStartTime()->getTimestamp());
	// Now in user TZ
	//$timeslot_start_date->add('h', -1*$tz_offset);

	//====================================
	// B. Calculations
	//====================================
	
	// Calculate the total amount of minutes of the current timeslot
	$current_timeslot_minutes = floor(($object->getEndTime()->getTimestamp() - $object->getStartTime()->getTimestamp()) / 60) - $object->getSubtract()/60;
	
	// Calculate the remaining normal minutes available for that day, checking other timeslots entered on this day
	$previous_worked_minutes = get_contact_worked_time($object, $timeslot_start_date);
	   
	// calculate the start aof the timeslot's week
	$week_limits = calculate_week_start_and_week_end($object);
	$week_start = $week_limits['start'];
	

	//====================================
	// C. Checks and actual splitting
	//====================================
	
	// CHECK #1) 7th CONSECUTIVE DAY RULE
	// CheckS if the user has worked 6 consecutive days before this timeslot.
	$has_worked_too_many_consecutive_days_for_overtime = check_consecutive_days($object, $timeslot_start_date);
	
	// If so, then all minutes are overtime. After 10 hours all are double overtime
	if ($has_worked_too_many_consecutive_days_for_overtime) {
		// no normal minutes left if user has worked all the consecutive days for overtime
		$week_remaining_normal_minutes = 0;
		
		// in this case double overtime is after 10 hours of overtime (12-2)
		$max_overtime_work_minutes -= 60 * 2; 
		
	} else {
	    // CHECK #2) 40 HOURS PER WEEK RULE
	    // Check how much regular-time work has been logged so far this week
		// Get the current week's total number of worked hours classified as regular (normal)
		$sql_classified_in_normal= "AND EXISTS (SELECT om.object_id FROM ".TABLE_PREFIX."object_members om WHERE om.member_id='$normal_work_member_id' AND om.object_id=t.object_id)";
		$current_week_normal_worked_minutes = get_contact_worked_time($object, $week_start, $sql_classified_in_normal, $timeslot_start_date);
		
		// If there are more than 40 hours, $week_remaining_normal_minutes is now 0.
		// (No more work will be under regular hours)
		// Update the caps to calculate normal and overtime hours if necesary (depending on the max hours per week)
		$week_remaining_normal_minutes = $max_normal_work_minutes_per_week - $current_week_normal_worked_minutes;
		if ($week_remaining_normal_minutes < 0) $week_remaining_normal_minutes = 0;
	}
	
	// This should go inside the 'else' above?
	// CHECK #3)
	// Check if the space for regular hours in the week (max of 40) is alreay below 10
	// In which case we'll have to fill-up with more overtime today;
	//   * up until 12 hours
	//     * after which everything will be double-overtime
	if ($max_normal_work_minutes > $week_remaining_normal_minutes) {
		// We do this by increasing the max_overtime_hours for today (from 2)
		$max_overtime_work_minutes += ($max_normal_work_minutes - $week_remaining_normal_minutes);
		// Update the normal work minutes variable to make the calculations using the remaining minutes of this week
		$max_normal_work_minutes = $week_remaining_normal_minutes;
	}
	
	// CHECK #4) IF ALL GOOD, SIMPLY ADD REGULAR HOURS
	// Check if the sum of (timeslot + previous work) still does not exceed the max amount of normal minutes available
	// If so, save timeslot as is.
	if ($current_timeslot_minutes + $previous_worked_minutes <= $max_normal_work_minutes) {
		// prevent duplicated classification
		ObjectMembers::removeObjectFromMembers($object, logged_user(), $worked_hours_member_ids, $worked_hours_member_ids);
		
		// only classify original timeslot in normal worked hours member
		$normal_work_member = Members::getMemberById($normal_work_member_id);
		$object->addToMembers(array($normal_work_member));
		return array();
				
	} else {
	    
		// Generate the splitted timeslots classified in each of the members set in configuration
		
		// get the current week's normal worked time
		//    This is already done before, correct?
		$sql_classified_in_normal= "AND EXISTS (SELECT om.object_id FROM ".TABLE_PREFIX."object_members om WHERE om.member_id='$normal_work_member_id' AND om.object_id=t.object_id)";
		$current_day_normal_worked_minutes = get_contact_worked_time($object, $timeslot_start_date, $sql_classified_in_normal);
		
		// discount the already worked minutes to get the max_normal_minutes remaining
		//$max_normal_minutes = $max_normal_work_minutes - $previous_worked_minutes;
		$max_normal_minutes = $max_normal_work_minutes - $current_day_normal_worked_minutes;
		if ($max_normal_minutes < 0) $max_normal_minutes = 0;
		
		// calculate normal and overtime minutes using current_timeslot_minutes
		$normal_minutes = $current_timeslot_minutes;
		$extra_minutes = 0;
		if ($current_timeslot_minutes > $max_normal_minutes) {
			$normal_minutes = $max_normal_minutes;
			$extra_minutes = $current_timeslot_minutes - $max_normal_minutes;
		}
		
		$splitted_timeslots = array();
		$member_ids = $original_member_ids;
		
		// build timeslots data for normal hours
		if ($normal_minutes > 0) {
			// remove the previous worked hours member id
			$member_ids = array_diff($member_ids, $worked_hours_member_ids);
			// add the new worked hours member id
			$normal_member_ids = array_merge($member_ids, array($normal_work_member_id));
			$ts_normal_data = build_timeslot_data($object, $timeslot_start_date, $normal_minutes, $normal_member_ids, array($overtime_work_member_id, $double_overtime_work_member_id));
		
			$splitted_timeslots[] = $ts_normal_data;
		}
		
		// build timeslots data for extra hours
		if ($extra_minutes > 0) {
			
			$sql_classified_in_extra = "AND EXISTS (SELECT om.object_id FROM ".TABLE_PREFIX."object_members om WHERE om.member_id='$overtime_work_member_id' AND om.object_id=t.object_id)";
			
			$previous_extra_minutes = get_contact_worked_time($object, $timeslot_start_date, $sql_classified_in_extra);
			
			// check if there are double overtime minutes
			$double_extra_minutes = 0;
			if ($extra_minutes + $previous_extra_minutes > $max_overtime_work_minutes) {
				$remaining_overtime_minutes = $max_overtime_work_minutes - $previous_extra_minutes;
				if ($remaining_overtime_minutes < 0) $remaining_overtime_minutes = 0;
				
				if ($extra_minutes > $remaining_overtime_minutes) {
					$double_extra_minutes = $extra_minutes - $remaining_overtime_minutes;
					$extra_minutes = $remaining_overtime_minutes;
				}
			}
			
			if ($extra_minutes > 0) {
				// remove the previous worked hours member id
				$member_ids = array_diff($member_ids, $worked_hours_member_ids);
				// add the new worked hours member id
				$extra_member_ids = array_merge($member_ids, array($overtime_work_member_id));
				$timeslot_start_date_extra = new DateTimeValue($timeslot_start_date->getTimestamp());
				$timeslot_start_date_extra->add('m', $normal_minutes);
				$timeslot_start_date_extra->add('h', $tz_offset);
				$ts_extra_data = build_timeslot_data($object, $timeslot_start_date_extra, $extra_minutes, $extra_member_ids, array($normal_work_member_id, $double_overtime_work_member_id));
			
				$splitted_timeslots[] = $ts_extra_data;
			}
			
			// double overtime data
			if ($double_extra_minutes > 0) {
				// remove the previous worked hours member id
				$member_ids = array_diff($member_ids, $worked_hours_member_ids);
				// add the new worked hours member id
				$double_extra_member_ids = array_merge($member_ids, array($double_overtime_work_member_id));
				$timeslot_start_date_extra2 = new DateTimeValue($timeslot_start_date->getTimestamp());
				$timeslot_start_date_extra2->add('m', $normal_minutes + $extra_minutes);
				$timeslot_start_date_extra2->add('h', $tz_offset);
				$ts_double_extra_data = build_timeslot_data($object, $timeslot_start_date_extra2, $double_extra_minutes, $double_extra_member_ids, array($normal_work_member_id, $overtime_work_member_id));
				
				$splitted_timeslots[] = $ts_double_extra_data;
			}
		}
		
		$all_timeslots = $splitted_timeslots;
	}
	
	return $all_timeslots;
}

/**
 * 	This function checks if the user has worked 6 consecutive days before this timeslot.
 *  If so, then all minutes are overtime
 *  
 */
function check_consecutive_days($timeslot, $timeslot_start_date) {

    $has_worked_past_max_consecutive_days_for_overtime = false;
    
    $max_consecutive_days_for_overtime = 6;

    $normal_work_member_id = config_option('normal_work_member_id');
    $overtime_work_member_id = config_option('overtime_work_member_id');
    $double_overtime_work_member_id = config_option('double_overtime_work_member_id');

    $tz_offset = logged_user()->getUserTimezoneHoursOffset();
    
    // start date to filter the query
    $current_day_start = new DateTimeValue($timeslot_start_date->getTimestamp());
    $current_day_start->add('h', $tz_offset);
    $current_day_start->beginningOfDay();
    $current_day_start->add('h', -1*$tz_offset);
    
    $worked_time_sql = "AND EXISTS (SELECT om2.object_id FROM ".TABLE_PREFIX."object_members om2
		WHERE om2.member_id IN ('$normal_work_member_id','$overtime_work_member_id','$double_overtime_work_member_id') AND om2.object_id=t.object_id)";
    
    $consecutive_days_sql = "
    		SELECT IF (t.original_timeslot_id = 0, DATE(start_time), (SELECT ts.`start_time` FROM ".TABLE_PREFIX."timeslots ts WHERE ts.object_id=t.original_timeslot_id)) AS day
			FROM ".TABLE_PREFIX."timeslots t
			INNER JOIN `".TABLE_PREFIX."objects` o ON o.id=t.object_id
			WHERE `contact_id` = ". $timeslot->getContactId() ." AND o.trashed_by_id = 0
			$worked_time_sql
			AND IF (t.original_timeslot_id = 0,
				start_time < '".$current_day_start->toMySQL()."'
				AND start_time > ADDDATE('".$current_day_start->toMySQL()."', INTERVAL -".$max_consecutive_days_for_overtime." DAY)
				,
				(SELECT ts.`start_time` FROM ".TABLE_PREFIX."timeslots ts WHERE ts.object_id=t.original_timeslot_id) < '".$current_day_start->toMySQL()."'
				AND (SELECT ts.`start_time` FROM ".TABLE_PREFIX."timeslots ts WHERE ts.object_id=t.original_timeslot_id) > ADDDATE('".$current_day_start->toMySQL()."', INTERVAL -".$max_consecutive_days_for_overtime." DAY)
			)
			GROUP BY DATE(day);
	";
			
	$consecutive_days_rows = DB::executeAll($consecutive_days_sql);
	
	// the amount of rows is the amount of worked days, check if it reached the $max_consecutive_days_for_overtime
	if (count($consecutive_days_rows) >= $max_consecutive_days_for_overtime) {
	    $has_worked_past_max_consecutive_days_for_overtime = true;
    }
    
    return $has_worked_past_max_consecutive_days_for_overtime;
}

/**
 * Builds an associative array with the timeslot information to pass to the controller and create it.
 * @param Contact $contact
 * @param DateTimeValue $date
 * @param integer $total_minutes
 * @param array $member_ids
 * @param array $remove_from_member_ids
 */
function build_timeslot_data($original_timeslot, $date, $total_minutes, $member_ids, $remove_from_member_ids = array()) {
	
	$ts_date = new DateTimeValue($date->getTimestamp());
	$minutes = $total_minutes % 60;
	$hours = floor($total_minutes / 60);
	
	$start_time = $ts_date->format('H:i');
	
	$data = array (
			'hours' => $hours,
			'minutes' => $minutes,
			'date' => $ts_date,
			'start_time' => $start_time,
			'description' => $original_timeslot->getDescription(),
			'contact_id' => $original_timeslot->getContactId(),
			'additional_member_ids' => array_unique(array_filter($member_ids)),
			'original_timeslot_id' => $original_timeslot->getId(),
	);
	if (count($remove_from_member_ids) > 0) {
		$data['ignore_member_ids'] = $remove_from_member_ids;
	}
	
	return $data;
}


/**
 * Returns the amount of time registered as worked (in minutes)
 *   - in the workday of the timeslot
 *   - by the same person
 *   
 * @param Timeslot $timeslot The original timeslot to exclude from the query
 * @param DateTimeValue $date The date to filter the query
 * @return integer The sum of worked minutes
 */
function get_contact_worked_time(Timeslot $timeslot, DateTimeValue $date, $additional_conditions='', $end_day=null) {
	
	$tz_offset = logged_user()->getUserTimezoneHoursOffset();
	
	// start date to filter the query
	$d = new DateTimeValue($date->getTimestamp());
	$d->add('h', $tz_offset);
	$d->beginningOfDay();
	$d->add('h', -1*$tz_offset);
	
	// end date to filter the query
	if (is_null($end_day)) {
		$end_day_time = new DateTimeValue($date->getTimestamp());
		$end_day_time->add('h', $tz_offset);
		$end_day_time->endOfDay();
	} else {
		$end_day_time = new DateTimeValue($end_day->getTimestamp());
	}
	$end_day_time->add('h', -1*$tz_offset);

	$contact_id = $timeslot->getContactId();
	$ts_id_to_exclude = $timeslot->getId();
	
	$normal_work_member_id = config_option('normal_work_member_id');
	$overtime_work_member_id = config_option('overtime_work_member_id');
	$double_overtime_work_member_id = config_option('double_overtime_work_member_id');
	
	$worked_time_sql = "AND EXISTS (SELECT om2.object_id FROM ".TABLE_PREFIX."object_members om2 
		WHERE om2.member_id IN ('$normal_work_member_id','$overtime_work_member_id','$double_overtime_work_member_id') AND om2.object_id=t.object_id)";
	
	// build sql
	$sql = "SELECT SUM(GREATEST(TIMESTAMPDIFF(MINUTE,start_time,end_time),0)) - SUM(subtract/60) as total
			FROM `".TABLE_PREFIX."timeslots` t
			INNER JOIN `".TABLE_PREFIX."objects` o ON o.id=t.object_id
			WHERE `contact_id` = ". $contact_id ." AND `object_id` != '$ts_id_to_exclude'
					$worked_time_sql
					AND o.trashed_by_id = 0
					AND IF (t.original_timeslot_id = 0,
						`start_time` >= ".DB::escape($d->toMySQL())."
						AND `start_time` <= ".DB::escape($end_day_time->toMySQL())."
						,
						(SELECT ts.`start_time` FROM ".TABLE_PREFIX."timeslots ts WHERE ts.object_id=t.original_timeslot_id) >= ".DB::escape($d->toMySQL())."
						AND (SELECT ts.`start_time` FROM ".TABLE_PREFIX."timeslots ts WHERE ts.object_id=t.original_timeslot_id) <= ".DB::escape($end_day_time->toMySQL())."
					)
					$additional_conditions";
	
	$row = DB::executeOne($sql);
	
	
	
	return array_var($row, "total", 0);
}

/**
 * Foreach next day in the week, after the date of $timeslot,
 * the timeslots will be grouped by day and the classification in overtime hours dimension will be recalculated
 * @param Timeslot $timeslot The original added/edited timeslot
 */
function recalculate_next_days_timeslots_overtime(Timeslot $timeslot) {
	
	$timeslot_start_date = $timeslot->getStartTime();
	
	$normal_work_member_id = config_option('normal_work_member_id');
	$overtime_work_member_id = config_option('overtime_work_member_id');
	$double_overtime_work_member_id = config_option('double_overtime_work_member_id');
	
	$worked_hours_member_ids = array($normal_work_member_id, $overtime_work_member_id, $double_overtime_work_member_id);
	
	// This variable is an array of all the timeslots that fall in the same week as the one we are processing
	$accumulated_timeslots = get_week_timeslots_grouped_by_date_and_classification($timeslot);
	
	// For each accumulated timeslot, call the split process and save the generated timesltos
	$time_controller = new TimeController();
	foreach ($accumulated_timeslots as $date_key => $objects_by_classification) {
		
		foreach ($objects_by_classification as $classification_key => $timeslots) {
		
			foreach ($timeslots as $object) {
			
				// This is an array of all the timeslots, after they have been split
				
				$generated_timeslots_data = split_timeslots_by_overtime_hours($object, $object->getMemberIds());
				
				if (count($generated_timeslots_data) > 0) {
						
					$original_ts_modified_data = array_shift($generated_timeslots_data);
						
					// modify original timeslot
					$end_time = new DateTimeValue($object->getStartTime()->getTimestamp());
					$end_time->add('m', $original_ts_modified_data['minutes'] + 60 * $original_ts_modified_data['hours']);
					$object->setEndTime($end_time);
					$object->save();
					
					$additional_member_ids = $original_ts_modified_data['additional_member_ids'];
					//$additional_member_ids = array_merge($additional_member_ids, $object->getMemberIds());
					
					// remove any worked hours member from the original member ids
					$original_member_ids = array_diff($object->getMemberIds(), $worked_hours_member_ids);
					
					// join the new worked hour members with the original ones
					$additional_member_ids = array_merge($additional_member_ids, $original_member_ids);
	
					$additional_member_ids = array_unique($additional_member_ids);
					$additional_members = Members::findAll(array("conditions" => "id IN (".implode(',',$additional_member_ids).")"));
					
					// prevent duplicated classification
					ObjectMembers::removeObjectFromMembers($object, logged_user(), $worked_hours_member_ids, $worked_hours_member_ids);
					// classify
					$object->addToMembers($additional_members);
						
					// save the other timeslots
					foreach ($generated_timeslots_data as $timeslot_data) {
						$parameters = array();
						$parameters['use_current_time'] = false;
						$parameters['object_id'] = $object->getRelObjectId();
						$parameters['timeslot'] = $timeslot_data;
				
						$new_timeslot_object = $time_controller->add_timeslot($parameters, false);
						
						// trigger event to add the new timeslot to the list
						//trigger_event_to_add_timeslot_to_list($new_timeslot_object);
					}
				}
			}
		}
	}
	
}

/**
 * Buidls an array for each day, containing the array of timeslots grouped by date and classification
 * Only original timeslots are included, the related timeslots are accumulated in the original one and then deleted.
 * @param Timeslot $timeslot
 * @return array
 */
function get_week_timeslots_grouped_by_date_and_classification(Timeslot $timeslot) {
	
	$timeslot_start_date = $timeslot->getStartTime();
	$tz_offset = logged_user()->getUserTimezoneHoursOffset();

	$week_limits = calculate_week_start_and_week_end($timeslot);
	$week_end = $week_limits['end'];
	// recalculate 4 weeks forward
	$week_end->add('w', 4);
	
	$date_from = new DateTimeValue($timeslot->getEndTime()->getTimestamp());
	$date_from->add('h', $tz_offset);
	$date_from->beginningOfDay();
	$date_from->add('h', -1 * $tz_offset);
	
	$normal_work_member_id = config_option('normal_work_member_id');
	$overtime_work_member_id = config_option('overtime_work_member_id');
	$double_overtime_work_member_id = config_option('double_overtime_work_member_id');
	$normal_work_member = Members::findById($normal_work_member_id);
	
	$worked_hours_member_ids = array($normal_work_member_id, $overtime_work_member_id, $double_overtime_work_member_id);
	
	$worked_time_sql = "AND EXISTS (SELECT om2.object_id FROM ".TABLE_PREFIX."object_members om2
	WHERE om2.member_id IN ('$normal_work_member_id','$overtime_work_member_id','$double_overtime_work_member_id') AND om2.object_id=o.id)";
	
	// get the next days timeslots, only the original ones
	$next_days_timeslots = Timeslots::findAll(array(
			"conditions" => "`contact_id` = ". $timeslot->getContactId() ."
				$worked_time_sql
				AND o.trashed_by_id = 0
				AND o.id <> ".$timeslot->getId()."
				AND `original_timeslot_id` = 0
				AND `start_time` > ".DB::escape($date_from->toMySQL())."
				AND `start_time` <= ".DB::escape($week_end->toMySQL()),
			"order" => "start_time"
	));
	
	// group them by date
	$grouped_timeslots = array();
	foreach ($next_days_timeslots as $ts) {
		$ts_start_date = new DateTimeValue($ts->getStartTime()->getTimestamp());
		$ts_start_date->add('h', $tz_offset);
		$date_key = $ts_start_date->format('Y-m-d');
		
		
		// initialize the array if not initialized yet
		if (!isset($grouped_timeslots[$date_key])) $grouped_timeslots[$date_key] = array();
		
		// generate the classification key
		$classification_key = get_timeslot_classification_key($ts);
		
		// initialize the array if not initialized yet
		if (!isset($grouped_timeslots[$date_key][$classification_key])) $grouped_timeslots[$date_key][$classification_key] = array();
		
		// get the accumulation of related timeslots worked minutes
		$related_ts_worked_minutes = get_related_timeslots_worked_minutes($ts);
		
		// modify end date and worked time columns in the original timeslot
		add_minutes_to_timeslot_object_and_save($ts, $related_ts_worked_minutes);
		
		// delete the related timeslots
		delete_related_timeslots($ts);
		
		// put the timeslot in the array
		$grouped_timeslots[$date_key][$classification_key][] = $ts;
	}
	// ensure that array is correctly sorted
	ksort($grouped_timeslots);
	
	return $grouped_timeslots;
/*
	// accumulate in one per day and classification, and delete the rest
	$accumulated_timeslots = array();
	foreach ($grouped_timeslots as $date_key => $ts_by_classification) {
		foreach ($ts_by_classification as $classification_key => $ts_array) {
			// accumulate in the first timeslot, keep it in $main_ts var and remove it from the array
			$main_ts = array_shift($ts_array);
		
			// iterate the other timeslots and accumulate the minutes, mark them to be deleted
			$to_delete = array();
			$minutes_to_add = 0;
			foreach ($ts_array as $ts) {
				$minutes_to_add += $ts->getColumnValue('worked_time');
				$to_delete[] = $ts;
			}
		
			// set the new end time in main timeslot
			if ($minutes_to_add > 0) {
				$new_end_time = new DateTimeValue($main_ts->getEndTime()->getTimestamp() + $minutes_to_add*60);
				$main_ts->setEndTime($new_end_time);
				$main_ts->save();
			}
		
			// remove timeslots from worked time members, and reset to normal member
			ObjectMembers::removeObjectFromMembers($main_ts, logged_user(), $worked_hours_member_ids, $worked_hours_member_ids);
			$main_ts->addToMembers(array($normal_work_member));
		
			// delete the other timeslots
			foreach ($to_delete as $ts_to_delete) {
				$ts_to_delete->delete();
			}
		
			// collect the accumulated timeslot of the day and classification
			if (!isset($accumulated_timeslots[$date_key][$classification_key])) $accumulated_timeslots[$date_key][$classification_key] = array();
			$accumulated_timeslots[$date_key][$classification_key] = $main_ts;
		}
	}
	// ensure that they are ordered according to the start date
	ksort($accumulated_timeslots);
	
	return $accumulated_timeslots;
*/
}

/**
 * Generates a key for the timeslot, using its members ids and type
 * Only certain member types are used for the key, not all members are used.
 * @param Timeslot $timeslot
 * @return string
 */
function get_timeslot_classification_key(Timeslot $timeslot) {
	
	$object_type_ids_for_key = ObjectTypes::findAll(array(
			"id" => true,
			"conditions" => "name IN ('workspace','project','hour_type','folder','project_folder')",
	));
	
	//$members = $timeslot->getMembers();
	$member_ids_rows = DB::executeAll("
			SELECT om.member_id FROM ".TABLE_PREFIX."object_members om
			INNER JOIN ".TABLE_PREFIX."members m ON m.id=om.member_id
			WHERE om.object_id=".$timeslot->getId()." 
			AND m.object_type_id IN (".implode(',', $object_type_ids_for_key).")
	");
	
	$member_ids = array_flat($member_ids_rows);
	$member_ids = array_filter($member_ids);
	
	$members = array();
	if (count($member_ids) > 0) {
		$members = Members::findAll(array("conditions" => "id IN (".implode(',', $member_ids).")"));
	}
	
	$key_array = array();
	foreach ($members as $member) {
		/* @var $member Member */
		if (in_array($member->getObjectTypeId(), $object_type_ids_for_key)) {
			$key_array[$member->getObjectTypeId()] = $member->getId();
		}
	}
	ksort($key_array);
	
	$ot_keys = array();
	foreach ($key_array as $otid => $mid) {
		$ot_keys[] = $otid."-".$mid;
	}
	$key_string = implode(',', $ot_keys);
	
	return $key_string;
}


function calculate_week_start_and_week_end(Timeslot $timeslot) {

	// start time in gmt
    $timeslot_start_date = new DateTimeValue($timeslot->getStartTime()->getTimestamp());
    $tz_offset = logged_user()->getUserTimezoneHoursOffset();
    
    // apply timezone to get the correct day
    $timeslot_start_date->add('h', $tz_offset);
    $day = $timeslot_start_date->format('w');
    
    
    $week_start = new DateTimeValue(strtotime('-'.$day.' days', $timeslot_start_date->getTimestamp()));
    $week_start->add('h', $tz_offset);
    $week_start->beginningOfDay();
    $week_start->add('h', -1 * $tz_offset);
    
    $week_end = new DateTimeValue(strtotime('+'.(6-$day).' days', $timeslot_start_date->getTimestamp()));
    $week_end->add('h', $tz_offset);
    $week_end->endOfDay();
    $week_end->add('h', -1 * $tz_offset);
    
    $result = array('start' => $week_start, 'end' => $week_end);
    
    return $result;
}


/**
 * This function saves all the timeslots created for the recently added one.
 * 
 * It also checks whether other timeslots for that person in that week might be affected
 * 
 * @param $original_timeslot Timeslot
 * @param $all_timeslots_to_be_added array
 */
function save_timeslots_and_recalculate_entire_week($original_timeslot, $all_timeslots_to_be_added) {

    // 1. Initialization of variables and reference values
    // These are the differnt types hours:
    // regular
    // overtime
    // double overtime
    $normal_work_member_id = config_option('normal_work_member_id');
    $overtime_work_member_id = config_option('overtime_work_member_id');
    $double_overtime_work_member_id = config_option('double_overtime_work_member_id');
    
    // This is the controller for adding and modifying timeslots
    $time_controller = new TimeController();
    
    if (count($all_timeslots_to_be_added) > 0) {
        
        //The first one in the array is the orignal timeslot, but potentially modified.
        $original_ts_modified_data = array_shift($all_timeslots_to_be_added);
        
        // modify original timeslot
        $end_time = new DateTimeValue($original_timeslot->getStartTime()->getTimestamp());
        $end_time->add('m', $original_ts_modified_data['minutes'] + 60 * $original_ts_modified_data['hours']);
        // take into consideration the paused time
        if ($original_timeslot->getSubtract() > 0) {
        	$end_time->add('s', $original_timeslot->getSubtract());
        }
        $original_timeslot->setEndTime($end_time);
        $original_timeslot->save();
        
        // verify that the timeslot is classified in only one member of overtime_hours dimension
        $additional_member_ids = $original_ts_modified_data['additional_member_ids'];
        if (in_array($double_overtime_work_member_id, $additional_member_ids)) {
            $tmp = array();
            foreach ($additional_member_ids as $mid) {
                if (!in_array($mid, array($overtime_work_member_id, $normal_work_member_id))) $tmp[] = $mid;
            }
            $additional_member_ids = $tmp;
        } else if (in_array($overtime_work_member_id, $additional_member_ids)) {
            $tmp = array();
            foreach ($additional_member_ids as $mid) {
                if (!in_array($mid, array($double_overtime_work_member_id, $normal_work_member_id))) $tmp[] = $mid;
            }
            $additional_member_ids = $tmp;
        } else if (in_array($normal_work_member_id, $additional_member_ids)) {
            $tmp = array();
            foreach ($additional_member_ids as $mid) {
                if (!in_array($mid, array($double_overtime_work_member_id, $overtime_work_member_id))) $tmp[] = $mid;
            }
            $additional_member_ids = $tmp;
        }
        
        $additional_members = Members::findAll(array("conditions" => "id IN (".implode(',',$additional_member_ids).")"));
        // prevent duplicated classification
        $work_member_ids = array($normal_work_member_id, $overtime_work_member_id, $double_overtime_work_member_id);
        ObjectMembers::removeObjectFromMembers($original_timeslot, logged_user(), $work_member_ids, $work_member_ids);
        // classify
        $original_timeslot->addToMembers($additional_members);
        
        // save the other timeslots
        foreach ($all_timeslots_to_be_added as $timeslot_data) {
            $parameters = array();
            $parameters['use_current_time'] = false;
            $parameters['object_id'] = $original_timeslot->getRelObjectId();
            $parameters['timeslot'] = $timeslot_data;
            
            $new_timeslot_object = $time_controller->add_timeslot($parameters, false);
            
            // trigger event to add the new timeslot to the list
            trigger_event_to_add_timeslot_to_list($new_timeslot_object);
        }
    }
    
    recalculate_next_days_timeslots_overtime($original_timeslot);
}

/**
 * Send the new timeslot data to the interface to insert it in the list
 * @param Timeslot $new_timeslot_object
 */
function trigger_event_to_add_timeslot_to_list($new_timeslot_object) {
	if ($new_timeslot_object instanceof Timeslot) {
		// build data to send to the list
		$grid_id = array_var($_REQUEST, 'current') . "_timeslots_module_grid";
		$event_data = array(
				"timeslot" => $new_timeslot_object->getArrayInfo(false, true, true), 
				"real_obj_id" => $new_timeslot_object->getRelObjectId(), 
				"grid_id" => $grid_id,
		);
		
		// trigger event to insert the data in the list
		evt_add('overtime_timeslot_generated', $event_data);
	}
}


/**
 * Returns the original timeslot of the chain.
 * If $timeslot is one of the automatically created, the return value will be the original timeslot that triggered the calculation
 * If $timeslot is the original one, it will be returned
 * 
 * @param Timeslot $timeslot Any timeslot
 */
function find_original_timeslot(Timeslot $timeslot) {
	
	$original_timeslot_id = $timeslot->getColumnValue('original_timeslot_id');
	
	if ($original_timeslot_id == 0) {
		return $timeslot;
	} else {
		
		$original_timeslot = Timeslots::findById($original_timeslot_id);
		return $original_timeslot;
		
	}
}

/**
 * Returns all the timeslots that have been automatically calculated from $timeslot
 * 
 * @param Timeslot $timeslot The original timeslot
 */
function find_related_timeslots(Timeslot $timeslot) {
	
	$related_timeslots = Timeslots::findAll(array(
		"conditions" => "original_timeslot_id=".$timeslot->getId()
	));
	
	if (!is_array($related_timeslots)) {
		$related_timeslots = array();
	}
	
	return $related_timeslots;
}

/**
 * Returns the total amount of worked minutes of all the timeslots that have been automatically calculated from $timeslot
 * 
 * @param Timeslot $timeslot The original timeslot
 */
function get_related_timeslots_worked_minutes($original_timeslot) {
	// get the related timeslots
	$related_timeslots = find_related_timeslots($original_timeslot);
	
	// accumulate the worked minutes of the related timeslots
	$related_ts_worked_minutes = 0;
	foreach ($related_timeslots as $rel_ts) {
		$worked_minutes = $rel_ts->getColumnValue('worked_time');
		$related_ts_worked_minutes += $worked_minutes;
	}
	
	return $related_ts_worked_minutes;
}

/**
 * Deletes all the timeslots that have been automatically calculated from $timeslot
 * 
 * @param Timeslot $timeslot The original timeslot
 */
function delete_related_timeslots(Timeslot $timeslot) {
	
	$related_timeslots = find_related_timeslots($timeslot);
	
	foreach ($related_timeslots as $t) {
		$t->delete();
	}
}


/**
 * Updates the timeslot's end_time and worked_time columns in the database after adding $minutes to the current timeslot
 * Returns the updated Timeslot object
 * 
 * @param Timeslot $timeslot
 * @param int $minutes
 */
function add_minutes_to_timeslot_object_and_save(Timeslot &$timeslot, $minutes) {
	// get the previous value of worked time column
	$original_ts_worked_minutes = $timeslot->getColumnValue('worked_time');

	// build the new end date
	$new_end_date = new DateTimeValue($timeslot->getEndTime()->getTimestamp());
	$new_end_date->add('m', $minutes);

	// sum the new minutes to the previous minutes
	$total_minutes = $original_ts_worked_minutes + $minutes;

	// update the timeslot end_time and worked_time columns
	DB::execute("UPDATE ".TABLE_PREFIX."timeslots
			SET worked_time='$total_minutes', end_time='".$new_end_date->toMySQL()."'
				WHERE object_id=".$timeslot->getId());

	// reload the object with the new data previusly saved
	$timeslot = Timeslots::findById($timeslot->getId(), true);
}






