<?php
Hook::register('overtime_calculations');

/**
 * Adds a new column to the Timestlots model to keep the original timeslot reference when a timeslot is generated automatically
 * @param string $type The class name of the model
 * @param array $ret The array of columns
 */
function overtime_calculations_object_definition($type, &$ret) {

	if ($type == 'Timeslot') {
		$ret['original_timeslot_id'] = DATA_TYPE_INTEGER;
	}
}

/**
 * Sets the new column for the original timeslot reference as a system column, to exclude it from reports and listings.
 * @param DataManager $manager manager class for the model
 * @param array $columns array of the system columns
 */
function overtime_calculations_additional_system_columns($manager, &$columns) {
	if ($manager instanceof Timeslots) {
		$columns[] = 'original_timeslot_id';
	}
}


function overtime_calculations_modify_timeslot_before_edit($params, &$timeslot) {
	// This is the helper file 'functions.php', where most of the calculations take place.
	Env::useHelper('functions', 'overtime_calculations');
	
	// find the original timeslot, and accumulate the hours of the related ones in it
	$original_timeslot = find_original_timeslot($timeslot);
	
	// check that object exists, if not then do nothing
	if (!$original_timeslot instanceof Timeslot) return;
	
	// accumulate the worked minutes of the related timeslots
	$related_ts_worked_minutes = get_related_timeslots_worked_minutes($original_timeslot);
	
	// get the original timeslot worked time in minutes
	$original_ts_worked_minutes = $original_timeslot->getColumnValue('worked_time');
	
	// modify end date and worked time columns in the original timeslot object
	$original_timeslot->setColumnValue('worked_time', $original_ts_worked_minutes + $related_ts_worked_minutes);
	$original_timeslot->getEndTime()->add('m', $related_ts_worked_minutes);
	
	// replace the output variable with the accumulated original timeslot
	$timeslot = $original_timeslot;

}

/**
 * This is where the calculation of overtime hours begins. 
 * 
 * It takes the timeslot, and checks whether or not there are changes to be made for both this timeslot and other timeslots.
 * If there are changes needed, those changes are made (in the functions it calls, in functions.php
 * 
 */
function overtime_calculations_after_add_to_members($object, &$members) {
	
	if ($object instanceof Timeslot) {
		// This is the helper file 'functions.php', where most of the calculations take place.
		Env::useHelper('functions', 'overtime_calculations');
		
		// calls the split function, save the generated timeslots, and then recalculate the rest of the week
		split_timeslots_by_overtime_hours_and_recalculate_the_rest_of_the_week($object);
	}
}

/**
 * After restoring a timeslot from the trash, call the function to split as it 
 * @param array $params Array containing the object that has been trashed
 * @param null $ignored
 */
function overtime_calculations_after_content_object_untrash($params, &$ignored) {
	$object = array_var($params, 'object');
	
	if ($object instanceof Timeslot) {
		// This is the helper file 'functions.php', where most of the calculations take place.
		Env::useHelper('functions', 'overtime_calculations');
		
		// calls the split function, save the generated timeslots, and then recalculate the rest of the week
		split_timeslots_by_overtime_hours_and_recalculate_the_rest_of_the_week($object);
	
	}
}

/**
 * After sending a timeslot to the trash, merge the related timeslots and delete them.
 * Only send to trash the original timeslot with the accumulated worked time.
 * After that the recalculation process for the rest of the week must be triggered.
 * @param array $params
 * @param null $ignored
 */
function overtime_calculations_after_content_object_trash($params, &$ignored) {
	$object = array_var($params, 'object');
	
	if ($object instanceof Timeslot) {
		$now = DateTimeValueLib::now();
		if (isset($_SESSION['after_add_to_members_processing']) && ($now->getTimestamp() - 10 > $_SESSION['after_add_to_members_processing'])) {
			// unset this flag if it is older than 10 seconds
			unset($_SESSION['after_add_to_members_processing']);
		}
		
		if (!array_var($_SESSION, 'after_add_to_members_processing')) {
			// Setting the flag...
			$now = DateTimeValueLib::now();
			$_SESSION['after_add_to_members_processing'] = $now->getTimestamp();
			
			Env::useHelper('functions', 'overtime_calculations');
			
			// find the original timeslot
			$original_timeslot = find_original_timeslot($object);
			
			// get the accumulation of related timeslots worked minutes
			$related_ts_worked_minutes = get_related_timeslots_worked_minutes($original_timeslot);
			
			// delete the related timeslots
			delete_related_timeslots($original_timeslot);
	
			// modify end date and worked time columns in the original timeslot
			add_minutes_to_timeslot_object_and_save($original_timeslot, $related_ts_worked_minutes);
			
			// send to trash the original timeslot, set the parameter to fire hook to false to prevent recursion.
			$original_timeslot->trash(null, false);
			
			// recalculate the rest of the week
			recalculate_next_days_timeslots_overtime($original_timeslot);
			
			unset($_SESSION['after_add_to_members_processing']);
		}
	}
}
