<?php

require_once 'plugins/quickbooks/application/helpers/TSheetsRestClient.php';


use TSheets\TSheetsRestClient;
use TSheets\ObjectType;

/***************************************************************************************/
/**************************** BASIC AND AUTH FUNCTIONS *********************************/
/***************************************************************************************/

function get_qbo_access_token() {
	
	$access_token = null;
	
	$serialized_token = config_option('qbo_access_token');
	if ($serialized_token != '') {
		$access_token = unserialize(base64_decode($serialized_token));
	}
	
	if ($access_token && !is_valid_qbo_access_token($access_token)) {
		$data_service = get_qbo_data_service($access_token); // this will refresh the token
		if ($data_service) {
			$access_token = $data_service->getOAuth2LoginHelper()->getAccessToken();
		} else {
			$access_token = null;
		}
	}
	
	return $access_token;
}

function save_qbo_access_token($access_token) {
	
	$serialized_token = base64_encode(serialize($access_token));
	
	set_config_option('qbo_access_token', $serialized_token);
}



function get_qbo_config() {
	
	//$callback_file = config_option('qbo_oauth_redirect_uri');
	
	if (!file_exists(with_slash(ROOT) . 'qbo_callback.php')) {
		@copy(ROOT . "/plugins/quickbooks/application/helpers/qbo_callback.php", ROOT . "/qbo_callback.php");
	}
	if (!file_exists(with_slash(ROOT) . 'qbo_callback.php')) {
		throw new Exception('cant syncronize qbo callback file dnx');
	}
	
	//$config = include('qbo_config.php');
	$config = array(
			'authorizationRequestUrl' => config_option('qbo_authorization_request_url'),
			'tokenEndPointUrl' => config_option('qbo_token_endpoint_url'),
			'client_id' => config_option('qbo_client_id'),
			'client_secret' => config_option('qbo_client_secret'),
			'oauth_scope' => config_option('qbo_oauth_scope'),
			'oauth_redirect_uri' => config_option('qbo_oauth_redirect_uri'),//'https://quickbooks.evxsoftware.com/callback.php',//with_slash(ROOT_URL) . $callback_file,
			'baseUrl' => config_option('qbo_base_url'),
	);
	
	return $config;
}

/**
 * 
 * @param array $config
 * @return \QuickBooksOnline\API\DataService\DataService
 */
function configure_qbo_dataservice($config, $access_token=null) {
	$settings = array(
			'auth_mode' => 'oauth2',
			'ClientID' => $config['client_id'],
			'ClientSecret' =>  $config['client_secret'],
			'RedirectURI' => $config['oauth_redirect_uri'],
			'scope' => $config['oauth_scope'],
			'baseUrl' => $config['baseUrl'],
	);
	
	if ($access_token instanceof QuickBooksOnline\API\Core\OAuth\OAuth2\OAuth2AccessToken) {
		$settings['accessTokenKey'] = $access_token->getAccessToken();
		$settings['refreshTokenKey'] = $access_token->getRefreshToken();
		$settings['QBORealmID'] = $access_token->getRealmID();
	}
	
	$dataService = QuickBooksOnline\API\DataService\DataService::Configure($settings);
	
	if ($access_token) {
		$dataService->updateOAuth2Token($access_token);
	}
	
	return $dataService;
}

/**
 * 
 * @param QuickBooksOnline\API\Core\OAuth\OAuth2\OAuth2AccessToken $access_token
 * @return \QuickBooksOnline\API\DataService\DataService
 */
function get_qbo_data_service($access_token = null) {
	/*
	 * Retrieve the accessToken value from session variable
	 */
	if (!$access_token) {
		$access_token = get_qbo_access_token();
	}
	
	$config = get_qbo_config();
	$dataService = configure_qbo_dataservice($config, $access_token);
	
	/*
	 * Update the OAuth2Token of the dataService object
	 */
	if ($access_token) {
		// if access token is not valid, then use the refres token (if still valid) to generate a new access token
		if (!is_valid_qbo_access_token($access_token) && is_valid_qbo_refresh_token($access_token)) {
			
			try {
				
				// use the login helper and the refresh token to get the new access token
				$oauth2LoginHelper = $dataService->getOAuth2LoginHelper();
				$new_access_token_obj = $oauth2LoginHelper->refreshAccessTokenWithRefreshToken($access_token->getRefreshToken());
				
				// save the new token in database
				save_qbo_access_token($new_access_token_obj);
				
				// update the access token variable to return
				$access_token = $new_access_token_obj;
				
				// update dataservice access token
				$dataService->updateOAuth2Token($access_token);
				
			} catch (Exception $e) {
				// if token could not be generated then unset the $access_token var to the user will be asked the credentials again
				$access_token = null;
				Logger::log_r("ERROR REFRESHING TOKEN: ". $e->getMessage()."\n".$e->getTraceAsString());
				return;
			}
		}
	}
	
	// set logs location, try to create it if not exists
	$qbo_log_dir = CACHE_DIR."/qbo";
	if (!file_exists($qbo_log_dir)) @mkdir($qbo_log_dir);
	if (file_exists($qbo_log_dir)) {
		$dataService->setLogLocation($qbo_log_dir);
		$dataService->enableLog();
	}
	
	return $dataService;
}


function get_qbo_auth_url() {
	
	$config = get_qbo_config();
	$dataService = configure_qbo_dataservice($config);
	
	$OAuth2LoginHelper = $dataService->getOAuth2LoginHelper();
	$authUrl = $OAuth2LoginHelper->getAuthorizationCodeURL();
	
	return $authUrl;
}


function is_valid_qbo_access_token($accessToken) {

	try {

		if (!$accessToken) return false;
		
		$token_expiration_date = $accessToken->getAccessTokenExpiresAt();
		$expiration_timestamp = strtotime($token_expiration_date);
		$now_timestamp = DateTimeValueLib::now()->getTimestamp();
		
		return $expiration_timestamp > $now_timestamp;

	} catch (Exception $e) {

		Logger::log_r("QBO SYNC - Error validating Access Token: " . $e->getMessage());

		// if there is an error with the access token then return false so the user can login again
		return false;
	}
	
}

function is_valid_qbo_refresh_token($accessToken) {
	
	$token_expiration_date = $accessToken->getRefreshTokenExpiresAt();
	$expiration_timestamp = strtotime($token_expiration_date);
	$now_timestamp = DateTimeValueLib::now()->getTimestamp();
	
	return $expiration_timestamp > $now_timestamp;
}


function build_qbo_error_message($error) {
	$message = "An error occurred syncronizing with Quickbooks:\n";
	$message .= "Error: ".$error->getIntuitErrorType()." (".$error->getIntuitErrorCode().")\n";
	$message .= $error->getIntuitErrorMessage()."\n".$error->getIntuitErrorDetail();
	
	return $message;
}



/**************************************************************************************/
/**************************** CONFIGURATION FUNCTIONS *********************************/
/**************************************************************************************/

function check_if_object_can_be_sync(ContentDataObject $object) {
	$can_sync = false;
	
	$from_sync_date = config_option('qbo_sync_entities_from_date');

	if ($from_sync_date instanceof DateTimeValue) {

		if ($object instanceof IncomeInvoice) {

			$can_sync = $object->getDate()->getTimestamp() >= $from_sync_date->getTimestamp();
 
		} else if ($object instanceof PaymentReceipt) {

			$can_sync = $object->getDate()->getTimestamp() >= $from_sync_date->getTimestamp();
 
		} else if ($object instanceof Timeslot) {

			$can_sync = $object->getStartTime()->getTimestamp() >= $from_sync_date->getTimestamp();
 
		} else if ($object instanceof InvoicePaymentReceipt) {
 
			$date_cp = CustomProperties::getCustomPropertyByCode($object->getObjectTypeId(), 'date');
			$date_cp_val = CustomPropertyValues::getCustomPropertyValue($object->getId(), $date_cp->getId());

			if ($date_cp_val) {
				try {
					$p_date = DateTimeValueLib::dateFromFormatAndString(DATE_MYSQL, $date_cp_val->getValue());
				} catch (Exception $e) {};
			}

			if ($p_date instanceof DateTimeValue) {
				$can_sync = $p_date->getTimestamp() >= $from_sync_date->getTimestamp();
			}

		} 
	}

	return $can_sync;
}

/**
 * @return QuickBooksOnline\API\Data\IPPPreferences
 */
function get_qbo_preferences($dataService = null) {
	if (!$dataService) $dataService = get_qbo_data_service();
	
	$preferences = $dataService->getCompanyPreferences();
	
	return $preferences;
}

function get_qbo_custom_fields_definition($preferences = null) {
	
	if (!$preferences) $preferences = get_qbo_preferences();
	//Logger::log_r($preferences);
	
	$result = array();
	if (!is_array($preferences->SalesFormsPrefs->CustomField)) return $result;
	
	$custom_fields = $preferences->SalesFormsPrefs->CustomField[1]->CustomField;
	
	if (!is_array($custom_fields)) $custom_fields = array($custom_fields);
	
	foreach ($custom_fields as $custom_field) {
		$cf = array(
				'id' => substr($custom_field->Name, -1),
				'name' => $custom_field->StringValue,
				'type' => $custom_field->Type,
				'code' => $custom_field->Name,
		);
		$result[] = $cf;
	}
	
	return $result;
}

function get_qbo_custom_field_value_from_invoice_and_prop(IncomeInvoice $invoice, $prop) {
	$val = null;
	
	if (str_starts_with($prop, "cp_")) {
		$cp_id = str_replace("cp_", "", $prop);
		$cp_val = CustomPropertyValues::getCustomPropertyValue($invoice->getId(), $cp_id);
		if ($cp_val instanceof CustomPropertyValue) {
			$val = trim($cp_val->getValue());
			$cp = CustomProperties::getCustomProperty($cp_id);
			if ($val != '' && $cp && ($cp->getType()=='date' || $cp->getType()=='datetime')) {
				$dtval = DateTimeValueLib::dateFromFormatAndString('Y-m-d', $val);
				if ($dtval instanceof DateTimeValue) $val = format_date($dtval, null, 0);
			}
		}
		
	} else if (str_starts_with($prop, "project_")) {
		
		$project_prop = str_replace("project_", "", $prop);
		if (str_starts_with($project_prop, "cp_")) {
			$cp_id = str_replace("cp_", "", $project_prop);
			
			$project_member = null;
			foreach ($invoice->getMembers() as $m) {
				if ($m->getObjectTypeId() == Projects::instance()->getObjectTypeId()) {
					$project_member = $m;
					break;
				}
			}
			
			if ($project_member instanceof Member) {
				$cp_val = CustomPropertyValues::getCustomPropertyValue($project_member->getObjectId(), $cp_id);
				if ($cp_val instanceof CustomPropertyValue) {
					$val = trim($cp_val->getValue());
					$cp = CustomProperties::getCustomProperty($cp_id);
					if ($val != '' && $cp && ($cp->getType()=='date' || $cp->getType()=='datetime')) {
						$dtval = DateTimeValueLib::dateFromFormatAndString('Y-m-d', $val);
						if ($dtval instanceof DateTimeValue) $val = format_date($dtval, null, 0);
					}
				}
			}
		}
		
	}
	return $val;
}

function get_qbo_custom_field_value_from_invoice(IncomeInvoice $invoice, $prop1, $prop2) {
	$values_array = array();
	
	$val1 = get_qbo_custom_field_value_from_invoice_and_prop($invoice, $prop1);
	if ($val1) $values_array[] = $val1;
	
	$val2 = get_qbo_custom_field_value_from_invoice_and_prop($invoice, $prop2);
	if ($val2) $values_array[] = $val2;
	
	return implode(' - ', array_filter($values_array));
}

function build_qbo_custom_fields_object(IncomeInvoice $invoice, $preferences) {
	
	$result = array();
	
	$custom_fields = get_qbo_custom_fields_definition($preferences);
	foreach ($custom_fields as $cf) {
		$row = DB::executeOne("SELECT * FROM ".TABLE_PREFIX."qbo_custom_field_match WHERE quickbooks_id=".$cf['id']);
		if ($row) {
			$invoice_cf_value = get_qbo_custom_field_value_from_invoice($invoice, $row['feng_property1'], $row['feng_property2']);
			if (strlen($invoice_cf_value) > 30) {
				$invoice_cf_value = substr($invoice_cf_value, 0, 30);
			}
			if ($invoice_cf_value) {
				$cf_obj = array(
						"DefinitionId" => $cf['id'],
						"StringValue" => $invoice_cf_value,
						"Name" => $cf['name'],
						"Type" => $cf['type'],
				);
				$result[] = $cf_obj;
			}
		}
	}
	
	return $result;
}


function get_qbo_items_definition($dataService = null, $type = null, $name = null) {
	if (!$dataService) $dataService = get_qbo_data_service();
	
	$where_clause = '';
	if ($type) {
		$type =  str_replace("'", "", $type);
		$where_clause = "where Type='$type'";
	} else if ($name) {
		$name =  str_replace("'", "", $name);
		$where_clause = "where Name='$name'";
		//$where_clause .= ($where_clause = "" ? "where " : " and "). "Name='$name'";
	}
	
	$items = $dataService->Query("select * from Item $where_clause maxresults 500");
	//Logger::log_r("QBO ITEMS");
	//Logger::log_r($items);
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting items list from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($items) && sizeof($items) > 0) {
			return $items;
		}
	}
	return array();
}

/**
 * Gets the qbo item definition using its id
 * @param \QuickBooksOnline\API\DataService\DataService $dataService
 * @param integer $id
 * @return \QuickBooksOnline\API\Data\IPPItem
 * @throws Exception
 */
function get_qbo_item_definition($dataService = null, $id) {
	
	// first check if we have built the cache
	if (isset($_SESSION['qbo_entities_cache'])) {
		$entity = array_var($_SESSION['qbo_entities_cache']['items'], $id);
		if ($entity) return $entity;
	}

	if (!$dataService) $dataService = get_qbo_data_service();
	$items = $dataService->Query("select * from Item where Id='$id'");
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting item $id from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	if (count($items) > 0) {
		return $items[0];
	}
}

function qbo_create_item($dataService = null, $item_name, $category_id = null, $qbo_account_id = null) {
	if (!$dataService) $dataService = get_qbo_data_service();

	// remove any space next to the ':' special character
	$item_name_arr = explode(':',$item_name);
	foreach ($item_name_arr as &$name_part) {
		$name_part = str_replace("'", "", trim($name_part));
	}
	$item_name = implode(':', $item_name_arr);
	
	// search item by full hierarchy name
	$items = $dataService->Query("select * from Item where FullyQualifiedName='$item_name'");
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting item $item_name from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	if (count($items) > 0) {
		return $items[0];
	} else {

		$default_income_account_id = null;

		// try to get the income account from the labor/expense category matching
		if ($category_id) {
			$default_income_account_id = get_qbo_account_id_by_category_id($category_id);
		}
		
		// if no income account specified by the category then look for the default one
		if (!$default_income_account_id) {
			$qbo_preferences = get_qbo_preferences($dataService);
	
			foreach ($qbo_preferences->OtherPrefs->NameValue as $other_pref) {
				if ($other_pref->Name == "UncategorizedIncomeAccountId") {
					$default_income_account_id = $other_pref->Value;
				}
			}
		}

		$qbo_account = null;

		// get the account using the id passed by parameter
		if ($qbo_account_id) {
			$qbo_account = get_qbo_account_by_id($dataService, $qbo_account_id);
		}

		// get the account object by its id if we found one in the previous queries
		if (!$qbo_account && $default_income_account_id) {
			$qbo_account = get_qbo_account_by_id($dataService, $default_income_account_id);
		}

		// if not found then search the "Sales" account
		if (!$qbo_account) {
			$qbo_account = get_qbo_account_by_name($dataService, "Sales");
		}
		
		if (!$qbo_account) return;

		// create the item and return the object
		$item_request_obj = QuickBooksOnline\API\Facades\Item::create([
				"Name" => escape_character($item_name),
				"Type" => "Service",
				"IncomeAccountRef" => [
						"name" => $qbo_account->Name,
						"value" => $qbo_account->Id,
				],
		]);
		$item_response_obj = $dataService->Add($item_request_obj);
		$error = $dataService->getLastError();
		if ($error) {
			Logger::log_r("Error adding item to Quickbooks: ".escape_character($item_name));
			Logger::log_r($error);
			throw new Exception(build_qbo_error_message($error));
		} else {
			return $item_response_obj;
		}
		
	}
}


/**
 * @param int $category_id the Labor/Expense category id
 * @return int The Quickbooks account id that matches with the catgory id
 */
function get_qbo_account_id_by_category_id($category_id) {

	$account_id = null;

	$row = DB::executeOne("SELECT qbo_account_id FROM ".TABLE_PREFIX."qbo_accounts_category_match WHERE category_id='$category_id'");
	if ($row) {
		$account_id = $row['qbo_account_id'];
	}

	// if there is no account defined for this category then use the parent category
	if (!$account_id) {
		$category_member = Members::getMemberById($category_id);
		if ($category_member instanceof Member && $category_member->getParentMemberId() > 0) {
			$account_id = get_qbo_account_id_by_category_id($category_member->getParentMemberId());
		}
	}

	return $account_id;
}


/**
 * @param int $category_id the Labor/Expense category id
 * @return int The Quickbooks item id that matches with the catgory id
 */
function get_qbo_item_id_by_category_id($category_id, $get_expense_item = false) {

	$item_id = null;

	$row = DB::executeOne("SELECT * FROM ".TABLE_PREFIX."qbo_items_category_match WHERE category_id='$category_id'");
	if ($row) {
		if ($get_expense_item) {
			$item_id = $row['qbo_expense_item_id'];
		} else {
			$item_id = $row['qbo_item_id'];
		}
	}

	// if there is no item defined for this category then use the parent category
	if (!$item_id) {
		$category_member = Members::getMemberById($category_id);
		if ($category_member instanceof Member && $category_member->getParentMemberId() > 0) {
			$item_id = get_qbo_item_id_by_category_id($category_member->getParentMemberId(), $get_expense_item);
		}
	}

	return $item_id;
}

/**
 * Returns the associated QBO Expense Item associated to a feng/evx object, using the expenses categories dimension
 * @param ContentDataObject $object
 * @return \QuickBooksOnline\API\Data\IPPItem
 */
function get_object_qbo_expense_item($object) {
	$expense_categories_dim = Dimensions::findByCode('expense_categories');
	if (!$expense_categories_dim) return;
	$exp_cat_member = null;

	// get the expense category member of the object
	$members = $object->getMembers();
	foreach ($members as $m) {
		if ($m->getDimensionId() == $expense_categories_dim->getId()) {
			$exp_cat_member = $m;
			break;
		}
	}

	// get the qbo item related to the expense category
	$qbo_item = null;
	if ($exp_cat_member instanceof Member) {
		$item_id = get_qbo_item_id_by_category_id($exp_cat_member->getId(), true);
		if (trim($item_id) != '') {
			$qbo_item = get_qbo_item_definition(null, $item_id);
		}
	}

	return $qbo_item;
}

/**
 * Returns the associated QBO Expense Account associated to a feng/evx object, using the expenses categories dimension
 * @param ContentDataObject $object
 * @return \QuickBooksOnline\API\Data\IPPAccount
 */
function get_object_qbo_expense_account($object) {
	
	// get the qbo item related to the object using the expense category
	$qbo_item = get_object_qbo_expense_item($object);

	// return the expense account related to the item
	if ($qbo_item) {
		return $qbo_item->ExpenseAccountRef;
	}
}


/**
 * Returns the associated QBO Labor Item associated to a feng/evx object, using the expenses categories dimension
 * @param ContentDataObject $object
 * @return \QuickBooksOnline\API\Data\IPPItem
 */
function get_object_qbo_labor_item($object) {
	$categories_dim = Dimensions::findByCode('hour_types');
	if (!$categories_dim) return;
	$cat_member = null;

	// get the labor category member of the object
	$members = $object->getMembers();
	foreach ($members as $m) {
		if ($m->getDimensionId() == $categories_dim->getId()) {
			$cat_member = $m;
			break;
		}
	}

	// get the qbo item related to the labor category
	$qbo_item = null;
	if ($cat_member instanceof Member) {
		$item_id = get_qbo_item_id_by_category_id($cat_member->getId());
		if (trim($item_id) != '') {
			$qbo_item = get_qbo_item_definition(null, $item_id);
		}
	}

	return $qbo_item;
}

/**
 * Returns the associated QBO Expense Account associated to a feng/evx object, using the expenses categories dimension
 * @param ContentDataObject $object
 * @return \QuickBooksOnline\API\Data\IPPAccount
 */
function get_object_qbo_labor_account($object) {
	
	// get the qbo item related to the object using the expense category
	$qbo_item = get_object_qbo_labor_item($object);

	// return the expense account related to the item
	if ($qbo_item) {
		return $qbo_item->IncomeAccountRef;
	}
}


/**
 * Gets the class definition using its id 
 * @param \QuickBooksOnline\API\DataService\DataService $dataService
 * @param integer $id
 * @throws Exception
 * @return \QuickBooksOnline\API\Data\IPPClass
 */
function get_qbo_class_def_by_id($dataService = null, $id) {
	
	// first check if we have built the cache
	if (isset($_SESSION['qbo_entities_cache'])) {
		$entity = array_var($_SESSION['qbo_entities_cache']['classes'], $id);
		if ($entity) return $entity;
	}
	
	$classes = $dataService->Query("select * from Class where Id='$id'");
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting class id:$id from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	if (count($classes) > 0) {
		// return class definition
		return $classes[0];
	}
}

/**
 * Gets the class definition that matches a given member by its name, if it doesn't exist then it is created in qbo
 * @param \QuickBooksOnline\API\DataService\DataService $dataService
 * @param Member $member
 * @param bool $create_if_not_exists
 * @throws Exception
 * @return \QuickBooksOnline\API\Data\IPPClass
 */
function get_qbo_class_def_by_member($dataService = null, Member $member, $create_if_not_exists = true) {

	$fully_qualified_name = null;

	// try to find the class name using the custom property qbo_class
	$member_ot = ObjectTypes::findById($member->getObjectTypeId());
	if ($member_ot->getType() == 'dimension_object') {
		$class_name_cp = CustomProperties::getCustomPropertyByCode($member_ot->getId(), 'qbo_class');
		if ($class_name_cp) {
			$class_name_cp_val = CustomPropertyValues::getCustomPropertyValue($member->getObjectId(), $class_name_cp->getId());
			if ($class_name_cp_val) $fully_qualified_name = trim($class_name_cp_val->getValue());
		}
	} else {
		$class_name_cp = MemberCustomProperties::getCustomPropertyByCode($member_ot->getId(), 'qbo_class');
		if ($class_name_cp) {
			$class_name_cp_val = MemberCustomPropertyValues::getMemberCustomPropertyValue($member->getId(), $class_name_cp->getId());
			if ($class_name_cp_val) $fully_qualified_name = trim($class_name_cp_val->getValue());
		}
	}
	
	// if qbo_class custom property does not exists or the member doesn't have a value for it, use member name to search the class
	if (!$fully_qualified_name) {
		// build query using fully qualified name of the member (concatenate all the hierarchy)
		$class_name = $member->getName();
		if ($member->getParentMemberId() > 0) {
			$member_parents = array_reverse($member->getAllParentMembersInHierarchy(false, false));
			
			$escaped_names = array();
			foreach ($member_parents as $member_parent) {
				$escaped_names[] = escape_character(escape_character($member_parent->getName(), ":"), "'");
			}
		}
		$escaped_names[] = escape_character(escape_character($class_name, ":"), "'");
		$fully_qualified_name = implode(':', $escaped_names);
	}
	
	// query the class to get the id
	$classes = $dataService->Query("select * from Class where FullyQualifiedName='$fully_qualified_name'");
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting class $class_name from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	if (count($classes) > 0) {
		// return class definition
		return $classes[0];
		
	} else if ($create_if_not_exists) {
		// if not found then create the class using name and parent definition if necessary
		$parent_class_def = null;
		if ($member->getParentMemberId() > 0) {
			$parent_class_def = get_qbo_class_def_by_member($dataService, $member->getParentMember());
		}
		
		// create class object
		$class = new QuickBooksOnline\API\Data\IPPClass();
		$class->Name = $class_name;
		if ($parent_class_def) {
			$class->ParentRef = $parent_class_def->Id;
			$class->SubClass = true;
		}
		$dataService->Add($class);
		
		// query the class to get the id
		$classes = $dataService->Query("select * from Class where FullyQualifiedName='$fully_qualified_name'");
		
		$error = $dataService->getLastError();
		if ($error) {
			Logger::log_r("Error getting class $class_name from Quickbooks (2)");
			Logger::log_r($error);
			throw new Exception(build_qbo_error_message($error));
		}
		if (count($classes) > 0) {
			// return class definition
			return $classes[0];
		}
	}
	
}

/**
 * Gets the qbo class definition for an invoice, expense, or any content object, using its classification and the dimension matched with classes
 * @param mixed $dataService
 * @param ContentDataObject $feng_object
 * @return QuickBooksOnline\API\Data\IPPClass
 */
function get_qbo_class_for_feng_object($dataService = null, $feng_object) {
	if (!$dataService) $dataService = get_qbo_data_service();
	
	$invoice_project_member = null;
	$projects_dim_id = Dimensions::findByCode('customer_project')->getId();
	$class_member = null;
	$dimension_id_for_qbo_classes = config_option('qbo_dimension_for_classes');

	$expense_cat_member = null;
	$expense_cat_dim = Dimensions::findByCode('expense_categories');
	$expense_cat_dim_id = $expense_cat_dim ? $expense_cat_dim->getId() : 0;
	
	// check if the invoice is classified in any member of the dimension for classes
	foreach ($feng_object->getMembers() as $m) {
		if ($m->getDimensionId() == $dimension_id_for_qbo_classes) {
			$class_member = $m;
		}
		if ($m->getDimensionId() == $projects_dim_id) {
			$invoice_project_member = $m;
		}
		// for expenses, get the expense category member to tru to find its related class
		if ($m->getDimensionId() == $expense_cat_dim_id && $feng_object instanceof PaymentReceipt) {
			$expense_cat_member = $m;
		}
	}

	// if object is an actual expense, try to get the class from expense category
	if ($expense_cat_member instanceof Member) {
		// if class not found using $expense_cat_member's custom property then don't create a new class and continue with normal flow
		$class_definition = get_qbo_class_def_by_member($dataService, $expense_cat_member, false);
		if ($class_definition) {
			return $class_definition;
		}
	}
	
	// if object is not classified in that dimension, then check if the invoice's project has a relation with that dimension and get the associated member of the project
	if (!$class_member && $invoice_project_member instanceof Member) {
		$class_dim = Dimensions::findById($dimension_id_for_qbo_classes);
		if ($class_dim) {
			$class_member_id = get_associated_status_member_id($invoice_project_member, $class_dim);
			if ($class_member_id > 0) {
				$class_member = Members::findById($class_member_id);
			}
		}
	}
	
	if ($class_member instanceof Member) {
		
		$sync_classes_with_first_level_members = config_option('qbo_sync_only_first_level_classes');// use config option
		if ($sync_classes_with_first_level_members) {
			$all_parents = $class_member->getAllParentMembersInHierarchy(true, false);
			$first_level_member = end($all_parents);
			if ($first_level_member instanceof Member) $class_member = $first_level_member;
		}
		
		return get_qbo_class_def_by_member($dataService, $class_member);
	}
}

/**
 * Gets the member associated to a qbo class, if it doesn't exist it will be created.
 * @param mixed $dataService
 * @param int $class_id
 * @return Member
 */
function get_member_for_qbo_class($dataService = null, $class_id) {
	if (!$dataService) $dataService = get_qbo_data_service();
	
	$class_member = null;
	$dimension_id_for_qbo_classes = config_option('qbo_dimension_for_classes');
	
	if ($dimension_id_for_qbo_classes > 0) {
		$class_ref = get_qbo_class_def_by_id($dataService, $class_id);
		if ($class_ref instanceof QuickBooksOnline\API\Data\IPPClass) {
			
			$parent_member = null;
			if ($class_ref->ParentRef > 0) {
				$parent_member = get_member_for_qbo_class($dataService = null, $class_ref->ParentRef);
			}
			
			$parent_cond = $parent_member == null ? "" : " AND parent_member_id=".$parent_member->getId();
			
			$class_member = Members::findOne(array("conditions" => "name LIKE '".$class_ref->Name."' $parent_cond"));
			if (!$class_member instanceof Member) {
				Env::useHelper('dimension');
				$mem_controller = new MemberController();
				$new_member = new Member();
				$dim_ots = DimensionObjectTypes::instance()->getObjectTypeIdsByDimension($dimension_id_for_qbo_classes);
				
				$member_data = array(
						'name' => $class_ref->Name,
						'object_type_id' => $dim_ots[0],
						'dimension_id' => $dimension_id_for_qbo_classes,
						'parent_member_id' => $parent_member ? $parent_member->getId() : 0
				);
				
				$class_member = $mem_controller->saveMember($member_data, $new_member);
			}
			return $class_member;
		}
	}
}

/**
 *
 * @param mixed $dataService
 * @param string $type
 * @param string $subtype
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPAccount
 */
function get_qbo_account_by_type($dataService, $type, $subtype) {
	$accountArray = $dataService->Query("select * from Account where AccountType='" . $type . "' and AccountSubType='" . $subtype . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($accountArray) && sizeof($accountArray) > 0) {
			return current($accountArray);
		}
	}
	
	// Create Income Account
	$incomeAccountRequestObj = QuickBooksOnline\API\Facades\Account::create([
			"AccountType" => $type,
			"AccountSubType" => $subtype,
			"Name" => $subtype
	]);
	$incomeAccountObject = $dataService->Add($incomeAccountRequestObj);
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		return $incomeAccountObject;
	}
}

/**
 *
 * @param mixed $dataService
 * @param string $name
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPAccount
 */
function get_qbo_account_by_name($dataService, $name) {
	$accountArray = $dataService->Query("select * from Account where Name='" . $name . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($accountArray) && sizeof($accountArray) > 0) {
			return current($accountArray);
		}
	}
}
/**
 *
 * @param mixed $dataService
 * @param int $id
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPAccount
 */
function get_qbo_account_by_id($dataService, $id) {
	if ($id == '') return null;
	$accountArray = $dataService->Query("select * from Account where Id='" . $id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($accountArray) && sizeof($accountArray) > 0) {
			return current($accountArray);
		}
	}
}

/**
 * Returns all the QuickBooksOnline\API\Data\IPPAccount objects
 * @param mixed $dataService
 * @throws Exception
 * @return array
 */
function get_qbo_all_accounts($dataService = null) {
	if (!$dataService) $dataService = get_qbo_data_service();

	$accountArray = $dataService->Query("select * from Account maxresults 500");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		return $accountArray;
	}
}

/**
 * Returns the QuickBooksOnline\API\Data\IPPTaxCode object
 * @param mixed $dataService
 * @throws Exception
 * @return array of QuickBooksOnline\API\Data\IPPTaxCode
 */
function get_all_qbo_taxes($dataService = null) {

	if (!$dataService) $dataService = get_qbo_data_service();

	$taxArray = $dataService->Query("select * from TaxCode");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		return $taxArray;
	}

}

/**
 * Returns the QuickBooksOnline\API\Data\IPPTaxCode object
 * @param mixed $dataService
 * @param integer $qbo_tax_id the tax id in quickbooks
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPTaxCode
 */
function get_qbo_tax($dataService = null, $qbo_tax_id) {

	if (!$dataService) $dataService = get_qbo_data_service();

	$taxArray = $dataService->Query("select * from TaxCode WHERE id = '$qbo_tax_id'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		return current($taxArray);
	}

}

/**
 * Returns the QuickBooksOnline\API\Data\IPPTaxRate object
 * @param mixed $dataService
 * @param integer $qbo_tax_id the tax id in quickbooks
 * @throws Exception
 * @return array QuickBooksOnline\API\Data\IPPTaxRate
 */
function get_qbo_tax_rate($dataService = null, $qbo_tax_rate_id) {

	if (!$dataService) $dataService = get_qbo_data_service();

	$taxArray = $dataService->Query("select * from TaxRate WHERE id = '$qbo_tax_rate_id'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		Logger::log_r(current($taxArray));
		return current($taxArray);
	}

}


function build_invoice_tax_data_to_sync($dataService = null, IncomeInvoice $feng_invoice) {

	$$qbo_invoice_taxes = array();

	$feng_taxes = $feng_invoice->getTaxes();

	foreach ($feng_taxes as $feng_tax) {

		$feng_tax_id = $feng_tax->getTaxId();
		$income_tax = IncomeTaxes::findById($feng_tax_id);
		if (!$income_tax) continue;

		// get the qbo tax id from the mapping
		$qbo_tax_id = $income_tax->getColumnValue('quickbooks_online_id');

		// get the qbo TaxCode object
		$qbo_tax = get_qbo_tax($dataService, $qbo_tax_id);
		
		if ($qbo_tax instanceof QuickBooksOnline\API\Data\IPPTaxCode) {

			$taxable_amount = $feng_invoice->getSubtotal();

			$qbo_tax_lines = [];

			// get the rates for this tax
			$qbo_tax_rates = is_array($qbo_tax->SalesTaxRateList) ? $qbo_tax->SalesTaxRateList : array($qbo_tax->SalesTaxRateList);
			
			// for each rate apply it to the taxable amount and build a tax line detail
			foreach ($qbo_tax_rates as $qbo_tax_rate_ref) {

				if ($qbo_tax_rate_ref->TaxRateRef) {
					$rate_ref_id = $qbo_tax_rate_ref->TaxRateRef;
				} else if ($qbo_tax_rate_ref->TaxRateDetail) {
					$rate_ref_id = $qbo_tax_rate_ref->TaxRateDetail->TaxRateRef;
				} else {
					continue;
				}

				$qbo_tax_rate = get_qbo_tax_rate($dataService, $rate_ref_id);
				if ($qbo_tax_rate instanceof QuickBooksOnline\API\Data\IPPTaxRate && $qbo_tax_rate->Active) {

					// get the rate (percentage)
					$tax_rate = $qbo_tax_rate->RateValue;

					if ($feng_tax->getValue() == $tax_rate) {
						// if tax has only one rate, then we can use the already calculated tax value in feng/evx
						$tax_amount = $feng_tax->getFixedValue();
					} else {
						// if qbo tax is splitted in more than one rate we have to calculate it for each rate
						$tax_amount = number_format($taxable_amount * $tax_rate / 100, 2);
					}

					// build the data to use at sync for this tax
					$tax_rate_data = [
						"DetailType" => "TaxLineDetail",
						"Amount" => $tax_amount,
						"TaxLineDetail" => [
							"NetAmountTaxable" => $taxable_amount,
							"TaxPercent" => $tax_rate,
							"TaxRateRef" => [
								"value" => $qbo_tax_rate->Id,
								"name" => $qbo_tax_rate->Name,
							],
							"PercentBased" => true
						]
					];

					$qbo_tax_lines[] = $tax_rate_data;

				}
			}

			$qbo_invoice_taxes[] = [
				"TxnTaxCodeRef" => [
					"value" => $qbo_tax->Id,
					"name" => $qbo_tax->Name,
				],
				"TotalTax" => $feng_tax->getFixedValue(),
				"TaxLine" => $qbo_tax_lines,
			];

		}
	}

	return $qbo_invoice_taxes;
}


/**************************************************************************************/
/**************************** MAIN WORKFLOW FUNCTIONS *********************************/
/**************************************************************************************/

/**
 * This function is used to remove unwanted characters from client and projects when trying to create them in Quickbooks
 * @param string The original client/project name
 * @return string The cleaned client/project name
 */
function qbo_clean_client_or_project_name($name) {

	$name = str_replace("–", "-", $name); // it seems the same char, but the one to replace is a bit longer and it's not standard and qb gives an error
	
	// remove the non authorized chars specified in api docs
	$name = str_replace(array(":","\n","\t"), " ", $name);

	return $name;
}

/**
 * Voids an invoice in quickbooks from a Feng's invoice
 * @param IncomeInvoice $invoice
 * @throws Exception
 * @return QuickBooksOnline\API\Facades\Invoice
 */
function void_invoice_in_qbo(IncomeInvoice $invoice) {

	if (!check_if_object_can_be_sync($invoice)) return;
	
	$dataService = get_qbo_data_service();
	
	$qbo_invoice_data = array();
	$qbo_invoice_id = $invoice->getColumnValue("quickbooks_online_id");
	
	$old_qbo_invoice_obj = get_qbo_invoice($dataService, $qbo_invoice_id);
	if ($old_qbo_invoice_obj) {
		$qbo_invoice_data['Id'] = $qbo_invoice_id;
		$qbo_invoice_data['SyncToken'] = $old_qbo_invoice_obj->SyncToken;
		$qbo_invoice_data['sparse'] = true; // only updates the sent information, does not override other data updated in qbo
	} else {
		return;
	}
	
	$entity = QuickBooksOnline\API\Facades\Invoice::create($qbo_invoice_data);
	
	$return_obj = $dataService->Void($entity);
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error voiding invoice in Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	
	return $return_obj;
}

/**
 * Deletes an invoice in quickbooks from a Feng's invoice
 * @param IncomeInvoice $invoice
 * @throws Exception
 * @return QuickBooksOnline\API\Facades\Invoice
 */
function delete_invoice_from_qbo(IncomeInvoice $invoice) {
	
	$dataService = get_qbo_data_service();
	
	$qbo_invoice_data = array();
	$qbo_invoice_id = $invoice->getColumnValue("quickbooks_online_id");
	
	$old_qbo_invoice_obj = get_qbo_invoice($dataService, $qbo_invoice_id);
	if ($old_qbo_invoice_obj) {
		$qbo_invoice_data['Id'] = $qbo_invoice_id;
		$qbo_invoice_data['SyncToken'] = $old_qbo_invoice_obj->SyncToken;
		$qbo_invoice_data['sparse'] = true; // only updates the sent information, does not override other data updated in qbo
	} else {
		return;
	}
	
	$entity = QuickBooksOnline\API\Facades\Invoice::create($qbo_invoice_data);
	
	$return_obj = $dataService->Delete($entity);
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error deleting invoice in Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	
	$invoice->setColumnValue("quickbooks_online_id", 0);
	$invoice->save();
	
	return $return_obj;
}

function sync_contact_with_qbo($feng_contact_id, $dataService = null)
{

	$qbo_employee_obj = null;

	if (is_null($dataService)) {
		$dataService = get_qbo_data_service();
	}

	$feng_contact = Contacts::findById($feng_contact_id);
	if (!$feng_contact) return null;

	$qbo_employee_id = $feng_contact->getColumnValue('quickbooks_online_id');

	// if employee qbo_id is in database
	if ($qbo_employee_id > 0) {

		$old_qbo_employee_obj = get_qbo_employee($dataService, $qbo_employee_id);

		if ($old_qbo_employee_obj) {
			$qbo_employee_obj = $old_qbo_employee_obj;
		}
	}
	
	

	if (!$qbo_employee_obj) {

		// if employee id is not in database try to find employee by email address
		$old_qbo_employee_array = $dataService->Query("select * from employee where PrimaryEmailAddr='" . strtolower($feng_contact->getEmailAddress()) . "'");

		if (count($old_qbo_employee_array) > 0) {
			$qbo_employee_obj = $old_qbo_employee_array[0];
		}

		if (!$qbo_employee_obj) {
			$old_qbo_employee_array = $dataService->Query("select * from employee where DisplayName='" . escape_single_quotes($feng_contact->getObjectName()) . "'");
			if (count($old_qbo_employee_array) > 0) {
				$qbo_employee_obj = $old_qbo_employee_array[0];
			}
		}

		// if couldn't find it by email then create it
		if (!$qbo_employee_obj) {

			$qbo_employee_data = array(
				'GivenName' => $feng_contact->getFirstName(),
				'FamilyName' => $feng_contact->getSurname(),

				'PrimaryEmailAddr' => array(
					'Address' => $feng_contact->getEmailAddress(),
				)
			);

			$contact_addresses = ContactAddresses::getByContact($feng_contact);

			if (sizeof($contact_addresses) > 0) {
				$addr = array();
				$addr['CountrySubDivisionCode'] = $contact_addresses[0]->getState();
				$addr['City'] = $contact_addresses[0]->getCity();
				$addr['PostalCode'] = $contact_addresses[0]->getZipCode();
				$addr['Line1'] = $contact_addresses[0]->getStreet();

				$qbo_employee_data['PrimaryAddr'] = $addr;
			}

			if(empty($addr['CountrySubDivisionCode']) || empty($addr['City']) || empty($addr['PostalCode']) || empty($addr['Line1'])){
				throw new Exception('User information incomplete');
			}

			$qbo_employee = QuickBooksOnline\API\Facades\Employee::create($qbo_employee_data);

			$qbo_employee_obj = $dataService->Add($qbo_employee);

			$error = $dataService->getLastError();
			if ($error) {
				Logger::log_r("Error creating employee at Quickbooks:\n".print_r($qbo_employee_data,1));
				Logger::log_r($error);
				throw new Exception(build_qbo_error_message($error));
			}			
		}

		//save employee id and sync token
		if ($qbo_employee_obj) {
			$feng_contact->setColumnValue('quickbooks_online_id', $qbo_employee_obj->Id);
			$feng_contact->setColumnValue('quickbooks_online_sync_token', $qbo_employee_obj->SyncToken);
			$feng_contact->save();
		}
	}

	return $qbo_employee_obj;
}


function delete_time_from_qbo(Timeslot $timeslot) {
	Logger::log_r("Por aca delete_time_from_qbo");
	$dataService = get_qbo_data_service();

	$qbo_timeslot_data = array();
	$qbo_timeslot_id = $timeslot->getColumnValue("quickbooks_online_id");

	$old_qbo_timeslot_obj = get_qbo_time_activity($dataService, $qbo_timeslot_id);

	if ($old_qbo_timeslot_obj) {
		$qbo_timeslot_data['Id'] = $qbo_timeslot_id;
		$qbo_timeslot_data['SyncToken'] = $old_qbo_timeslot_obj->SyncToken;
		$qbo_timeslot_data['sparse'] = false; // only updates the sent information, does not override other data updated in qbo
	} else {
		return;
	}
	
	$entity = QuickBooksOnline\API\Facades\TimeActivity::create($qbo_timeslot_data);
	
	$return_obj = $dataService->Delete($entity);
	
	$error = $dataService->getLastError();
	if ($error) {
		throw new Exception(build_qbo_error_message($error));
	}
	
	$timeslot->setColumnValue("quickbooks_online_id", 0);
	$timeslot->setColumnValue("quickbooks_online_sync_token", 0);
	$timeslot->save();
	
	return $return_obj;
}

function save_timeslot_to_qbo(Timeslot $timeslot, $dataService=null) {

	// Quicktimes needs both start_time and end_time to save correctly, when starting timer, end_time is not set
	if(is_null($timeslot->getEndTime())) return;

	if (!check_if_object_can_be_sync($timeslot)) return;
	
	if (is_null($dataService)) {
		$dataService = get_qbo_data_service();
	}

	$qbo_time_activity_id = $timeslot->getColumnValue('quickbooks_online_id');
	$qbo_time_activity_token = $timeslot->getColumnValue('quickbooks_online_sync_token');	

	$client_ot = ObjectTypes::findByName('customer');
	$project_ot = ObjectTypes::findByName('project');


	// get the item class
	$qbo_preferences = get_qbo_preferences($dataService);
	$use_class_in_whole_invoice = $qbo_preferences->AccountingInfoPrefs->ClassTrackingPerTxn;
	$use_class_in_invoice_lines = $qbo_preferences->AccountingInfoPrefs->ClassTrackingPerTxnLine;

	$class_definition = null;
	if ($use_class_in_invoice_lines || $use_class_in_whole_invoice) {
		$class_definition = get_qbo_class_for_feng_object($dataService, $timeslot);
	}


	$tz_offset = $timeslot->getTimezoneValue();

	// format txndate time field
	$txn_date = new DateTimeValue($timeslot->getStartTime()->getTimestamp());
	$txn_date->advance($tz_offset);
	$txn_date = $txn_date->format('Y-m-d');

	// format start time field
	$start_time = new DateTimeValue($timeslot->getStartTime()->getTimestamp());
	$start_time->add('s', $tz_offset);
	$start_time_str = $start_time->toMySQL();
	$start_time_str = str_replace(' ', 'T', $start_time_str);

	// format end time field
	$end_time = new DateTimeValue($timeslot->getEndTime()->getTimestamp());
	$end_time->add('s', $tz_offset);
	$end_time_str = $end_time->toMySQL();
	$end_time_str = str_replace(' ', 'T', $end_time_str);

	// get the time entry item depending on the labor category
	$hours_item_ref = get_object_qbo_labor_item($timeslot);
	if (!$hours_item_ref) {
		$hours_item_array = get_qbo_items_definition($dataService, null, 'Hours');
		if (is_array($hours_item_array) && count($hours_item_array) > 0) {
			$hours_item_ref = $hours_item_array[0];
		}
	}

	// find client and project members
	$feng_timeslot_members = $timeslot->getMembers();
	$client_member = null;
	$project_member = null;
	foreach ($feng_timeslot_members as $member) {

		if ($member->getObjectTypeId() == $client_ot->getId()) {
			$client_member = $member;
		}
		if ($member->getObjectTypeId() == $project_ot->getId()) {
			$project_member = $member;
		}
	}
	
	// get qbo customer for the evx client
	if ($client_member) {
		$customerRef = get_qbo_customer_object($dataService, $client_member);
	}
	
	// get qbo job/project for the evx project
	if ($customerRef && $project_member instanceof Member) {
		$projectRef = get_qbo_project_object($dataService, $project_member, $customerRef);
	}
		
	// get the employee object for timeslot's contact
	$qbo_employee_obj = sync_contact_with_qbo($timeslot->getContactId());

	// set as billable / not billable
	if(($timeslot->getColumnValue('invoicing_status') != 'non_billable') && ($projectRef || $customerRef)) {
		$ts_status = 'Billable';	
	} else {
		$ts_status = 'NotBillable';	
	}

	$total_minutes = $timeslot->getColumnValue('worked_time');
	$hours = floor($total_minutes / 60);
	$minutes = $total_minutes % 60;
	
	// build the data to sync
	$time_activity_data = array(
		'NameOf' => 'Employee',
		'Description' => escape_single_quotes($timeslot->getDescription()),
		'EmployeeRef' => array(
			'value' => $qbo_employee_obj ? $qbo_employee_obj->Id : '0',
			'name' => $qbo_employee_obj ? escape_single_quotes($qbo_employee_obj->DisplayName) : '',
		),
		'TxnDate' => $txn_date,
	//	'StartTime' => $start_time_str,
	//	'EndTime' => $end_time_str,
		'Hours' => $hours,
		'Minutes' => $minutes,
		'BillableStatus' => $ts_status,
		'HourlyRate' => $timeslot->getHourlyBilling(),
	);

	// set the project / customer of the time entry
	if ($projectRef || $customerRef) {
		$time_activity_data['CustomerRef'] = array(
			'value' => $projectRef ? $projectRef->Id : $customerRef->Id
		);			
	}

	// set the item reference of the time entry
	if ($hours_item_ref) {
		$time_activity_data['ItemRef'] = array(
			'value' => $hours_item_ref->Id,
			'name' => $hours_item_ref->Name,
		);
	}

	// set the class
	if ($use_class_in_invoice_lines && $class_definition) {
		$time_activity_data["ClassRef"] = array(
				'value' => $class_definition->Id,
				'name' => $class_definition->Name,
		);
	}

	// get the current time activity object fo quickbooks 
	if ($qbo_time_activity_id > 0) {
		
		$old_qbo_time_activity_obj = get_qbo_time_activity($dataService, $qbo_time_activity_id);
		
		if ($old_qbo_time_activity_obj) {		
			$time_activity_data['Id'] = $qbo_time_activity_id;
			$time_activity_data['SyncToken'] = $old_qbo_time_activity_obj->SyncToken;
			$time_activity_data['sparse'] = true; // only updates sent information, does not override other data updated in qbo 
		} else {
			$qbo_time_activity_id = 0;
		}
	
	}
	
	//Logger::log_r($time_activity_data);
	$qbo_time_activity = QuickBooksOnline\API\Facades\TimeActivity::create($time_activity_data);

	if (isset($time_activity_data['Id'])) {
		// update in qbo
		$resultingTimeActObj = $dataService->Update($qbo_time_activity);
	} else {
		// add to qbo
		$resultingTimeActObj = $dataService->Add($qbo_time_activity);		
	}

	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error when trying to synch timeslot with Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	
	$timeActivityId = $resultingTimeActObj->Id;

	//save time id and sync token
	if ($qbo_time_activity_id == 0) {
		$timeslot->setColumnValue('quickbooks_online_id', $timeActivityId);	
	}
	$timeslot->setColumnValue('quickbooks_online_sync_token', $resultingTimeActObj->SyncToken);
	$timeslot->save();

	return $timeActivityId;
}

/**
 * Saves an invoice in quickbooks from a Feng's invoice
 * @param IncomeInvoice $invoice
 * @throws Exception
 * @return null
 */
function save_invoice_to_qbo(IncomeInvoice $feng_invoice, $dataService=null) {

	if (!check_if_object_can_be_sync($feng_invoice)) return;

	if (is_null($dataService)) {
		$dataService = get_qbo_data_service();
	}
	
	$feng_invoice_members = $feng_invoice->getMembers();
	$client_member = null;
	$project_member = null;
	$client_ot = ObjectTypes::findByName('customer');
	$project_ot = ObjectTypes::findByName('project');
	
	foreach ($feng_invoice_members as $member) {
		if ($member->getObjectTypeId() == $client_ot->getId()) {
			$client_member = $member;
		}
		if ($member->getObjectTypeId() == $project_ot->getId()) {
			$project_member = $member;
		}
	}
	
	$qbo_preferences = get_qbo_preferences($dataService);
	
	$use_class_in_whole_invoice = $qbo_preferences->AccountingInfoPrefs->ClassTrackingPerTxn;
	$use_class_in_invoice_lines = $qbo_preferences->AccountingInfoPrefs->ClassTrackingPerTxnLine;
	$allow_shipping = $qbo_preferences->SalesFormsPrefs->AllowShipping;
	
	$class_definition = null;
	if ($use_class_in_invoice_lines || $use_class_in_whole_invoice) {
		$class_definition = get_qbo_class_for_feng_object($dataService, $feng_invoice);
	}

	// build transaction tax data
	$qbo_invoice_tax = build_invoice_tax_data_to_sync($dataService, $feng_invoice);
	
	$bill_charges_already_included_in_line_linked_transaction = array();
	$times_already_included_in_line_linked_transaction = array();

	//$total_discount_amount = 0;
	$markup_line_amount = 0;
	$qb_lines = array();
	$invoice_lines = $feng_invoice->getInvoiceLines();
	foreach ($invoice_lines as $invoice_line) {
		/*  @var $invoice_line IncomeInvoiceLine */
		
		if ($invoice_line->getTotal() < 0) {
			
			// add a discount
			$qb_line = array(
				"Amount" => abs($invoice_line->getTotal()),
				"DetailType" => "DiscountLineDetail",
				"DiscountLineDetail" => array(
						"PercentBased" => false,
						//"DiscountPercent" => abs($invoice_line->getTotal()),
						/*"DiscountAccountRef" => [
								"value" => "86",
								"name" => "Discounts given"
						]*/
				),
				//"Description" => $invoice_line->getDetail()
			);
			
		} else {
		
			//$itemRef = get_qbo_invoice_item_object($dataService, $invoice_line);
			
			$qb_line = array(
					"Amount" => str_replace(',', '', $invoice_line->getTotal()),
					"DetailType" => "SalesItemLineDetail",
					"Description" => $invoice_line->getDetail(),
					"LinkedTxn" => array(),
					"SalesItemLineDetail" => [
							"Qty" => str_replace(',', '', $invoice_line->getAmount()),
						//	"UnitPrice" => str_replace(',', '', $invoice_line->getUnitValue()),
							"UnitPrice" => round(str_replace(',', '', $invoice_line->getTotal()) / str_replace(',', '', $invoice_line->getAmount()), 7),
					]
			);
			
			// if invoice line has a date, add it to the qbo line detail data
			if ($invoice_line->getDate() instanceof DateTimeValue) {
				$qb_line["SalesItemLineDetail"]["ServiceDate"] = $invoice_line->getDate()->format("Y-m-d");
			}
			
			$itemRef = null;
			
			// If it is a line generated from timeslots then try to associate the user to its service item defined in Quickbooks
			if ($invoice_line->getFromObjectTypeId() == 0 || $invoice_line->getFromObjectTypeId() == Timeslots::instance()->getObjectTypeId()) {
				
				$use_general_hours_item = false;

				if ($invoice_line->getCategoryId() > 0) {

					$item_id = get_qbo_item_id_by_category_id($invoice_line->getCategoryId());
					if (trim($item_id) != '') {
						$itemRef = get_qbo_item_definition($dataService, $item_id);
					}
					if (!$itemRef) {
						$use_general_hours_item = true;
					}
					
				}

				if (!$itemRef && $invoice_line->getUserId() > 0) {
					
					$row = DB::executeOne("SELECT * FROM ".TABLE_PREFIX."qbo_user_hour_items_match WHERE user_id=".$invoice_line->getUserId());
					if ($row && $row['qbo_item_id'] > 0) {
						// use item linked to user if exsists
						$itemRef = get_qbo_item_definition($dataService, $row['qbo_item_id']);
					} else {
						$use_general_hours_item = true;
					}
					
				}
				
				// if no qbo item was found for the user, then use the labor category to get or generate the item
				if (!$itemRef && $invoice_line->getCategoryId() > 0) {
					
					$labor_cat = Members::getMemberById($invoice_line->getCategoryId());
					if ($labor_cat instanceof Member) {
						$full_labor_cat_name = $labor_cat->getPath();
						$full_labor_cat_name .= ($full_labor_cat_name == "" ? "" : "/") . $labor_cat->getName();
						
						$qbo_laborcat_array = get_qbo_items_definition($dataService, null, $full_labor_cat_name);
						if (is_array($qbo_laborcat_array) && count($qbo_laborcat_array) > 0) {
							$itemRef = $qbo_laborcat_array[0];
						} else {
							// create item in qbo
							$qbo_item = qbo_create_item($dataService, $full_labor_cat_name, $invoice_line->getCategoryId());
							if ($qbo_item) $itemRef = $qbo_item;
						}
					}
				}
				
				if (!$itemRef && $use_general_hours_item) {
					// if user is not linked to an hour type, then try to use the generic 'Hours' service item
					$hours_array = get_qbo_items_definition($dataService, null, 'Hours');
					if (is_array($hours_array) && count($hours_array) > 0) {
						$itemRef = $hours_array[0];
					}
				}


				if ($invoice_line->getRelObjectId() > 0) {

					$line_timeslot = Timeslots::findById($invoice_line->getRelObjectId());
					if ($line_timeslot) {

						if ($line_timeslot->getColumnValue('quickbooks_online_id') > 0 && !in_array($line_timeslot->getColumnValue('quickbooks_online_id'), $times_already_included_in_line_linked_transaction)) {
							$timeslot_id_qb = $line_timeslot->getColumnValue('quickbooks_online_id');
						} else {
							$timeslot_id_qb = save_timeslot_to_qbo($line_timeslot, $dataService);
						}
			
						if($timeslot_id_qb){

							$qb_line['LinkedTxn'][] = array(
								"TxnId" => $timeslot_id_qb,
								"TxnType" => "TimeActivity"
							);
			
							$times_already_included_in_line_linked_transaction[] = $line_timeslot->getColumnValue('quickbooks_online_id');
						}
					}
				}
				
			} else if ($invoice_line->getFromObjectTypeId() == PaymentReceipts::instance()->getObjectTypeId()) {

				// get the item id using the expense category mapping
				if ($invoice_line->getCategoryId() > 0) {
					$item_id = get_qbo_item_id_by_category_id($invoice_line->getCategoryId());
					if (trim($item_id) != '') {
						$itemRef = get_qbo_item_definition($dataService, $item_id);
					}
				}
				
				// if not found using category, get the item using product type name
				if (!$itemRef && $invoice_line->getProductTypeId() > 0) {
					$pt = ProductTypes::getById($invoice_line->getProductTypeId());
					if ($pt instanceof ProductType) {
						$qbo_pt_array = get_qbo_items_definition($dataService, null, $pt->getName());
						if (is_array($qbo_pt_array) && count($qbo_pt_array) > 0) {
							$itemRef = $qbo_pt_array[0];
						} else {
							// create item in qbo
							$qbo_item = qbo_create_item($dataService, $pt->getName(), $invoice_line->getCategoryId());
							if ($qbo_item) $itemRef = $qbo_item;
						}
					}
				}

				// send this line with cost and accumulate markup amount in another new line
				if ($invoice_line->getUnitValue() != $invoice_line->getUnitCost()) {

					$this_line_markup = $invoice_line->getAmount() * ($invoice_line->getUnitValue() - $invoice_line->getUnitCost());
					$markup_line_amount += $this_line_markup;

					// put the cost as the line amount
					$qb_line["Amount"] = round($invoice_line->getAmount() * $invoice_line->getUnitCost(), 3);
					// put the unit cost as the rate for this line
					$qb_line["SalesItemLineDetail"]["UnitPrice"] = $invoice_line->getUnitCost();
/*
					$markup_percent = (($invoice_line->getUnitValue() / $invoice_line->getUnitCost()) - 1) * 100;

					$qb_line["SalesItemLineDetail"]["MarkupInfo"] = array(
						"Percent" => number_format($markup_percent, 3),
					);
*/
				}


				// link expense bill charge id if possible
				if ($invoice_line->getRelObjectId() > 0) {

					$feng_actual_exp = PaymentReceipts::findById($invoice_line->getRelObjectId());
					
					// if not synchronized yet, then send the expense to quickbooks
					if ($feng_actual_exp->getColumnValue('quickbooks_online_id') == 0 && in_array($feng_actual_exp->getExpenseType(), config_option("expense_types_to_sync"))) {
						$invoiced_expense = save_actual_expense_to_qbo($feng_actual_exp, $dataService);
					}
					// get the reimburse charge id (billabe expense charge)
					$billable_expense_id = $feng_actual_exp->getColumnValue('qbo_reimburse_charge_id');
					if ($billable_expense_id > 0 && !in_array($billable_expense_id, $bill_charges_already_included_in_line_linked_transaction)) {

						// ensure that reimburse charge exists
						$reimburse_charge_obj = get_reimburse_charge_object_by_id($dataService, $billable_expense_id);

						// if billable expense charge exists in qbo then send it as an associated object to the transaction
						if ($reimburse_charge_obj) {

							$qb_line['LinkedTxn'][] = array(
								"TxnId" => $billable_expense_id,
								"TxnType" => "ReimburseCharge",
								"TxnLineId" => 1
							);

							$bill_charges_already_included_in_line_linked_transaction[] = $billable_expense_id;
						}
					}
				}
				
			}
			
			// if item found then build the item structure for the line
			if ($itemRef) {
				$qb_line["SalesItemLineDetail"]["ItemRef"] = [
						"value" => $itemRef->Id,
						"name" => $itemRef->Name,
				];

				// set the line item's account using the account specified in the item object
				$itemAccountRef = null;
				if ($itemRef->IncomeAccountRef) {
					if (is_numeric($itemRef->IncomeAccountRef)) {
						$item_account_id = $itemRef->IncomeAccountRef;
						$itemAccountRef = get_qbo_account_by_id($dataService, $item_account_id);
						
					} else {
						$itemAccountRef = $itemRef->IncomeAccountRef;
					}
				}
				
				if ($itemAccountRef) {
					$qb_line["SalesItemLineDetail"]["ItemAccountRef"] = [
						"value" => $itemAccountRef->Id,
						"name" => $itemAccountRef->Name,
					];
				}
			}
			
			if ($use_class_in_invoice_lines && $class_definition) {
				$qb_line["SalesItemLineDetail"]["ClassRef"] = array(
						'value' => $class_definition->Id,
						'name' => $class_definition->Name,
				);
			}

			// if the invoice includes taxes then set the TaxCodeRef attributes of the invoice lines
			if (!empty($qbo_invoice_tax)) {
				$qb_line["SalesItemLineDetail"]["TaxCodeRef"] = [
					"value" => "TAX",
					"name" => "TAX",
				];
			}
			
		}
		
		$qb_lines[] = $qb_line;
	}

	
	if ($markup_line_amount > 0) {

		//$markup_line_amount = number_format($markup_line_amount, 3);
		$markup_line_amount = round($markup_line_amount, 3);
		
		$qbo_markup_item_array = get_qbo_items_definition($dataService, null, "Markup");
		if (is_array($qbo_markup_item_array) && count($qbo_markup_item_array) > 0) {
			$markupItemRef = $qbo_markup_item_array[0];
		} else {
			$markupItemRef = qbo_create_item($dataService, "Markup", null, $qbo_preferences->VendorAndPurchasesPrefs->DefaultMarkupAccount);
		}

		$markupAccountRef = null;
		if ($qbo_preferences->VendorAndPurchasesPrefs->DefaultMarkupAccount) {
			$markupAccountRef = get_qbo_account_by_id($dataService, $qbo_preferences->VendorAndPurchasesPrefs->DefaultMarkupAccount);
		}

		$markup_qb_line = array(
			"Amount" => $markup_line_amount,
			"DetailType" => "SalesItemLineDetail",
			"Description" => "Markup",
			"SalesItemLineDetail" => [
					"Qty" => "1",
					"UnitPrice" => $markup_line_amount,
					"ItemRef" => [
						"value" => $markupItemRef->Id,
						"name" => $markupItemRef->Name,
					],
			]
		);
		if ($markupAccountRef) {
			$markup_qb_line["SalesItemLineDetail"]["ItemAccountRef"] = [
				"value" => $markupAccountRef->Id,
				"name" => $markupAccountRef->Name,
			];
		}
		$qb_lines[] = $markup_qb_line;
	}
	
	
	// if invoice already exists in qbo, then use this id to update it
	$qbo_invoice_id = $feng_invoice->getColumnValue('quickbooks_online_id');
	
	$customerRef = get_qbo_customer_object($dataService, $client_member);
	
	$projectRef = null;
	if ($customerRef && $project_member instanceof Member) {
		$projectRef = get_qbo_project_object($dataService, $project_member, $customerRef);
	}
	
	$billing_email = $feng_invoice->getCustomerEmail();
	
	$address_line4 = trim($feng_invoice->getCustomerCity());
	$address_line4 .= ($address_line4=='' ? '' : ', ') . trim($feng_invoice->getCustomerState()." ".$feng_invoice->getCustomerZipcode());
	$billing_address = array(
			"Line1" => $client_member->getName(),
			"Line2" => "",//$client_member->getName(),
			"Line3" => $feng_invoice->getCustomerAddress(),
			"Line4" => $address_line4,
	);
	if ($feng_invoice->getCustomerCountry()) {
		$billing_address["Country"] = Countries::getCountryNameByCode($feng_invoice->getCustomerCountry());
	}
	
	$currency = Currencies::instance()->getCurrency($feng_invoice->getCurrencyId());
	
	$qbo_invoice_data = [
			"Line" => $qb_lines,
			"TxnDate" => $feng_invoice->getDate()->format("Y-m-d"),
			"CustomerRef"=> [
					"value"=> $projectRef ? $projectRef->Id : $customerRef->Id
			],
			"BillAddr" => $billing_address,
			"BillEmail" => [
					"Address" => $billing_email
			],
			"CurrencyRef" => [
					"value" => $currency->getShortName()
			]
	];

	// Set the invoice tax data
	if (!empty($qbo_invoice_tax)) {
		$qbo_invoice_data['TxnTaxDetail'] = $qbo_invoice_tax;
	}

	
	// build custom fields data to sync
	$custom_fields_formatted = build_qbo_custom_fields_object($feng_invoice, $qbo_preferences);
	if ($custom_fields_formatted) {
		$qbo_invoice_data["CustomField"] = $custom_fields_formatted;
	}
	
	// set invoice memo with our additional information field
	if ($feng_invoice->getAdditionalInformation() != '') {
		$qbo_invoice_data["CustomerMemo"] = [
				"value" => $feng_invoice->getAdditionalInformation(),
		];
	}
	
	// get invoice payment period value
	$inv_payment_period = null;
	$period_cp = CustomProperties::getCustomPropertyByCode($feng_invoice->getObjectTypeId(), "invoice_payment_period");
	if ($period_cp) {
		$cp_val = CustomPropertyValues::getCustomPropertyValue($feng_invoice->getId(), $period_cp->getId());
		if ($cp_val) $inv_payment_period = $cp_val->getValue();
	}
	
	// if invoice is "Due on receipt", set the due date equal to invoice date
	if ($inv_payment_period == 'Due on receipt') {
		$qbo_invoice_data["DueDate"] = $qbo_invoice_data["TxnDate"];
	}
	if ($feng_invoice->getExpirationDate() instanceof DateTimeValue) {
		$qbo_invoice_data["DueDate"] = $feng_invoice->getExpirationDate()->format("Y-m-d");
	}
	
	// use Quickbooks default terms (e.g.: "Net 30")
	$term_id = $qbo_preferences->SalesFormsPrefs->DefaultTerms;
	$term_obj = get_qbo_payment_term_by_name($dataService, $inv_payment_period);
	if ($term_obj) $term_id = $term_obj->Id;
	if ($term_id) {
		$qbo_invoice_data["SalesTermRef"] = [
				"value" => $term_id
		];
	}
	
	// set the class id and name
	if ($use_class_in_whole_invoice && $class_definition) {
		$qbo_invoice_data["ClassRef"] = array(
				'value' => $class_definition->Id,
				'name' => $class_definition->Name,
		);
	}
	
	// add shipping info
	if ($allow_shipping) {
		$shipping_address = null;
		$shipping_address_cp = CustomProperties::getCustomPropertyByCode($feng_invoice->getObjectTypeId(), "shipping_address");
		if ($shipping_address_cp) {
			$cp_val = CustomPropertyValues::getCustomPropertyValue($feng_invoice->getId(), $shipping_address_cp->getId());
			if ($cp_val) $shipping_address = $cp_val->getValue();
		}
		if ($shipping_address) {
			$exploded = explode("|", escape_character($shipping_address));
			if (count($exploded) > 0) {
				$ship_street = array_var($exploded, 1, '');
				$ship_city = array_var($exploded, 2, '');
				$ship_state = array_var($exploded, 3, '');
				//$ship_country = array_var($exploded, 4, '');
				$ship_zip_code = array_var($exploded, 5, '');
				
				$qbo_invoice_data["ShipAddr"] = array(
						'City' => $ship_city,
						'PostalCode' => $ship_zip_code,
						'CountrySubDivisionCode' => $ship_state,
						'Line1' => $client_member->getName(),
						'Line2' => '',
						'Line3' => $ship_street,
						//'Line4' => ($ship_city == '' ? '' : "$ship_city, ") . trim("$ship_state $ship_zip_code"),
				);
			}
		} else {
			$qbo_invoice_data["ShipAddr"] = $billing_address;
		}
		
		$shipping_method = null;
		$shipping_method_cp = CustomProperties::getCustomPropertyByCode($feng_invoice->getObjectTypeId(), "shipping_method");
		if ($shipping_method_cp) {
			$cp_val = CustomPropertyValues::getCustomPropertyValue($feng_invoice->getId(), $shipping_method_cp->getId());
			if ($cp_val) $shipping_method = trim($cp_val->getValue());
		}
		if ($shipping_method) {
			$qbo_invoice_data["ShipMethodRef"] = array("value" => $shipping_method);
		}
		
		$shipping_date = null;
		$shipping_date_cp = CustomProperties::getCustomPropertyByCode($feng_invoice->getObjectTypeId(), "shipping_date");
		if ($shipping_date_cp) {
			$cp_val = CustomPropertyValues::getCustomPropertyValue($feng_invoice->getId(), $shipping_date_cp->getId());
			if ($cp_val) $shipping_date = trim($cp_val->getValue());
		}
		if ($shipping_date) {
			$qbo_invoice_data["ShipDate"] = $shipping_date;
		}
		
		$shipping_track_num = null;
		$shipping_track_num_cp = CustomProperties::getCustomPropertyByCode($feng_invoice->getObjectTypeId(), "shipping_tracking_num");
		if ($shipping_track_num_cp) {
			$cp_val = CustomPropertyValues::getCustomPropertyValue($feng_invoice->getId(), $shipping_track_num_cp->getId());
			if ($cp_val) $shipping_track_num = trim($cp_val->getValue());
		}
		if ($shipping_track_num) {
			$qbo_invoice_data["TrackingNum"] = $shipping_track_num;
		}
	}
	
	
	
	
	// if editing an existing invoice set the values for sparse and synctoken attributes
	if ($qbo_invoice_id > 0) {
		// get the invoice's SyncToken, it is needed to update it
		$old_qbo_invoice_obj = get_qbo_invoice($dataService, $qbo_invoice_id);
		if ($old_qbo_invoice_obj) {
			$qbo_invoice_data['Id'] = $qbo_invoice_id;
			$qbo_invoice_data['SyncToken'] = $old_qbo_invoice_obj->SyncToken;
			$qbo_invoice_data['sparse'] = true; // only updates the sent information, does not override other data updated in qbo
		} else {
			$qbo_invoice_id = 0;
		}
	}
	
	
	if ($feng_invoice->getInvoiceNumber()) { // send invoice number only if is new
		$doc_number_exists = get_qbo_invoice_by_docnumber($dataService, $feng_invoice->getInvoiceNumber());
		if (!$doc_number_exists) {
			$qbo_invoice_data["DocNumber"] = $feng_invoice->getInvoiceNumber();
		}
	}
	// create the qbo invoice object to use in the api call
	$invoiced_timeslots = Timeslots::findAll(array("conditions" => array("invoice_id=?", $feng_invoice->getId())));
	if($invoiced_timeslots){
		$qbo_invoice_data['LinkedTxn'] = array();

		foreach($invoiced_timeslots as $timeslot){
			if ($timeslot->getColumnValue('quickbooks_online_id') > 0) {
				$timeslot_id_qb = $timeslot->getColumnValue('quickbooks_online_id');
			} else {
				$timeslot_id_qb = save_timeslot_to_qbo($timeslot, $dataService);
			}

			if($timeslot_id_qb && !in_array($timeslot->getColumnValue('quickbooks_online_id'), $times_already_included_in_line_linked_transaction)){
				$qbo_invoice_data['LinkedTxn'][] = array(
					"TxnId" => $timeslot_id_qb,
					"TxnType" => "TimeActivity"
				);

				$times_already_included_in_line_linked_transaction[] = $timeslot->getColumnValue('quickbooks_online_id');
			}
		}
	}

	// sync the related expenses before synchronizing the invoice
	if (Plugins::instance()->isActivePlugin('expenses2')) {
		$invoiced_expenses = PaymentReceipts::findAll(array("conditions" => array("invoicing_status = 'invoiced' AND invoice_id=?", $feng_invoice->getId())));
		foreach ($invoiced_expenses as $invoiced_expense) {
			// if not synchronized yet, then send the expense to quickbooks
			if ($invoiced_expense->getColumnValue('quickbooks_online_id') == 0 && in_array($invoiced_expense->getExpenseType(), config_option("expense_types_to_sync"))) {
				$invoiced_expense = save_actual_expense_to_qbo($invoiced_expense, $dataService);
			}

			// get the reimburse charge id (billabe expense charge)
			$billable_expense_id = $invoiced_expense->getColumnValue('qbo_reimburse_charge_id');
			if ($billable_expense_id > 0) {
				// ensure that reimburse charge exists
				$reimburse_charge_obj = get_reimburse_charge_object_by_id($dataService, $billable_expense_id);

				// if billable expense charge exists in qbo then send it as an associated object to the transaction
				if ($reimburse_charge_obj) {
					$qbo_invoice_data['LinkedTxn'][] = array(
						"TxnId" => $billable_expense_id,
						"TxnType" => "ReimburseCharge"
					);

					$bill_charges_already_included_in_line_linked_transaction[] = $billable_expense_id;
				}
			}
		}
	}

	$invoiceObj = QuickBooksOnline\API\Facades\Invoice::create($qbo_invoice_data);
	
	if (isset($qbo_invoice_data['Id'])) {
		// update in qbo
		$resultingInvoiceObj = $dataService->Update($invoiceObj);
	} else {
		// add to qbo
		$resultingInvoiceObj = $dataService->Add($invoiceObj);
	}
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error uploading invoice to Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	/* @var $resultingInvoiceObj QuickBooksOnline\API\Data\IPPInvoice */
	$invoiceId = $resultingInvoiceObj->Id;
	
	if ($qbo_invoice_id == 0) {
		$feng_invoice->setColumnValue('quickbooks_online_id', $invoiceId);
	}
	
	// put the same invoice number used in quickbooks, if different than the one in evx then alert the user
	if ($feng_invoice->getInvoiceNumber() != $resultingInvoiceObj->DocNumber) {
		$feng_invoice->setInvoiceNumber($resultingInvoiceObj->DocNumber);
		//flash_success("Invoice number had to be changed to '".$resultingInvoiceObj->DocNumber."' to keep compatibility with your Quickbooks Online account.");
	}
	
	$feng_invoice->setColumnValue('quickbooks_online_sync_token', $resultingInvoiceObj->SyncToken);
	$feng_invoice->dont_update_totals = true;
	$feng_invoice->save();
	return $feng_invoice;
}

/**
 * Queries an invoice in Quickbooks by id
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param integer $qbo_invoice_id
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPInvoice
 */
function get_qbo_invoice($dataService, $qbo_invoice_id) {
	
	$invoice_array = $dataService->Query("select * from Invoice where Id='" . $qbo_invoice_id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting invoice from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($invoice_array) && sizeof($invoice_array) > 0) {
			return current($invoice_array);
		}
	}
}
  
function get_qbo_employee($dataService, $qbo_employee_id) {
		
	$employee_array = $dataService->Query("select * from Employee where Id='" . $qbo_employee_id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting employee from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($employee_array) && sizeof($employee_array) > 0) {
			return current($employee_array);
		}
	}
}


function get_qbo_time_activity($dataService, $qbo_time_activity_id) {
		
	$time_activity_array = $dataService->Query("select * from TimeActivity where Id='" . $qbo_time_activity_id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting Time Activity from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($time_activity_array) && sizeof($time_activity_array) > 0) {
			return current($time_activity_array);
		}
	}
}



/**
 * Queries an invoice in Quickbooks by id
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param string $doc_number
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPInvoice
 */
function get_qbo_invoice_by_docnumber($dataService, $doc_number) {
	
	$invoice_array = $dataService->Query("select * from Invoice where DocNumber='" . $doc_number . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting invoice from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($invoice_array) && sizeof($invoice_array) > 0) {
			return current($invoice_array);
		}
	}
}


/************************************************************/
/* PAYMENTS */
/************************************************************/



/**
 * Deletes an invoice in quickbooks from a Feng's invoice
 * @param IncomeInvoice $payment
 * @throws Exception
 * @return string
 */
function delete_invoice_payment_from_qbo(InvoicePaymentReceipt $payment) {
	
	$dataService = get_qbo_data_service();
	
	$qbo_payment_data = array();
	$qbo_payment_id = $payment->getColumnValue("quickbooks_online_id");
	
	$old_qbo_invoice_obj = get_qbo_payment($dataService, $qbo_payment_id);
	if ($old_qbo_invoice_obj) {
		$qbo_payment_data['Id'] = $qbo_payment_id;
		$qbo_payment_data['SyncToken'] = $old_qbo_invoice_obj->SyncToken;
		$qbo_payment_data['sparse'] = true; // only updates the sent information, does not override other data updated in qbo
	} else {
		return;
	}
	
	$entity = QuickBooksOnline\API\Facades\Payment::create($qbo_payment_data);
	
	$return_obj = $dataService->Delete($entity);
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error deleting payment in Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	
	$payment->setColumnValue("quickbooks_online_id", 0);
	$payment->save();
	
	return $return_obj;
}

/**
 * Saves an invoice in quickbooks from a Feng's invoice
 * @param InvoicePaymentReceipt $feng_payment
 * @throws Exception
 * @return null
 */
function save_invoice_payment_to_qbo(InvoicePaymentReceipt $feng_payment) {

	if (!check_if_object_can_be_sync($feng_payment)) return;
	
	$dataService = get_qbo_data_service();
	
	$feng_invoice_members = $feng_payment->getMembers();
	$client_member = null;
	$project_member = null;
	$client_ot = ObjectTypes::findByName('customer');
	$project_ot = ObjectTypes::findByName('project');
	
	foreach ($feng_invoice_members as $member) {
		if ($member->getObjectTypeId() == $client_ot->getId()) {
			$client_member = $member;
		}
		if ($member->getObjectTypeId() == $project_ot->getId()) {
			$project_member = $member;
		}
	}
	
	//$qbo_preferences = get_qbo_preferences($dataService);
	
	$qb_lines = array();
	$affected_invoices = InvoicePaymentReceiptInvoices::instance()->getByPaymentReceiptId($feng_payment->getId());
	foreach ($affected_invoices as $affected_invoice) {
		/*  @var $affected_invoice InvoicePaymentReceiptInvoice */
		$inv = $affected_invoice->getInvoice();
		
		// if invoice not syncronized then create it in qbo
		if ($inv->getColumnValue('quickbooks_online_id') == 0) {
			$inv = save_invoice_to_qbo($inv, $dataService);
		}
			
		$qb_line = array(
				"Amount" => $affected_invoice->getAmount(),
				"LinkedTxn" => array(
						array(
								"TxnType" => "Invoice",
								"TxnId" => $inv->getColumnValue('quickbooks_online_id'),
						)
				),
		);
		
		$qb_lines[] = $qb_line;
	}
	
	// if invoice already exists in qbo, then use this id to update it
	$qbo_payment_id = $feng_payment->getColumnValue('quickbooks_online_id');
	
	$customerRef = get_qbo_customer_object($dataService, $client_member);
	
	$projectRef = null;
	if ($customerRef && $project_member instanceof Member) {
		$projectRef = get_qbo_project_object($dataService, $project_member, $customerRef);
	}
	
	$date_cp = CustomProperties::getCustomPropertyByCode($feng_payment->getObjectTypeId(), 'date');
	$date_cp_val = CustomPropertyValues::getCustomPropertyValue($feng_payment->getId(), $date_cp->getId());
	$pm_cp = CustomProperties::getCustomPropertyByCode($feng_payment->getObjectTypeId(), 'payment_method');
	$pm_cp_val = CustomPropertyValues::getCustomPropertyValue($feng_payment->getId(), $pm_cp->getId());
	$pref_cp = CustomProperties::getCustomPropertyByCode($feng_payment->getObjectTypeId(), 'external_transaction_id');
	$pref_cp_val = CustomPropertyValues::getCustomPropertyValue($feng_payment->getId(), $pref_cp->getId());
	$desc_cp = CustomProperties::getCustomPropertyByCode($feng_payment->getObjectTypeId(), 'description');
	$desc_cp_val = CustomPropertyValues::getCustomPropertyValue($feng_payment->getId(), $desc_cp->getId());
	
	$currency = Currencies::instance()->getCurrency($feng_payment->getCurrencyId());
	
	$qbo_payment_data = [
			"TotalAmt" => $feng_payment->getAmount(),
			"UnappliedAmt" => $feng_payment->getBalance(),
			"Line" => $qb_lines,
			"TxnDate" => $date_cp_val ? $date_cp_val->getValue() : '',
			"CustomerRef"=> [
					"value"=> $projectRef ? $projectRef->Id : $customerRef->Id
			],
			"CurrencyRef" => [
					"value" => $currency->getShortName()
			]
	];
	
	$qbo_account = get_qbo_account_by_name($dataService, "Undeposited Funds");
	if ($qbo_account) {
		$qbo_payment_data["DepositToAccountRef"] = [
				"value" => $qbo_account->Id,
				"name" => $qbo_account->Name,
		];
	}
	
	if ($pm_cp_val) {
		$pm_string = $pm_cp_val->getValue();
		if (Localization::instance()->lang_exists($pm_string)) {
			$pm_string = lang($pm_string);
		}
		$qbo_payment_method = get_qbo_payment_method($dataService, $pm_string); // Cash, Check. MO, Credit card, Other
		if ($qbo_payment_method) {
			$qbo_payment_data["PaymentMethodRef"] = [
					"value" => $qbo_payment_method->Id
			];
		}
	}
	if ($pref_cp_val && trim($pref_cp_val->getValue()) != '') {
		$qbo_payment_data["PaymentRefNum"] = trim($pref_cp_val->getValue());
	}
	if ($desc_cp_val && trim($desc_cp_val->getValue()) != '') {
		$qbo_payment_data["PrivateNote"] = trim($desc_cp_val->getValue());
	}
	
	
	
	// if editing an existing invoice set the values for sparse and synctoken attributes
	if ($qbo_payment_id > 0) {
		// get the invoice's SyncToken, it is needed to update it
		$old_qbo_payment_obj = get_qbo_payment($dataService, $qbo_payment_id);
		if ($old_qbo_payment_obj) {
			$qbo_payment_data['Id'] = $qbo_payment_id;
			$qbo_payment_data['SyncToken'] = $old_qbo_payment_obj->SyncToken;
			$qbo_payment_data['sparse'] = true; // only updates the sent information, does not override other data updated in qbo
		} else {
			$qbo_payment_id = 0;
		}
	}
	
	// create the qbo invoice object to use in the api call
	$paymentObj = QuickBooksOnline\API\Facades\Payment::create($qbo_payment_data);
	
	if (isset($qbo_payment_data['Id'])) {
		// update in qbo
		$resultingPaymentObj = $dataService->Update($paymentObj);
	} else {
		// add to qbo
		$resultingPaymentObj = $dataService->Add($paymentObj);
	}
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error uploading payment to Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}
	
	$paymentId = $resultingPaymentObj->Id;
	
	if ($qbo_payment_id == 0) {
		$feng_payment->setColumnValue('quickbooks_online_id', $paymentId);
	}
	
	$feng_payment->setColumnValue('quickbooks_online_sync_token', $resultingPaymentObj->SyncToken);
	$feng_payment->save();
	
}

/**
 * Queries an invoice in Quickbooks by id
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param integer $qbo_payment_id
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPPayment
 */
function get_qbo_payment($dataService, $qbo_payment_id) {
	
	$invoice_array = $dataService->Query("select * from Payment where Id='" . $qbo_payment_id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting payment from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($invoice_array) && sizeof($invoice_array) > 0) {
			return current($invoice_array);
		}
	}
}

/**
 * 
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param string $methodName
 * @param int $methodId
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPPaymentMethod
 */
function get_qbo_payment_method($dataService, $methodName, $methodId=null) {
	
	if (!$methodId) {
		$method_array = $dataService->Query("select * from PaymentMethod where Name='" . $methodName . "'");
	} else {
		// first check if we have built the cache
		if (isset($_SESSION['qbo_entities_cache'])) {
			$entity = array_var($_SESSION['qbo_entities_cache']['payment_methods'], $methodId);
			if ($entity) return $entity;
		}
		$method_array = $dataService->Query("select * from PaymentMethod where Id='" . $methodId . "'");
	}
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting payment method from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($method_array) && sizeof($method_array) > 0) {
			return current($method_array);
		}
	}
}

function get_qbo_all_payment_methods($dataService) {
	
	$method_array = $dataService->Query("select * from PaymentMethod");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting payment methods from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		return $method_array;
	}
}


function get_qbo_payment_term($dataService, $id) {
	
	// first check if we have built the cache
	if (isset($_SESSION['qbo_entities_cache'])) {
		$entity = array_var($_SESSION['qbo_entities_cache']['terms'], $id);
		if ($entity) return $entity;
	}
	
	$method_array = $dataService->Query("select * from Term where Id='" . $id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting term from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($method_array) && sizeof($method_array) > 0) {
			return current($method_array);
		}
	}
}


function get_qbo_payment_term_by_name($dataService, $term) {
	
	$method_array = $dataService->Query("select * from Term where Name='" . $term . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting term from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($method_array) && sizeof($method_array) > 0) {
			return current($method_array);
		}
	}
	
	$dd = 30;
	if (strtolower($term) == 'net d') {
		$dd = 1;
	}
	// create the term
	$term_request_obj = QuickBooksOnline\API\Facades\Term::create([
			"Name" => escape_character($term),
			"DueDays" => $dd
	]);
	$term_response_obj = $dataService->Add($term_request_obj);
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error adding payment term to Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		return $term_response_obj;
	}
}




/**
 * Queries a customer in Quickbooks by name, if it doesn't exist it creates it and returns the resulting object
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param Member $client_member
 * @throws Exception
 * @return QuickBooksOnline\API\Facades\Customer
 */
function get_qbo_customer_object($dataService, Member $client_member) {
	$customerObj = null;

	if ($client_member->getColumnValue('quickbooks_online_id') > 0) {
		// if we already have the id of the matching project in QB then make the query by id
		$customerObj = get_qbo_customer_object_by_id($dataService, $client_member->getColumnValue('quickbooks_online_id'));
	}

	if ($customerObj) {

		return $customerObj;
	
	} else {
			
		// search client by CompanyName
		$customerName = $client_member->getName();
		$customerName = qbo_clean_client_or_project_name($customerName);
		$customerArray = $dataService->Query("select * from Customer where CompanyName='" . escape_character($customerName) . "'");
		$error = $dataService->getLastError();
		if ($error) {
			Logger::log_r("Error getting customer from Quickbooks: '" . escape_character($customerName) . "'");
			Logger::log_r($error);
			throw new Exception(build_qbo_error_message($error));
		} else {
			if (is_array($customerArray) && sizeof($customerArray) > 0) {
				$customerObj = current($customerArray);
			}
		}

		// if client was not found using CompnayName, try using DisplayName
		if (!$customerObj) {
			$customerArray = $dataService->Query("select * from Customer where DisplayName='" . escape_character($customerName) . "'");
			$error = $dataService->getLastError();
			if ($error) {
				Logger::log_r("Error getting customer from Quickbooks: '" . escape_character($customerName) . "'");
				Logger::log_r($error);
				throw new Exception(build_qbo_error_message($error));
			} else {
				if (is_array($customerArray) && sizeof($customerArray) > 0) {
					$customerObj = current($customerArray);
				}
			}
		}

		// if customer object found then associate its id to the client member for future queries
		if ($customerObj) {
			// save the qbo id in our member
			$client_member->setColumnValue('quickbooks_online_id', $customerObj->Id);
			$client_member->setColumnValue('quickbooks_online_sync_token', $customerObj->SyncToken);
			$client_member->save();

			return $customerObj;
		} else {
			// debug
			//$customerArray = $dataService->Query("select * from Customer MAXRESULTS 1000");
			//file_put_contents(CACHE_DIR."/all_qbo_customers", print_r($customerArray,1));
		}

		// if customer object not found in qbo then create the customer object		
		$customer = Customers::findById($client_member->getObjectId());
		$customer_contact = $customer ? $customer->getContact() : null;
		
		// get billing address
		$customer_address = get_client_billing_address($client_member);

		// create object to send in the request
		$customerRequestObj = QuickBooksOnline\API\Facades\Customer::create([				
				"BillAddr" => [
					"Line1"=>  $customer_address ? $customer_address->getStreet() : '',
					"City"=>  $customer_address ? $customer_address->getCity() : '',
					"Country"=>  $customer_address ? $customer_address->getCountry() : '',
					"CountrySubDivisionCode"=>  $customer_address ? $customer_address->getState() : '',
					"PostalCode"=> $customer_address ? $customer_address->getZipCode() : ''
				],				
				"DisplayName"=>  escape_character($customerName),
				"PrimaryPhone"=>  [
					"FreeFormNumber"=> $customer_contact ? $customer_contact->getPhoneNumber('work') : ''
				],
				"PrimaryEmailAddr"=>  [
					"Address" => $customer_contact ? $customer_contact->getEmailAddress() : ''
				]
				
		]);
		$customerResponseObj = $dataService->Add($customerRequestObj);
		$error = $dataService->getLastError();
		if ($error) {
			Logger::log_r("Error adding customer to Quickbooks: ".escape_character($customerName));
			Logger::log_r($error);
			throw new Exception(build_qbo_error_message($error));
		} else {

			// save the qbo id in our member
			$client_member->setColumnValue('quickbooks_online_id', $customerResponseObj->Id);
			$client_member->setColumnValue('quickbooks_online_sync_token', $customerResponseObj->SyncToken);
			$client_member->save();

			return $customerResponseObj;
		}
	}
}

/**
 * Queries a customer in Quickbooks by name, if it doesn't exist it creates it and returns the resulting object
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param int $id
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPCustomer
 */
function get_qbo_customer_object_by_id($dataService, $id) {
	
	// first check if we have built the cache
	if (isset($_SESSION['qbo_entities_cache'])) {
		$entity = array_var($_SESSION['qbo_entities_cache']['clients'], $id);
		if ($entity) return $entity;
	}

	$customerArray = $dataService->Query("select * from Customer where Id='" . $id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting customer from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($customerArray) && sizeof($customerArray) > 0) {
			$client_obj = current($customerArray);
			if (isset($_SESSION['qbo_entities_cache'])) {
				$_SESSION['qbo_entities_cache']['clients'][$id] = $client_obj;
			}
			return $client_obj;
		}
	}

	// if not found then query the api using active=false, to see if the client has been deleted and still exists
	$customerArray = $dataService->Query("select * from Customer where Id='" . $id . "' and active=false");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting customer from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($customerArray) && sizeof($customerArray) > 0) {
			$client_obj = current($customerArray);
			if (isset($_SESSION['qbo_entities_cache'])) {
				$_SESSION['qbo_entities_cache']['clients'][$id] = $client_obj;
			}
			return $client_obj;
		}
	}
}




/**
 * Queries a customer in Quickbooks by name, if it doesn't exist it creates it and returns the resulting object
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param Member $client_member
 * @throws Exception
 * @return QuickBooksOnline\API\Facades\Vendor
 */
function get_qbo_vendor_object($dataService, $client_member = null, $employee = null) {
	$customerObj = null;

	// if we already have the id of the matching project in QB then make the query by id
	if ($client_member instanceof Member && $client_member->getColumnValue('quickbooks_online_id') > 0) {
		$customerObj = get_qbo_vendor_object_by_id($dataService, $client_member->getColumnValue('quickbooks_online_id'));

	} else if ($employee instanceof Contact && $employee->getColumnValue('quickbooks_online_vendor_id') > 0) {
		$customerObj = get_qbo_vendor_object_by_id($dataService, $employee->getColumnValue('quickbooks_online_vendor_id'));
	}

	if ($customerObj) {

		return $customerObj;
	
	} else {
			
		// search client by CompanyName
		if ($client_member instanceof Member) {
			$customerName = $client_member->getName();
		} else if ($employee instanceof Contact) {
			$customerName = $employee->getObjectName();
		} else {
			return null;
		}

		$customerName = qbo_clean_client_or_project_name($customerName);
		
		$customerArray = $dataService->Query("select * from Vendor where DisplayName='" . escape_character($customerName) . "'");
		$error = $dataService->getLastError();
		if ($error) {
			Logger::log_r("Error getting vendor from Quickbooks: '" . escape_character($customerName) . "'");
			Logger::log_r($error);
			throw new Exception(build_qbo_error_message($error));
		} else {
			if (is_array($customerArray) && sizeof($customerArray) > 0) {
				$customerObj = current($customerArray);
			}
		}

		// if it is for an employee and no vendor found with its name, try using the suffix "(Vendor)"
		if (!$customerObj && $employee instanceof Contact) {
			$customerName .= " (Vendor)";

			$customerArray = $dataService->Query("select * from Vendor where DisplayName='" . escape_character($customerName) . "'");
			$error = $dataService->getLastError();
			if ($error) {
				Logger::log_r("Error getting vendor from Quickbooks: '" . escape_character($customerName) . "'");
				Logger::log_r($error);
				throw new Exception(build_qbo_error_message($error));
			} else {
				if (is_array($customerArray) && sizeof($customerArray) > 0) {
					$customerObj = current($customerArray);
				}
			}
		}
		

		// if customer object found then associate its id to the client member for future queries
		if ($customerObj) {
			// save the qbo id in our member
			if ($client_member instanceof Member) {
				$client_member->setColumnValue('quickbooks_online_id', $customerObj->Id);
				$client_member->setColumnValue('quickbooks_online_sync_token', $customerObj->SyncToken);
				$client_member->save();
			} else if ($employee instanceof Contact) {
				$employee->setColumnValue('quickbooks_online_vendor_id', $customerObj->Id);
				$employee->save();
			}

			return $customerObj;
		}

		// if customer object not found in qbo then create the customer object
		if ($client_member instanceof Member) {
			$customer = Customers::findById($client_member->getObjectId());
			$customer_contact = $customer ? $customer->getContact() : null;
		} else if ($employee instanceof Contact) {
			$customer_contact = $employee;
		}
		$customer_address = null;
		if ($customer_contact instanceof Contact) {
			$customer_address = $customer_contact->getAddress('work');
		}
		
		$vendorRequestObj = QuickBooksOnline\API\Facades\Vendor::create([				
				"BillAddr" => [
					"Line1"=>  $customer_address ? $customer_address->getStreet() : '',
					"City"=>  $customer_address ? $customer_address->getCity() : '',
					"Country"=>  $customer_address ? $customer_address->getCountry() : '',
					"CountrySubDivisionCode"=>  $customer_address ? $customer_address->getState() : '',
					"PostalCode"=> $customer_address ? $customer_address->getZipCode() : ''
				],
				"CompanyName"=>  escape_character($customerName),
				"DisplayName"=>  escape_character($customerName),
				"PrimaryPhone"=>  [
					"FreeFormNumber"=> $customer_contact ? $customer_contact->getPhoneNumber('work') : ''
				],
				"PrimaryEmailAddr"=>  [
					"Address" => $customer_contact ? $customer_contact->getEmailAddress() : ''
				]
				
		]);
		$vendorResponseObj = $dataService->Add($vendorRequestObj);
		$error = $dataService->getLastError();
		if ($error) {
			Logger::log_r("Error adding customer to Quickbooks: ".escape_character($customerName));
			Logger::log_r($error);
			throw new Exception(build_qbo_error_message($error));
		} else {

			if ($client_member instanceof Member) {
				// save the qbo id in our member
				$client_member->setColumnValue('quickbooks_online_id', $vendorResponseObj->Id);
				$client_member->setColumnValue('quickbooks_online_sync_token', $vendorResponseObj->SyncToken);
				$client_member->save();
			} else if ($employee instanceof Contact) {
				//Logger::log_r("guardando id en contact: ".$vendorResponseObj->Id);
				$employee->setColumnValue('quickbooks_online_vendor_id', $vendorResponseObj->Id);
				$employee->save();
			}

			return $vendorResponseObj;
		}
	}
}

/**
 * Queries a Vendor in Quickbooks by name, if it doesn't exist it creates it and returns the resulting object
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param int $id
 * @throws Exception
 * @return QuickBooksOnline\API\Data\IPPVendor
 */
function get_qbo_vendor_object_by_id($dataService, $id) {
	
	// first check if we have built the cache
	if (isset($_SESSION['qbo_entities_cache'])) {
		$entity = array_var($_SESSION['qbo_entities_cache']['vendors'], $id);
		if ($entity) return $entity;
	}

	$vendorArray = $dataService->Query("select * from Vendor where Id='" . $id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting Vendor from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($vendorArray) && sizeof($vendorArray) > 0) {
			$vendor_obj = current($vendorArray);
			if (isset($_SESSION['qbo_entities_cache'])) {
				$_SESSION['qbo_entities_cache']['vendors'][$id] = $vendor_obj;
			}
			return $vendor_obj;
		}
	}
}



/**
 * Queries a project in Quickbooks by id, project number or name, and ensures that it belongs to the client $customerRef, if it doesn't exist it creates it and returns the resulting object
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param Member $project_member
 * @param QuickBooksOnline\API\Facades\Customer $customerRef
 * @throws Exception
 * @return QuickBooksOnline\API\Facades\Customer
 */
function get_qbo_project_object($dataService, Member $param_project_member, $customerRef) {	

	$project_member = $param_project_member;

	// if option to use only parent project then iterate the project hierachy to get the parent project.
	$use_only_parent_project = config_option("qbo_sync_only_in_first_level_projects");
	if ($use_only_parent_project) {

		$parent_member = $param_project_member->getParentMember();
		while ($parent_member instanceof Member) {
			if ($parent_member->getObjectTypeId() == $project_member->getObjectTypeId()) {
				$project_member = $parent_member;
			}
			$parent_member = $parent_member->getParentMember();
		}
	}
	
	if ($project_member->getColumnValue('quickbooks_online_id') > 0) {
		// if we already have the id of the matching project in QB then make the query by id
		
		return get_qbo_customer_object_by_id($dataService, $project_member->getColumnValue('quickbooks_online_id'));

	} else {
		$projectRef = null;
		$project_number = null;

		// search the project object by project number or name
		$project_id_cp = CustomProperties::getCustomPropertyByCode($project_member->getObjectTypeId(), 'project_id');
		if ($project_id_cp) {

			$cp_val = CustomPropertyValues::getCustomPropertyValue($project_member->getObjectId(), $project_id_cp->getId());
			if ($cp_val instanceof CustomPropertyValue) {
				$project_number = trim($cp_val->getValue());
			}
			// try to query the project whose name starts with the project number
			if ($project_number) {
				$project_number = str_replace(':','',$project_number);
				$project_number = qbo_clean_client_or_project_name($project_number);
				$query = "select * from Customer where DisplayName LIKE '" . escape_character($project_number) . "%'";
				
				$projectArray = $dataService->Query($query);
				$error = $dataService->getLastError();
				if ($error) {
					Logger::log_r("Error getting project from Quickbooks");
					Logger::log_r($error);
					throw new Exception(build_qbo_error_message($error));
				} else {
					if (is_array($projectArray) && sizeof($projectArray) > 0) {
						$projectRef = current($projectArray);
					}
				}
			}

		}

		// if not found by qbo id or project number, query the project by name and verify that it belongs to the right client
		if (!$projectRef) {

			$projectName = $project_member->getName();
			$projectName = str_replace(':','',$projectName);
			$projectName = qbo_clean_client_or_project_name($projectName);
			
			$query = "select * from Customer where Job=true AND DisplayName='" . escape_character($projectName) . "'";
			$projectArray = $dataService->Query($query);
			
			
			$error = $dataService->getLastError();
			if ($error) {
				Logger::log_r("Error getting project from Quickbooks");
				Logger::log_r($error);
				throw new Exception(build_qbo_error_message($error));
			} else {
				if (is_array($projectArray) && sizeof($projectArray) > 0) {
					// verify that it belongs to the right client
					foreach ($projectArray as $projectObj) {
						if ($projectObj->ParentRef == $customerRef->Id) {
							$projectRef = $projectObj;
						}
					}
					
				}
			}
		}


		if ($projectRef) {
			// save the qbo id in our member
			$project_member->setColumnValue('quickbooks_online_id', $projectRef->Id);
			$project_member->setColumnValue('quickbooks_online_sync_token', $projectRef->SyncToken);
			$project_member->save();

			return $projectRef;

		} else {

			// call to the project creation, if it is a child of another project then it creates the whole project hierarchy
			return create_qbo_project_object($dataService, $project_member, $customerRef);

		}
		
	}

}

/**
 * Creates a project in quickbooks and returns the resulting object, if it is a subproject then it queries/creates the parent projects hierarchy to replicate the same structure
 * @param QuickBooksOnline\API\DataService\DataService $dataService
 * @param Member $project_member
 * @param QuickBooksOnline\API\Facades\Customer $customerRef
 * @throws Exception
 * @return QuickBooksOnline\API\Facades\Customer
 */
function create_qbo_project_object($dataService, Member $project_member, $customerRef) {	

	$project_ot_id = $project_member->getObjectTypeId();
	$parent_member = $project_member->getParentMember();
	$qbo_parent_object = null;

	if ($parent_member instanceof Member && $parent_member->getObjectTypeId() == $project_ot_id) {
		$qbo_parent_object = get_qbo_project_object($dataService, $parent_member, $customerRef);
	}

	if ($qbo_parent_object == null) {
		$qbo_parent_object = $customerRef;
	}


	// build the name to create at quickbooks, add the project number before the name if available
	$projectName = $project_member->getName();
	$projectName = qbo_clean_client_or_project_name($projectName);

	$project_number = null;
	$project_id_cp = CustomProperties::getCustomPropertyByCode($project_member->getObjectTypeId(), 'project_id');
	if ($project_id_cp) {
		$cp_val = CustomPropertyValues::getCustomPropertyValue($project_member->getObjectId(), $project_id_cp->getId());
		if ($cp_val instanceof CustomPropertyValue) {
			$project_number = trim($cp_val->getValue());
		}
	}
	if ($project_number) {
		$projectName = "$project_number - $projectName";
	}
	$projectName = str_replace(':','',$projectName);

	// Create Project in Quickbooks online						
	$cli_ot = ObjectTypes::findByName('customer');						
	$clients_dim = Dimensions::findByCode("customers");
	$member = Members::findById($project_member->getId());

	$customer_member_id = get_associated_status_member_id($member, $clients_dim, $cli_ot, false, 'project_customer');	
	$customer_member = Members::getMemberById($customer_member_id);
	if ($customer_member) {
		$customer = Customers::findById($customer_member->getObjectId());
	}
	if ($customer) {
		$customer_contact = $customer ? $customer->getContact() : null;
	}
	
	$customer_address = get_client_billing_address($customer_member);

	$projectRequestObj = QuickBooksOnline\API\Facades\Customer::create([

		"BillAddr" => [
			"Line1"=>  $customer_address ? $customer_address->getStreet() : '',
			"City"=>  $customer_address ? $customer_address->getCity() : '',
			"Country"=>  $customer_address ? $customer_address->getCountry() : '',
			"CountrySubDivisionCode"=>  $customer_address ? $customer_address->getState() : '',
			"PostalCode"=> $customer_address ? $customer_address->getZipCode() : ''
		],				
		"DisplayName" => escape_character($projectName),
			// specify that this is a job and not a subclient
			"Job" => true, 
			// link to the client
			"ParentRef" => array(
					"value" => $qbo_parent_object->Id,
					"name" => $qbo_parent_object->DisplayName,
			),
			// tell quickbooks to use parent client's data to bill
			"BillWithParent" => true,
		"PrimaryPhone"=>  [
			"FreeFormNumber"=> $customer_contact ? $customer_contact->getPhoneNumber('work') : ''
		],
		"PrimaryEmailAddr"=>  [
			"Address" => $customer_contact ? $customer_contact->getEmailAddress() : ''
		]
			
	]);

	$projectResponseObj = $dataService->Add($projectRequestObj);
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error adding project to Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {

		// save the qbo id in our member
		$project_member->setColumnValue('quickbooks_online_id', $projectResponseObj->Id);
		$project_member->setColumnValue('quickbooks_online_sync_token', $projectResponseObj->SyncToken);
		$project_member->save();

		return $projectResponseObj;
	}

}

/**
 * gets the billing address of a client member
 * first try to find billing address cp, if not exists or is empty then use the customer normal address
 */
function get_client_billing_address($customer_member) {
	if (!$customer_member) return;
	$customer_address = null;

	$bill_addr_cp = CustomProperties::getCustomPropertyByCode($customer_member->getObjectTypeId(), "billing_address");
	if ($bill_addr_cp instanceof CustomProperty) {
		$cp_val = CustomPropertyValues::getCustomPropertyValue($customer_member->getObjectId(), $bill_addr_cp->getId());
		if ($cp_val) {
			$address_string = trim($cp_val->getValue());
			if ($address_string != '') {

				$exploded = explode("|", escape_character($address_string));
				if (count($exploded) > 0) {
					$street = array_var($exploded, 1, '');
					$city = array_var($exploded, 2, '');
					$state = array_var($exploded, 3, '');
					$country = array_var($exploded, 4, '');
					$zip_code = array_var($exploded, 5, '');
					
					$customer_address = new ContactAddress();
					$customer_address->setStreet($street);
					$customer_address->setCity($city);
					$customer_address->setState($state);
					$customer_address->setCountry($country);
					$customer_address->setZipCode($zip_code);
				}
			}
		}
	}

	if (!$customer_address) {
		$customer = Customers::findById($customer_member->getObjectId());
		$customer_contact = $customer ? $customer->getContact() : null;
		if ($customer_contact instanceof Contact) {
			$customer_address = $customer_contact->getAddress('work');
			// if we didnt find work address, try with other types
			if (!$customer_address) $customer_address = $customer_contact->getAddress('home');
			if (!$customer_address) $customer_address = $customer_contact->getAddress('postal');
			if (!$customer_address) $customer_address = $customer_contact->getAddress('other');
		}
	}

	return $customer_address;
}


/**
 * @deprecated : before creating items we need to ensure that QBO version has "inventory" module enabled or if it is a "plus edition" 
 * Find if an Item is present , if not create new Item
 */
function get_qbo_invoice_item_object($dataService, IncomeInvoiceLine $invoice_line) {
	
	$itemName = $invoice_line->getDetail();
	$itemArray = $dataService->Query("select * from Item WHERE Name='" . escape_character($itemName) . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting item from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($itemArray) && sizeof($itemArray) > 0) {
			return current($itemArray);
		}
	}
	
	// Fetch IncomeAccount, ExoenseAccount and AssetAccount Refs needed to create an Item
	$incomeAccount = getIncomeAccountObj($dataService);
	$expenseAccount = getExpenseAccountObj($dataService);
	$assetAccount = getAssetAccountObj($dataService);
	
	// Create Item
	$dateTime = new \DateTime('NOW');
	$ItemObj = QuickBooksOnline\API\Facades\Item::create([
			"Name" => $itemName,
			"Description" => $itemName,
			"Active" => true,
			"FullyQualifiedName" => $itemName,
			"Taxable" => true,
			"UnitPrice" => $invoice_line->getUnitValue(),
			"Type" => "Inventory",
			"IncomeAccountRef"=> [
					"value"=>  $incomeAccount->Id
			],
			"PurchaseDesc"=> "",
			"PurchaseCost"=> 1,
			"ExpenseAccountRef"=> [
					"value"=> $expenseAccount->Id
			],
			"AssetAccountRef"=> [
					"value"=> $assetAccount->Id
			],
			"TrackQtyOnHand" => true,
			"QtyOnHand"=> 99999,
			"InvStartDate"=> $dateTime
	]);
	$resultingItemObj = $dataService->Add($ItemObj);
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error adding item to Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		return $resultingItemObj;
	}
	
}


/***************************************************************************************/
/******* TEMPORARY FUNCTIONS TO QUERY/CREATE THE NECESSARY ACCOUNTS FOR THE ITEMS ******/
/***************************************************************************************/

const QBO_INCOME_ACCOUNT_TYPE = "Income";
const QBO_INCOME_ACCOUNT_SUBTYPE = "SalesOfProductIncome";
const QBO_EXPENSE_ACCOUNT_TYPE = "Cost of Goods Sold";
const QBO_EXPENSE_ACCOUNT_SUBTYPE = "SuppliesMaterialsCogs";
const QBO_ASSET_ACCOUNT_TYPE = "Other Current Asset";
const QBO_ASSET_ACCOUNT_SUBTYPE = "Inventory";

/*
 Find if an account of Income type exists, if not, create one
 */
function getIncomeAccountObj($dataService) {
	return get_qbo_account_by_type($dataService, QBO_INCOME_ACCOUNT_TYPE, QBO_INCOME_ACCOUNT_SUBTYPE);
}
function getExpenseAccountObj($dataService) {
	return get_qbo_account_by_type($dataService, QBO_EXPENSE_ACCOUNT_TYPE, QBO_EXPENSE_ACCOUNT_SUBTYPE);
}
function getAssetAccountObj($dataService) {
	return get_qbo_account_by_type($dataService, QBO_ASSET_ACCOUNT_TYPE, QBO_ASSET_ACCOUNT_SUBTYPE);
}

/*
 Sync with qb time / tsheets functions
 */


$access_token = 'S.13__c6a1e7a713d4c6118dd754979abb0b1bca238bd1';
$tsheets_client = new TSheetsRestClient(1, $access_token);

//feng contact / tsheets user
function sync_contact_with_qb_time($contact_id, $tsheets_client = null)
{

	if (is_null($tsheets_client)) {
		$tsheets_client = new TSheetsRestClient(1, $access_token = 'S.13__c6a1e7a713d4c6118dd754979abb0b1bca238bd1');
	}

	
	//contact(feng)-employee(qbo)
	$feng_contact_id = $contact_id;
	$feng_contact = Contacts::findById($feng_contact_id);


	$qb_time_user_id = $feng_contact->getColumnValue('quickbooks_time_id');
	
	$result = null;
	// if contact qb_time_user_id is in database


	$request = array();
	$request[] = array(
		'username' => $feng_contact->getEmailAddress(),
		'first_name' => $feng_contact->getFirstName(),
		'last_name' => $feng_contact->getSurname(),
		"email" => $feng_contact->getEmailAddress()           
	);
	


	if ($qb_time_user_id > 0) {

        $user = $tsheets_client->get(ObjectType::Users, ['ids' => $qb_time_user_id]);
		$request = array();
		$request[] = array(
			'username' => $feng_contact->getEmailAddress(),
			'first_name' => $feng_contact->getFirstName(),
			'last_name' => $feng_contact->getSurname(),
			"email" => $feng_contact->getEmailAddress()           
		);
		
        $result = $tsheets_client->edit(ObjectType::Users, $request);

	} else {
		
		$result = $tsheets_client->add(ObjectType::Users, $request);		
        //save contact qb_time_user_id in feng contact
        $feng_contact->setColumnValue('quickbooks_time_id', $result['results']['users'][1]['id']);
        $feng_contact->save();
        
	}

	return $result;
}

// feng timeslots /qb time time-entries
function save_timeslot_to_qb_time(Timeslot $timeslot, $tsheets_client = null) {

	if (!check_if_object_can_be_sync($timeslot)) return;
	
    if (is_null($tsheets_client)) {
		$tsheets_client = new TSheetsRestClient(1, $access_token = 'S.13__c6a1e7a713d4c6118dd754979abb0b1bca238bd1');
	}

	$qbo_time_id = $timeslot->getColumnValue('quickbooks_time_id');
	

	$tz_offset = $timeslot->getTimezoneValue();

	$tz_hours = str_pad(floor(abs($tz_offset) / 3600), 2, '0', STR_PAD_LEFT);

	$tz_mins = str_pad(abs($tz_offset) % 60, 2, '0', STR_PAD_LEFT);

	$txn_date = new DateTimeValue($timeslot->getStartTime()->getTimestamp());

	$txn_date->advance($tz_offset);

	$txn_date = $txn_date->format('Y-m-d');

	$start_time = new DateTimeValue($timeslot->getStartTime()->getTimestamp());

	$start_time->add('s', $tz_offset);

	$end_time = new DateTimeValue($timeslot->getEndTime()->getTimestamp());

	$end_time->add('s', $tz_offset);

	$start_time_str = $start_time->toMySQL();

	$start_time_str = str_replace(' ', 'T', $start_time_str);

	$end_time_str = $end_time->toMySQL();

	$end_time_str = str_replace(' ', 'T', $end_time_str);



	$feng_timeslot_members = $timeslot->getMembers();

	$client_member = null;

	$project_member = null;

	$client_ot = ObjectTypes::findByName('customer');

	$project_ot = ObjectTypes::findByName('project');

	
	foreach ($feng_timeslot_members as $member) {

		if ($member->getObjectTypeId() == $client_ot->getId()) {
			$client_member = $member;
		}

		if ($member->getObjectTypeId() == $project_ot->getId()) {
			$project_member = $member;
		}

	}
	
 	if($client_member) {		
		//parent
		
	}

	
	//si tiene: proyecto, si no: cliente

	if ($project_member instanceof Member) {
		//jobcode		
		
	} 
	
	$jobcode_billable = ($timeslot->getColumnValue('invoicing_status') != 'non_billable') ? 'yes' : 'no';
	
	$jobcode_request = array();
	$jobcode_request[] = array(
		"name" => $project_member->getName(),		
		//"short_code" => "c1",
		"billable" => $jobcode_billable,
		//"assigned_to_all" => "yes",
		//"billable_rate" => "37.50",
		//"parent_id" => client
		//"connect_with_quickbooks" => true     
	);

	
	$jobcode_result = $tsheets_client->add(ObjectType::Jobcodes, $jobcode_request);
	

	$user = sync_contact_with_qb_time($timeslot->getContactId());

	$user_id = $user['results']['users'][1]['id'];

	if($jobcode_result) {
		$jobcode_id = $jobcode_result['results']['jobcodes'][1]['id'];
	}
	

	$time_entry_request = array();
	$time_entry_request[] = array(
		'user_id' => $user_id,
	 	'jobcode_id' => $jobcode_id ? $jobcode_id : 0 ,		 
		'type' => 'manual',
		"date" => $txn_date,
		"duration" => "15552",
		'start' => $start_time_str,
		'end' => $end_time_str, 
		'customfields' => [
		"1302158" => "Hours"        
		]
		
	);

	$time_entry_result = $tsheets_client->add(ObjectType::Timesheets, $time_entry_request);


}

function save_actual_expense_to_qbo(PaymentReceipt $feng_payment, $dataService=null) {

	if (!check_if_object_can_be_sync($feng_payment)) return;

	if (is_null($dataService)) {
		$dataService = get_qbo_data_service();
	}

	if ($feng_payment->isNew() && !in_array($feng_payment->getExpenseType(), config_option("expense_types_to_sync"))){
		return;
	}else if (!in_array($feng_payment->getExpenseType(), config_option("expense_types_to_sync"))){
		delete_actual_expense_from_qbo($feng_payment);
		return;
	}

	$feng_payment_members = $feng_payment->getMembers();
	$qbo_bill_id = $feng_payment->getColumnValue('quickbooks_online_id');
	$client_member = null;
	$project_member = null;
	$supplier_member = null;
	$client_ot = ObjectTypes::findByName('customer');
	$project_ot = ObjectTypes::findByName('project');
	$suppliers_dim = Dimensions::findByCode('suppliers');
	$payment_receipt_obj = ObjectTypes::findByName('payment_receipt');

	
	foreach ($feng_payment_members as $member) {
		if ($member->getObjectTypeId() == $client_ot->getId()) {
			$client_member = $member;
		}
		if ($member->getObjectTypeId() == $project_ot->getId()) {
			$project_member = $member;
		}
		if ($suppliers_dim && $member->getDimensionId() == $suppliers_dim->getId()) {
			$supplier_member = $member;
		}
	}

	$customerRef = get_qbo_customer_object($dataService, $client_member);
	
	$projectRef = null;
	if ($customerRef && $project_member instanceof Member) {
		$projectRef = get_qbo_project_object($dataService, $project_member, $customerRef);
	}

	// Reference to quickbooks customer
	$entity_ref = null;
	if ($projectRef) {
		$entity_ref = $projectRef;
	} else if ($customerRef) {
		$entity_ref = $customerRef;
	}
	// --

	// get the item class
	$qbo_preferences = get_qbo_preferences($dataService);
	$use_class_in_whole_invoice = $qbo_preferences->AccountingInfoPrefs->ClassTrackingPerTxn;
	$use_class_in_invoice_lines = $qbo_preferences->AccountingInfoPrefs->ClassTrackingPerTxnLine;

	$class_definition = null;
	if ($use_class_in_invoice_lines || $use_class_in_whole_invoice) {
		$class_definition = get_qbo_class_for_feng_object($dataService, $feng_payment);
	}

	$bill_no = 0;
	$payment_bill_no = CustomProperties::getCustomPropertyByCode($payment_receipt_obj->getId(), "payment_receipt_bill_no");
	if ($payment_bill_no instanceof CustomProperty) {
		$cpval = CustomPropertyValues::getCustomPropertyValue($feng_payment->getId(), $payment_bill_no->getId());
		if ($cpval instanceof CustomPropertyValue) {
			$bill_no = $cpval->getValue();
		}
	}

	$doc_num_sel = 0;
	$config_option_exists = CustomProperties::findOne(array('conditions' => "code='invoice_number_vendor'"));

	if ($config_option_exists instanceof CustomProperty) {
		$invoice_number = CustomProperties::getCustomPropertyByCode($payment_receipt_obj->getId(), "invoice_number_vendor");
		if ($invoice_number instanceof CustomProperty) {
			$cpval = CustomPropertyValues::getCustomPropertyValue($feng_payment->getId(), $invoice_number->getId());
			if ($cpval instanceof CustomPropertyValue) {
				$doc_num_sel = $cpval->getValue();
			}
		}
	}else if($bill_no){
		$doc_num_sel = $bill_no;
	}else{
		$doc_num_sel = $feng_payment->getId();
	}

	// invoice_number_vendor 
	$request = [
		"DocNumber" => $doc_num_sel
	];

	// set paid by id or vendor
	if ($feng_payment->getExpenseType() == 'reimbursable') {
/*		$qbo_employee_obj = sync_contact_with_qbo($feng_payment->getPaidById(), $dataService);
		if (is_array($qbo_employee_obj) && count($qbo_employee_obj) > 0) {
			$qbo_employee_obj = $qbo_employee_obj[0];
		}
*/
		$paid_by_contact = Contacts::findById($feng_payment->getPaidById());
		if ($paid_by_contact) {
			$qbo_vendor_obj = get_qbo_vendor_object($dataService, null, $paid_by_contact);
		}
		
		if ($qbo_vendor_obj) {
			$request["VendorRef"] = [
				"name" => $qbo_vendor_obj->DisplayName,
				"value" => $qbo_vendor_obj->Id,
			];
		}
		
	} else {

		if ($supplier_member instanceof Member) {
			$qbo_vendor_obj = get_qbo_vendor_object($dataService, $supplier_member);
			
			if ($qbo_vendor_obj) {
				$request["VendorRef"] = [
					"name" => $qbo_vendor_obj->DisplayName,
					"value" => $qbo_vendor_obj->Id,
				];
			}
		}
	}
	// --

	// set description
	if (trim($feng_payment->getDescription()) != '') {
		$request["PrivateNote"] = escape_single_quotes(trim($feng_payment->getDescription()));
	}
	// --


	// date and due date
	if ($feng_payment->getDate() instanceof DateTimeValue) {
		$request["TxnDate"] = $feng_payment->getDate()->format("Y-m-d");
	}

	if ($feng_payment->getDate() instanceof DateTimeValue) {
		$request["DueDate"] = $feng_payment->getDueDate()->format("Y-m-d");
	}
	// --

	// payment terms
	Env::useHelper("functions", "expenses2");
	$term_name = actual_expense_get_payment_term_name($feng_payment->getPaymentTermId());
	if (trim($term_name) != '') {
		$qbo_term_obj = get_qbo_payment_term_by_name($dataService, $term_name);
		if ($qbo_term_obj) {
			$request["SalesTermRef"] = [
				"value" => $qbo_term_obj->Id,
				"name" => $qbo_term_obj->Name,
			];
		}
	}
	// --

	if ($qbo_bill_id > 0) {
		// get the payment's SyncToken, it is needed to update it
		$old_qbo_payment_obj = get_qbo_actual_expense($dataService, $qbo_bill_id);
		if ($old_qbo_payment_obj) {
			$request['Id'] = $qbo_bill_id;
			$request['SyncToken'] = $old_qbo_payment_obj->SyncToken;
			$request['sparse'] = true; // only updates the sent information, does not override other data updated in qbo
		} else {
			$qbo_bill_id = 0;
		}
	}
/*
	$qbo_preferences = get_qbo_preferences($dataService);
	$send_item_based_line = $qbo_preferences->ProductAndServicesPrefs->ForPurchase;

	if ($send_item_based_line) {
*/
		// get quickbooks item
		$item = get_object_qbo_expense_item($feng_payment);
		if ($item == null) {
			$hours_array = get_qbo_items_definition($dataService, null, 'Direct Costs/Fees');
			if (is_array($hours_array) && count($hours_array) > 0) {
				$item = $hours_array[0];
			}
		}
		// --

		// expense status
		$bill_status = "NotBillable";
		if ($feng_payment->getIsBillable()) {
			if ($feng_payment->getColumnValue('invoicing_status') == 'invoiced') {
				$bill_status = "HasBeenBilled";
			} else {
				$bill_status = "Billable";
			}
		}
		// --

		// build line from item
		$line_item_detail = array(
			"TaxCodeRef" => array(
				"value" => "NON"
			),
			"BillableStatus" => $bill_status,
			"Qty" => $feng_payment->getQuantity(),
			"UnitPrice" => $feng_payment->getUnitCost(),
		);
		// set customer
		if ($entity_ref) {
			$line_item_detail["CustomerRef"] = [
				"name" => $entity_ref->DisplayName,
				"value" => $entity_ref->Id,
			];
		}
		// set item
		if ($item) {
			$line_item_detail["ItemRef"] = [
				"name" => $item ? $item->Name : '',
				"value" => $item ? $item->Id : '0'
			];
		}

		// set markup so qb can calculate the sales amount
		if ($feng_payment->getTotalPrice() != $feng_payment->getAmount() && $feng_payment->getAmount() > 0 && $feng_payment->getTotalPrice() > 0) {
			$markup = (($feng_payment->getTotalPrice() / $feng_payment->getAmount()) - 1 ) * 100;
			$line_item_detail["MarkupInfo"] = [
				"Percent" => number_format($markup, 3)
			];
		}

		if ($use_class_in_invoice_lines && $class_definition) {
			$line_item_detail["ClassRef"] = array(
					'value' => $class_definition->Id,
					'name' => $class_definition->Name,
			);
		}

		// add line
		$request['Line'] = array(
			"DetailType" =>"ItemBasedExpenseLineDetail", 
			"Amount" => $feng_payment->getAmount(),
			"ItemBasedExpenseLineDetail" => $line_item_detail,
		);

/*
	} else {

		// get quickbooks item account
		$item_account = get_object_qbo_expense_account($feng_payment);
		if ($item_account == null) {
			$item_account = get_qbo_account_by_name($dataService, "Cash and cash equivalents");			
		}
		// --

		$request['Line'] = array(
			"DetailType" =>"AccountBasedExpenseLineDetail", 
			"Amount" => $feng_payment->getAmount(),
			"AccountBasedExpenseLineDetail" => array(
				"TaxCodeRef" => array(
					"value" => "NON"
				),
				"AccountRef" => array(
					"name" => $item_account ? $item_account->Name : '',
					"value" => $item_account ? $item_account->Id : '0'
				),
				"BillableStatus" => $feng_payment->getIsBillable() ? "Billable" : "NotBillable",
			)
		);
	}
*/

	$qbo_bill_object = QuickBooksOnline\API\Facades\Bill::create($request);

	if (isset($request['Id'])) {
		// update in qbo
		$resulting_bill_obj = $dataService->Update($qbo_bill_object);
	} else {
		// add to qbo
		$resulting_bill_obj = $dataService->Add($qbo_bill_object);
	}

	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error uploading actual expense to Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}

	$feng_payment->setColumnValue('quickbooks_online_id', $resulting_bill_obj->Id);
	$feng_payment->setColumnValue('quickbooks_online_sync_token', $resulting_bill_obj->SyncToken);

	$reimburse_charge_obj = get_reimburse_charge_id_for_expense($dataService, $resulting_bill_obj, $entity_ref);
	if ($reimburse_charge_obj) {
		$feng_payment->setColumnValue('qbo_reimburse_charge_id', $reimburse_charge_obj->Id);
	}

	$feng_payment->save();

	return $feng_payment;
}

function get_qbo_actual_expense($dataService, $qbo_bill_id) {
	$bill_array = $dataService->Query("select * from Bill where Id='" . $qbo_bill_id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting Bill from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($bill_array) && sizeof($bill_array) > 0) {
			return current($bill_array);
		}
	}
}

function get_reimburse_charge_object_by_id($dataService, $qbo_reimburse_charge_id) {
	$rc_array = $dataService->Query("select * from ReimburseCharge where Id='" . $qbo_reimburse_charge_id . "'");
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting ReimburseCharge from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($rc_array) && sizeof($rc_array) > 0) {
			return current($rc_array);
		}
	}
}

function get_reimburse_charge_id_for_expense($dataService, $bill, $entity_ref) {
	
	$qbo_customer_id = $entity_ref ? $entity_ref->Id : 0;
	$rc_array = $dataService->Query("select * from ReimburseCharge where HasBeenInvoiced=false AND CustomerRef='$qbo_customer_id' ORDERBY Id DESC MAXRESULTS 1");
	
	$error = $dataService->getLastError();
	if ($error) {
		Logger::log_r("Error getting ReimburseCharge from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	} else {
		if (is_array($rc_array) && sizeof($rc_array) > 0) {
			return current($rc_array);
		}
	}
}


function delete_actual_expense_from_qbo(PaymentReceipt $actual_expense) {
	
	$dataService = get_qbo_data_service();

	$qbo_bill_data = array();
	$qbo_bill_id = $actual_expense->getColumnValue("quickbooks_online_id");

	$old_qbo_bill_obj = get_qbo_actual_expense($dataService, $qbo_bill_id);

	if ($old_qbo_bill_obj) {
		$qbo_bill_data['Id'] = $qbo_bill_id;
		$qbo_bill_data['SyncToken'] = $old_qbo_bill_obj->SyncToken;
		$qbo_bill_data['sparse'] = false; // only updates the sent information, does not override other data updated in qbo
	} else {
		return;
	}
	
	$entity = QuickBooksOnline\API\Facades\Bill::create($qbo_bill_data);
	
	$return_obj = $dataService->Delete($entity);
	
	$error = $dataService->getLastError();
	if ($error) {
		throw new Exception(build_qbo_error_message($error));
	}
	
	$actual_expense->setColumnValue("quickbooks_online_id", 0);
	$actual_expense->setColumnValue("quickbooks_online_sync_token", 0);
	$actual_expense->save();
	
	return $return_obj;
}

function qbo_sync_accounts_and_payment_methods(){
	$dataService = get_qbo_data_service();

	$all_types = QuickbooksAccountTypes::instance()->findAll();
	$all_accounts = QuickbooksAccounts::instance()->findAll();

	$final_accounts_qb = array();
	foreach($all_types as $type){
		$type_name = $type->getName();
		$type_id = $type->getAccountTypeId();

		$accounts = $dataService->Query("Select * from Account where AccountType = '$type_name'");
		$error = $dataService->getLastError();

		if ($error) {
			Logger::log_r("Error getting account type $type from Quickbooks");
			Logger::log_r($error);
			throw new Exception(build_qbo_error_message($error));
		}

		if(is_array($accounts))
		foreach($accounts as $account_information){
			$exists_account = QuickbooksAccounts::instance()->findOne(
				array("conditions" => array("quickbooks_id=?", $account_information->Id))
			);
			
			if($exists_account && $exists_account instanceof QuickbooksAccount){
				$exists_account->setName($account_information->Name);
				$exists_account->save();	
			}else{
				$qb_account = new QuickbooksAccount();
				$qb_account->setAccountTypeId($type_id);
				$qb_account->setName($account_information->Name);
				$qb_account->setQuickbooksId($account_information->Id);
				$qb_account->save();
			}

			$final_accounts_qb[] = $account_information->Id;
		}
	}

	// Turn off
	foreach($all_accounts as $account){
		if(!in_array($account->getQuickbooksId(), $final_accounts_qb) && $account instanceof QuickbooksAccount){
			$account->setVisible(0);
			$account->save();
		}else{
			$account->setVisible(1);
			$account->save();
		}
	}

	$all_payments = QuickbooksPaymentMethods::instance()->findAll();

	$payments = $dataService->Query("Select * from PaymentMethod");
	$error = $dataService->getLastError();

	if ($error) {
		Logger::log_r("Error getting PaymentMethod from Quickbooks");
		Logger::log_r($error);
		throw new Exception(build_qbo_error_message($error));
	}

	$final_payments_qb = array();
	if(is_array($payments))
	foreach($payments as $payment_information){

		$exists_account = QuickbooksPaymentMethods::instance()->findOne(
			array("conditions" => array("quickbooks_id=?", $payment_information->Id))
		);
		
		if($exists_account && $exists_account instanceof QuickbooksPaymentMethod){
			$exists_account->setName($payment_information->Name);
			$exists_account->save();	
		}else{
			$payment_method = new QuickbooksPaymentMethod();
			$payment_method->setQuickbooksId($payment_information->Id);
			$payment_method->setName($payment_information->Name);
			$payment_method->setType($payment_information->Type);
			$payment_method->setVisible(1);
			$payment_method->save();
		}

		$final_payments_qb[] = $payment_information->Id;
	}
	
	// Turn off
	foreach($all_payments as $payment){
		if(!in_array($payment->getQuickbooksId(), $final_payments_qb) && $payment instanceof QuickbooksPaymentMethod){
			$payment->setVisible(0);
			$payment->save();
		}else{
			$payment->setVisible(1);
			$payment->save();
		}
	}
}